/*   X File ǂݍ  Ver 0.50   */
/*                               */
/*                               */
/*   by Atsushi       1999/9/9   */


#include "basic.h"
#include "dxfile.h"

#include "lib/math/Vector3.h"
#include "lib/math/Matrix4.h"
#include "lib/math/TransformMatrix.h"
#include "lib/math/Bounds.h"
#include "lib/math/BoundingSphere.h"
#include "lib/math/BoundingBox.h"
#include "lib/util/Enumeration.h"
#include "lib/util/HashtableEnumeration.h"
#include "lib/util/Hashtable.h"
#include "ImageLoader.h"
#include "TextureLoader.h"
#include "XFileLoader.h"
#include "XFileObjectNode.h"
#include "XFileObjectMesh.h"
#include "XFileObjectFrame.h"
#include "XFileObject.h"
#include "XFileObjectCompiler.h"
#include "CompiledXFileObject.h"


/*   ---   RXgN^   ---   */

CompiledXFileObject::CompiledXFileObject(XFileObjectNode* shapeTree,
                                         XFileObjectMesh** shapeMeshes,
                                         int shapeMeshLength)
                    :XFileObject(shapeTree)
{
	this->shapeMeshes    =shapeMeshes;
	this->shapeMeshLength=shapeMeshLength;
	
	pointCache=new Hashtable<String,TransformMatrix>();
}


/*   ---   `   ---   */

void CompiledXFileObject::paint(IDirect3DDevice7* device)
{
	for (int i=0;i<=shapeMeshLength-1;i++)
	{
		shapeMeshes[i]->paint(device);
	}
}


/*   ---   m[h   ---   */

void CompiledXFileObject::detachNode(String* name)
{
	XFileObjectCompiler* compiler;

	XFileObject::detachNode(name);
	
	compiler=new XFileObjectCompiler();
	compiler->recompile(this);
	delete compiler;
}


/*   ---      ---   */

TransformMatrix* CompiledXFileObject::getTransformMatrix(String* name)
{
	TransformMatrix* point;

	if (!pointCache->containsKey(name))
	{
		point=XFileObject::getTransformMatrix(name);
		if (point!=NULL)
		{
			pointCache->put(new String(name),point);
		}
	}
	else
	{
		point=pointCache->get(name);
	}
	
	if (point==NULL)
	{
		return NULL;
	}
	else
	{
		return new TransformMatrix(point);
	}
}


/*   ---   ύX   ---   */

void CompiledXFileObject::setAlpha(int value)
{
	D3DMATERIAL7 material;

	for (int i=0;i<=shapeMeshLength-1;i++)
	{
		material=shapeMeshes[i]->getMaterial();
		material.diffuse.a=value/255.0f;
		shapeMeshes[i]->setMaterial(material);
	}
}


/*   ---   ݒ   ---   */

void CompiledXFileObject::setCompiledData(XFileObjectMesh** shapeMeshes,
                                          int shapeMeshLength)
{
	for (int i=0;i<=shapeMeshLength-1;i++)
	{
		delete this->shapeMeshes[i];
	}
	delete[] this->shapeMeshes;
	
	this->shapeMeshes    =shapeMeshes;
	this->shapeMeshLength=shapeMeshLength;
}


/*   ---   fXgN^   ---   */

CompiledXFileObject::~CompiledXFileObject(void)
{
	for (int i=0;i<=shapeMeshLength-1;i++)
	{
		delete shapeMeshes[i];
	}
	delete[] shapeMeshes;
	
	pointCache->deleteAll();
	delete pointCache;
}
