/*   DirectX7 ǗNX  Ver 0.80   */
/*                                   */
/*                                   */
/*   by Atsushi         1999/11/22   */


#include "basic.h"

#include "DirectXException.h"
#include "DirectXManager.h"


/*   ---   RXgN^   ---   */

DirectXManager::DirectXManager(HWND window)
{
	this->window=window;
	
	ddraw=NULL;
	primary=NULL;
	backBuffer=NULL;
	zBuffer=NULL;
	clipper=NULL;
	
	d3d=NULL;
	device=NULL;
}


/*   ---      ---   */

void DirectXManager::init2D(int width,int height,boolean fullscreen)
{
	init2D(width,height,16,fullscreen);
}

void DirectXManager::init3D(int width,int height,boolean fullscreen)
{
	init3D(width,height,16,fullscreen);
}

void DirectXManager::init2D(int width,int height,int bpp,boolean fullscreen)
{
	this->fullscreen=fullscreen;
	
	initDDraw(width,height,bpp,fullscreen);
}

void DirectXManager::init3D(int width,int height,int bpp,boolean fullscreen)
{
	this->fullscreen=fullscreen;
	
	initDDraw(width,height,bpp,fullscreen);
	initD3D(width,height,fullscreen);
}


/*   ---   DirectDraw    ---   */

void DirectXManager::initDDraw(int width,int height,int bpp,boolean fullscreen)
{
	DDSURFACEDESC2 ddsd;
	DDSCAPS2 ddscaps;

	if (ddraw==NULL)
	{
		if (DirectDrawCreateEx(NULL,(LPVOID*)(&ddraw),
		                       IID_IDirectDraw7,NULL)!=DD_OK)
		{
			throw new DirectXException();
		}
	}
	
	if (fullscreen)
	{
		ddraw->SetCooperativeLevel(window,
		    DDSCL_FULLSCREEN | DDSCL_EXCLUSIVE |
		    DDSCL_NOWINDOWCHANGES | DDSCL_ALLOWREBOOT);
	}
	else
	{
		ddraw->SetCooperativeLevel(window,DDSCL_NORMAL);
	}
	
	if (fullscreen)
	{
		ddraw->SetDisplayMode(width,height,bpp,0,0);
		
		ZeroMemory(&ddsd,sizeof(DDSURFACEDESC2));
		ddsd.dwSize=sizeof(DDSURFACEDESC2);
		ddsd.dwFlags=DDSD_CAPS | DDSD_BACKBUFFERCOUNT;
		ddsd.ddsCaps.dwCaps=
		    DDSCAPS_PRIMARYSURFACE | DDSCAPS_FLIP | DDSCAPS_COMPLEX |
		    DDSCAPS_3DDEVICE;
		ddsd.dwBackBufferCount=1;
		
		ddraw->CreateSurface(&ddsd,&primary,NULL);
		
		ZeroMemory(&ddscaps,sizeof(DDSCAPS2));
		ddscaps.dwCaps=DDSCAPS_BACKBUFFER | DDSCAPS_VIDEOMEMORY;
		primary->GetAttachedSurface(&ddscaps,&backBuffer);
	}
	else
	{
		ZeroMemory(&ddsd,sizeof(DDSURFACEDESC2));
		ddsd.dwSize=sizeof(DDSURFACEDESC2);
		ddsd.dwFlags=DDSD_CAPS;
		ddsd.ddsCaps.dwCaps=DDSCAPS_PRIMARYSURFACE;
		ddsd.ddsCaps.dwCaps2=0;
		
		ddraw->CreateSurface(&ddsd,&primary,NULL);
		
		ddsd.dwFlags=DDSD_CAPS | DDSD_WIDTH | DDSD_HEIGHT;
		ddsd.dwWidth =width;
		ddsd.dwHeight=height;
		ddsd.ddsCaps.dwCaps=DDSCAPS_OFFSCREENPLAIN | DDSCAPS_SYSTEMMEMORY |
		                    DDSCAPS_3DDEVICE;
		
		ddraw->CreateSurface(&ddsd,&backBuffer,NULL);
	}
	
	if (!fullscreen)
	{
		ddraw->CreateClipper(0,&clipper,NULL);
		clipper->SetHWnd(0,window);
		primary->SetClipper(clipper);
	}
}


/*   ---   Direct3D    ---   */

class DeviceData
{
public:
	GUID FAR guid;
	boolean fullscreen;
	boolean hal;
	int priority;
};

void DirectXManager::initD3D(int width,int height,boolean fullscreen)
{
	DeviceData* data;
	DDPIXELFORMAT pixelFormat;
	DDSURFACEDESC2 ddsd;

	if (d3d==NULL)
	{
		if (ddraw->QueryInterface(IID_IDirect3D7,(LPVOID*)(&d3d))!=S_OK)
		{
			throw new DirectXException();
		}
	}
	
	data=new DeviceData();
	data->fullscreen=fullscreen;
	data->hal=FALSE;
	data->priority=-1;
	pixelFormat.dwZBufferBitDepth=0;
	d3d->EnumDevices(D3DEnumDevicesCallback,data);
	d3d->EnumZBufferFormats(data->guid,D3DEnumPixelFormatsCallback,
	                        &pixelFormat);
	
	ZeroMemory(&ddsd,sizeof(DDSURFACEDESC2));
	ddsd.dwSize=sizeof(DDSURFACEDESC2);
	ddsd.dwFlags=DDSD_CAPS | DDSD_WIDTH | DDSD_HEIGHT | DDSD_PIXELFORMAT;
	ddsd.dwWidth =width;
	ddsd.dwHeight=height;
	if (data->hal && fullscreen)
	{
		ddsd.ddsCaps.dwCaps=DDSCAPS_ZBUFFER | DDSCAPS_VIDEOMEMORY;
	}
	else
	{
		ddsd.ddsCaps.dwCaps=DDSCAPS_ZBUFFER | DDSCAPS_SYSTEMMEMORY;
	}
	memcpy(&ddsd.ddpfPixelFormat,&pixelFormat,sizeof(DDPIXELFORMAT));
	
	ddraw->CreateSurface(&ddsd,&zBuffer,NULL);
	backBuffer->AddAttachedSurface(zBuffer);
	
	hal=(data->hal && fullscreen);
	if (d3d->CreateDevice(data->guid,backBuffer,&device)!=DD_OK)
	{
		if (d3d->CreateDevice(IID_IDirect3DRGBDevice,
		                      backBuffer,&device)!=DD_OK)
		{
			throw new DirectXException();
		}
	}
}

HRESULT CALLBACK DirectXManager::D3DEnumDevicesCallback(
    LPSTR deviceDescription,LPSTR deviceName,
    D3DDEVICEDESC7* desc,void* userData)
{
	DeviceData* data;
	int priority;
	boolean hal;

	if (IsEqualGUID(desc->deviceGUID,
	                IID_IDirect3DNullDevice)) return D3DENUMRET_OK;
	
	data=(DeviceData*)(userData);
	hal=(desc->dwDevCaps & D3DDEVCAPS_HWRASTERIZATION!=0);
	
	priority=(hal ? 10000 : 0)+
	         (desc->dwDevCaps & D3DDEVCAPS_HWTRANSFORMANDLIGHT ? 10000 : 0);
	
	priority+=(desc->dwDeviceZBufferBitDepth & DDBD_8  ? 1 : 0)+
	          (desc->dwDeviceZBufferBitDepth & DDBD_16 ? 2 : 0)+
	          (desc->dwDeviceZBufferBitDepth & DDBD_24 ? 4 : 0)+
	          (desc->dwDeviceZBufferBitDepth & DDBD_32 ? 8 : 0);
	priority+=(desc->dpcLineCaps.dwRasterCaps &
	           D3DPRASTERCAPS_WBUFFER ? 16 : 0);
	priority+=(desc->dpcTriCaps.dwRasterCaps &
	           D3DPRASTERCAPS_WBUFFER ? 16 : 0);
	
	if (data->priority<priority)
	{
		memcpy(&(data->guid),&(desc->deviceGUID),sizeof(GUID));
		data->priority=priority;
		data->hal=hal;
	}
	
	return D3DENUMRET_OK;
}

HRESULT CALLBACK DirectXManager::D3DEnumPixelFormatsCallback(
    DDPIXELFORMAT* pixelFormat,VOID* userData)
{
	DDPIXELFORMAT* data;

	data=(DDPIXELFORMAT*)(userData);
	if (pixelFormat->dwFlags & DDPF_ZBUFFER)
	{
		if (pixelFormat->dwZBufferBitDepth==16)
		{
			memcpy(data,pixelFormat,sizeof(DDPIXELFORMAT));
		}
	}
	
	return D3DENUMRET_OK;
}


/*   ---   DirectDraw   ---   */

IDirectDraw7* DirectXManager::getDirectDraw(void)
{
	return ddraw;
}

IDirectDrawSurface7* DirectXManager::getPrimarySurface(void)
{
	return primary;
}

IDirectDrawSurface7* DirectXManager::getBackBuffer(void)
{
	return backBuffer;
}

IDirectDrawSurface7* DirectXManager::createSurface(DDSURFACEDESC2* desc)
{
	IDirectDrawSurface7* surface;

	ddraw->CreateSurface(desc,&surface,NULL);
	
	return surface;
}

IDirectDrawPalette* DirectXManager::createPalette(DWORD flags,
                                                  PALETTEENTRY* entry)
{
	IDirectDrawPalette* palette;

	ddraw->CreatePalette(flags,entry,&palette,NULL);
	
	return palette;
}

IDirectDrawClipper* DirectXManager::createClipper(void)
{
	IDirectDrawClipper* clipper;

	ddraw->CreateClipper(0,&clipper,NULL);
	
	return clipper;
}


/*   ---   Direct3D   ---   */

IDirect3D7* DirectXManager::getDirect3D(void)
{
	return d3d;
}

IDirect3DDevice7* DirectXManager::get3DDevice(void)
{
	return device;
}


/*   ---   tXN[   ---   */

void DirectXManager::changeDisplayMode(int width,int height,
                                       boolean fullscreen)
{
	release();
	
	init3D(width,height,fullscreen);
}

boolean DirectXManager::isFullscreen(void)
{
	return fullscreen;
}


/*   ---   HAL/HEL   ---   */

boolean DirectXManager::isHAL(void)
{
	return hal;
}


/*   ---   A   ---   */

void DirectXManager::restore(void)
{
	if (ddraw!=NULL) ddraw->RestoreAllSurfaces();
}


/*   ---   ㏈   ---   */

void DirectXManager::release(void)
{
	releaseD3D();
	releaseDDraw();
}

void DirectXManager::releaseDDraw(void)
{
	if (clipper   !=NULL) { clipper   ->Release(); clipper   =NULL; }
	if (zBuffer   !=NULL) { zBuffer   ->Release(); zBuffer   =NULL; }
	if (backBuffer!=NULL) { backBuffer->Release(); backBuffer=NULL; }
	if (primary   !=NULL) { primary   ->Release(); primary   =NULL; }
}

void DirectXManager::releaseD3D(void)
{
	if (device!=NULL) { device->Release(); device=NULL; }
}


/*   ---   fXgN^   ---   */

DirectXManager::~DirectXManager(void)
{
	release();
	
	if (ddraw!=NULL)
	{
		ddraw->Release();
		ddraw=NULL;
	}
	if (d3d!=NULL)
	{
		d3d->Release();
		d3d=NULL;
	}
}
