/*   摜ǂݍ  Ver 0.60   */
/*                            */
/*                            */
/*   by Atsushi  1999/10/20   */


#include "basic.h"

#include "lib/io/IOException.h"
#include "lib/io/FileNotFoundException.h"
#include "ImageLoader.h"


/*   ---   RXgN^   ---   */

ImageLoader::ImageLoader(IDirectDraw7* ddraw,HINSTANCE hInstance)
{
	this->ddraw=ddraw;
	this->hInstance=hInstance;
}


/*   ---   ǂݍ   ---   */

IDirectDrawSurface7* ImageLoader::load(WORD resource)
{
	HBITMAP bitmap;

	bitmap=LoadBitmap(hInstance,MAKEINTRESOURCE(resource));
	
	if (bitmap==NULL)
	{
		throw new FileNotFoundException(String::valueOf(resource));
	}
	
	return load(bitmap);
}

IDirectDrawSurface7* ImageLoader::load(String* filename)
{
	HBITMAP bitmap;

	bitmap=(HBITMAP)(LoadImage(hInstance,filename->toCharArray(),
	                           IMAGE_BITMAP,0,0,
	                           LR_LOADFROMFILE | LR_CREATEDIBSECTION));
	
	if (bitmap==NULL)
	{
		throw new FileNotFoundException(filename);
	}
	
	return load(bitmap);
}

IDirectDrawSurface7* ImageLoader::load(char* filename)
{
	HBITMAP bitmap;

	bitmap=(HBITMAP)(LoadImage(hInstance,filename,IMAGE_BITMAP,0,0,
	                           LR_LOADFROMFILE | LR_CREATEDIBSECTION));
	
	if (bitmap==NULL)
	{
		throw new FileNotFoundException(new String(filename));
	}
	
	return load(bitmap);
}

IDirectDrawSurface7* ImageLoader::load(HBITMAP bitmap)
{
	IDirectDrawSurface7* image;
	BITMAP data;
	DDSURFACEDESC2 ddsd;

	GetObject(bitmap,sizeof(BITMAP),&data);
	ZeroMemory(&ddsd,sizeof(DDSURFACEDESC2));
	ddsd.dwSize=sizeof(DDSURFACEDESC2);
	ddsd.dwFlags=DDSD_CAPS | DDSD_WIDTH | DDSD_HEIGHT;
	ddsd.dwWidth =data.bmWidth;
	ddsd.dwHeight=data.bmHeight;
	ddsd.ddsCaps.dwCaps=DDSCAPS_OFFSCREENPLAIN;
	ddsd.ddsCaps.dwCaps2=0;
	
	ddraw->CreateSurface(&ddsd,&image,NULL);
	
	copyBitmap(image,bitmap);
	setPalette(image,bitmap);
	
	DeleteObject(bitmap);
	
	return image;
}


/*   ---   ǂݍ (Color Key)   ---   */

IDirectDrawSurface7* ImageLoader::load(WORD resource,DDCOLORKEY colorKey)
{
	HBITMAP bitmap;

	bitmap=LoadBitmap(hInstance,MAKEINTRESOURCE(resource));
	
	if (bitmap==NULL)
	{
		throw new FileNotFoundException(String::valueOf(resource));
	}
	
	return load(bitmap,colorKey);
}

IDirectDrawSurface7* ImageLoader::load(String* filename,DDCOLORKEY colorKey)
{
	HBITMAP bitmap;

	bitmap=(HBITMAP)(LoadImage(hInstance,filename->toCharArray(),
	                           IMAGE_BITMAP,0,0,
	                           LR_LOADFROMFILE | LR_CREATEDIBSECTION));
	
	if (bitmap==NULL)
	{
		throw new FileNotFoundException(filename);
	}
	
	return load(bitmap,colorKey);
}

IDirectDrawSurface7* ImageLoader::load(char* filename,DDCOLORKEY colorKey)
{
	HBITMAP bitmap;

	bitmap=(HBITMAP)(LoadImage(hInstance,filename,IMAGE_BITMAP,0,0,
	                           LR_LOADFROMFILE | LR_CREATEDIBSECTION));
	
	if (bitmap==NULL)
	{
		throw new FileNotFoundException(new String(filename));
	}
	
	return load(bitmap,colorKey);
}

IDirectDrawSurface7* ImageLoader::load(HBITMAP bitmap,DDCOLORKEY colorKey)
{
	IDirectDrawSurface7* image;
	BITMAP data;
	DDSURFACEDESC2 ddsd;

	GetObject(bitmap,sizeof(BITMAP),&data);
	ZeroMemory(&ddsd,sizeof(DDSURFACEDESC2));
	ddsd.dwSize=sizeof(DDSURFACEDESC2);
	ddsd.dwFlags=DDSD_CAPS | DDSD_WIDTH | DDSD_HEIGHT | DDSD_CKSRCBLT;
	ddsd.dwWidth =data.bmWidth;
	ddsd.dwHeight=data.bmHeight;
	ddsd.ddckCKSrcBlt=colorKey;
	ddsd.ddsCaps.dwCaps=DDSCAPS_OFFSCREENPLAIN;
	ddsd.ddsCaps.dwCaps2=0;
	
	ddraw->CreateSurface(&ddsd,&image,NULL);
	
	copyBitmap(image,bitmap);
	setPalette(image,bitmap);
	
	DeleteObject(bitmap);
	
	return image;
}


/*   ---   Bitmap    ---   */

void ImageLoader::copyBitmap(IDirectDrawSurface7* surface,HBITMAP bitmap)
{
	IDirectDrawSurface7* temp;
	IDirectDrawPalette* palette;
	HDC imageDC,dc;
	BITMAP data;
	DDSURFACEDESC2 ddsd;
	DWORD table[256];
	int length;

	GetObject(bitmap,sizeof(BITMAP),&data);
	
	ddsd.dwSize=sizeof(DDSURFACEDESC2);
	ddsd.ddpfPixelFormat.dwSize=sizeof(DDPIXELFORMAT);
	surface->GetSurfaceDesc(&ddsd);
	
	ddsd.dwFlags=DDSD_CAPS | DDSD_WIDTH | DDSD_HEIGHT | DDSD_PIXELFORMAT;
	ddsd.dwWidth =data.bmWidth;
	ddsd.dwHeight=data.bmHeight;
	ddsd.ddsCaps.dwCaps =DDSCAPS_TEXTURE | DDSCAPS_SYSTEMMEMORY;
	ddsd.ddsCaps.dwCaps2=0;
	ddraw->CreateSurface(&ddsd,&temp,NULL);
	
	imageDC=CreateCompatibleDC(NULL);
	SelectObject(imageDC,bitmap);
	
	if (ddsd.ddpfPixelFormat.dwRGBBitCount==8)
	{
		length=GetDIBColorTable(imageDC,0,256,(RGBQUAD*)(table));
		for (int i=0;i<=length-1;i++)
		{
			table[i]=0xff000000+RGB(GetBValue(table[i]),
			                        GetGValue(table[i]),
			                        GetRValue(table[i]));
/*          Set alpha for transparent pixels
            if( dwFlags & D3DTEXTR_TRANSPARENTBLACK )
            {
                if( (table[i]&0x00ffffff) == 0x00000000 )
                    table[i] &= 0x00ffffff;
            }
            else if( dwFlags & D3DTEXTR_TRANSPARENTWHITE )
            {
                if( (table[i]&0x00ffffff) == 0x00ffffff )
                    table[i] &= 0x00ffffff;
            }
*/
		}
/*
        if( dwFlags & (D3DTEXTR_TRANSPARENTWHITE|D3DTEXTR_TRANSPARENTBLACK) )
            ddraw->CreatePalette( DDPCAPS_8BIT|DDPCAPS_ALPHA, (PALETTEENTRY*)table, &palette, NULL );
        else
*/
		ddraw->CreatePalette(DDPCAPS_8BIT,(PALETTEENTRY*)(table),
		                      &palette,NULL);
		temp->SetPalette(palette);
		surface->SetPalette(palette);
		palette->Release();
	}
	
	temp->GetDC(&dc);
	StretchBlt(dc,0,0,ddsd.dwWidth,ddsd.dwHeight,
	           imageDC,0,0,data.bmWidth,data.bmHeight,SRCCOPY);
	temp->ReleaseDC(dc);
	
	DeleteDC(imageDC);
	
	surface->Blt(NULL,temp,NULL,0,NULL);
	
	if (ddsd.ddpfPixelFormat.dwRGBBitCount==8)
	{
		temp->GetPalette(&palette);
		palette->GetEntries(0,0,256,(PALETTEENTRY*)(&table));
		palette->Release();
		
		surface->GetPalette(&palette);
		palette->SetEntries(0,0,256,(PALETTEENTRY*)(&table));
		palette->Release();
	}
	
	temp->Release();

/*
    // For textures with real alpha (not palettized), set transparent bits
    if( ddsd.ddpfPixelFormat.dwRGBAlphaBitMask )
    {
        if( dwFlags & (D3DTEXTR_TRANSPARENTWHITE|D3DTEXTR_TRANSPARENTBLACK) )
        {
            // Lock the texture surface
            DDSURFACEDESC2 ddsd;
            D3DUtil_InitSurfaceDesc( ddsd );
            while( pddsTarget->Lock( NULL, &ddsd, 0, NULL ) ==
                   DDERR_WASSTILLDRAWING );
             
            DWORD dwAlphaMask = ddsd.ddpfPixelFormat.dwRGBAlphaBitMask;
            DWORD dwRGBMask   = ( ddsd.ddpfPixelFormat.dwRBitMask |
                                  ddsd.ddpfPixelFormat.dwGBitMask |
                                  ddsd.ddpfPixelFormat.dwBBitMask );
            DWORD dwColorkey  = 0x00000000; // Colorkey on black
            if( dwFlags & D3DTEXTR_TRANSPARENTWHITE ) 
                dwColorkey = dwRGBMask;     // Colorkey on white

            // Add an opaque alpha value to each non-colorkeyed pixel
            for( DWORD y=0; y<ddsd.dwHeight; y++ )
            {
                WORD*  p16 =  (WORD*)((BYTE*)ddsd.lpSurface + y*ddsd.lPitch);
                DWORD* p32 = (DWORD*)((BYTE*)ddsd.lpSurface + y*ddsd.lPitch);
        
                for( DWORD x=0; x<ddsd.dwWidth; x++ )
                {
                    if( ddsd.ddpfPixelFormat.dwRGBBitCount == 16 )
                    {
                        if( ( *p16 &= dwRGBMask ) != dwColorkey )
                            *p16 |= dwAlphaMask;
                        p16++;
                    }
                    if( ddsd.ddpfPixelFormat.dwRGBBitCount == 32 )
                    {
                        if( ( *p32 &= dwRGBMask ) != dwColorkey )
                            *p32 |= dwAlphaMask;
                        p32++;
                    }
                }
            }
            pddsTarget->Unlock( NULL );
        }
    }
*/
}

void ImageLoader::setPalette(IDirectDrawSurface7* surface,HBITMAP bitmap)
{
	LPDIRECTDRAWPALETTE palette;
	HDC dc;
	DDSURFACEDESC2 ddsd;
	DWORD table[256];
	UINT length;

	ddsd.dwSize=sizeof(DDSURFACEDESC2);
	surface->GetSurfaceDesc(&ddsd);
	
	if (ddsd.ddpfPixelFormat.dwRGBBitCount==8)
	{
		dc=CreateCompatibleDC(NULL);
		
		length=GetDIBColorTable(dc,0,256,(RGBQUAD*)(table));
		for (int i=0;i<=length-1;i++)
		{
			table[i]=0xff000000+RGB(GetBValue(table[i]),
			                        GetGValue(table[i]),
			                        GetRValue(table[i]));
		}
		ddraw->CreatePalette(DDPCAPS_8BIT,(PALETTEENTRY*)(table),
		                     &palette,NULL);
		surface->SetPalette(palette);
		
		DeleteDC(dc);
	}
}


/*   ---   fXgN^   ---   */

ImageLoader::~ImageLoader(void)
{
}
