/*   eNX`[ǂݍ  Ver 0.60   */
/*                                    */
/*                                    */
/*   by Atsushi          1999/11/23   */


#include "basic.h"

#include "DirectXException.h"
#include "ImageLoader.h"
#include "TextureLoader.h"


/*   ---   RXgN^   ---   */

TextureLoader::TextureLoader(IDirectDraw7* ddraw,IDirect3DDevice7* device,
                             HINSTANCE hInstance)
{
	this->ddraw    =ddraw;
	this->device   =device;
	this->hInstance=hInstance;
	
	imageLoader=new ImageLoader(ddraw,hInstance);
	
	init();
}


/*   ---      ---   */

void TextureLoader::init(void)
{
	for (int i=0;i<=31;i++) format.bit[i]=FALSE;
	format.priority=-1;
	
	device->EnumTextureFormats(D3DEnumPixelFormatsCallback,&format);
}

HRESULT CALLBACK TextureLoader::D3DEnumPixelFormatsCallback(
    DDPIXELFORMAT* pixelFormat,VOID* userData)
{
	TEXTURE_FORMAT* format;
	int priority;

	format=(TEXTURE_FORMAT*)(userData);
	if (!(pixelFormat->dwFlags & DDPF_RGB))
	{
		return D3DENUMRET_OK;
	}
	
	format->bit[pixelFormat->dwRGBBitCount-1]=TRUE;
	
/*
	if (pixelFormat->dwFlags & (DDPF_PALETTEINDEXED1 | DDPF_PALETTEINDEXED2 |
	                            DDPF_PALETTEINDEXED4 | DDPF_PALETTEINDEXED8))
	{
		return D3DENUMRET_OK;
	}
	if (pixelFormat->dwFlags & (DDPF_LUMINANCE | DDPF_ALPHA))
	{
		return D3DENUMRET_OK;
	}
	if (pixelFormat->dwFlags & (DDPF_YUV | DDPF_FOURCC))
	{
		return D3DENUMRET_OK;
	}
	if (pixelFormat->dwFlags & (DDPF_LUMINANCE | DDPF_BUMPLUMINANCE |
	                            DDPF_BUMPDUDV))
	{
		return D3DENUMRET_OK;
	}
	if (pixelFormat->dwFourCC!=0)
	{
		return D3DENUMRET_OK;
	}
*/
	
	priority=(pixelFormat->dwRGBBitCount==16 ? 3 : 0)+
	         (pixelFormat->dwRGBBitCount==24 ? 2 : 0)+
	         (pixelFormat->dwRGBBitCount==32 ? 1 : 0);
	if (!(pixelFormat->dwFlags & DDPF_ALPHAPIXELS)) priority+=4;
	
	if (format->priority<=priority)
	{
		memcpy(&(format->validFormat),pixelFormat,sizeof(DDPIXELFORMAT));
		format->priority=priority;
	}
	
	return D3DENUMRET_OK;
}


/*   ---   ǂݍ   ---   */

IDirectDrawSurface7* TextureLoader::load(WORD resource)
{
	return load(imageLoader->load(resource));
}

IDirectDrawSurface7* TextureLoader::load(String* filename)
{
	return load(imageLoader->load(filename));
}

IDirectDrawSurface7* TextureLoader::load(char* filename)
{
	return load(imageLoader->load(filename));
}

IDirectDrawSurface7* TextureLoader::load(WORD resource,DDCOLORKEY colorKey)
{
	return load(imageLoader->load(resource,colorKey));
}

IDirectDrawSurface7* TextureLoader::load(String* filename,DDCOLORKEY colorKey)
{
	return load(imageLoader->load(filename,colorKey));
}

IDirectDrawSurface7* TextureLoader::load(char* filename,DDCOLORKEY colorKey)
{
	return load(imageLoader->load(filename,colorKey));
}

IDirectDrawSurface7* TextureLoader::load(WORD* resource,int length)
{
	IDirectDrawSurface7** image;

	image=new IDirectDrawSurface7*[length];
	for (int i=0;i<=length-1;i++)
	{
		image[i]=imageLoader->load(resource[i]);
	}
	return load(image,length);
}

IDirectDrawSurface7* TextureLoader::load(String** filename,int length)
{
	IDirectDrawSurface7** image;

	image=new IDirectDrawSurface7*[length];
	for (int i=0;i<=length-1;i++)
	{
		image[i]=imageLoader->load(filename[i]);
	}
	return load(image,length);
}

IDirectDrawSurface7* TextureLoader::load(char** filename,int length)
{
	IDirectDrawSurface7** image;

	image=new IDirectDrawSurface7*[length];
	for (int i=0;i<=length-1;i++)
	{
		image[i]=imageLoader->load(filename[i]);
	}
	return load(image,length);
}

IDirectDrawSurface7* TextureLoader::load(WORD* resource,int length,
                                       DDCOLORKEY colorKey)
{
	IDirectDrawSurface7** image;

	image=new IDirectDrawSurface7*[length];
	for (int i=0;i<=length-1;i++)
	{
		image[i]=imageLoader->load(resource[i],colorKey);
	}
	return load(image,length);
}

IDirectDrawSurface7* TextureLoader::load(String** filename,int length,
                                       DDCOLORKEY colorKey)
{
	IDirectDrawSurface7** image;

	image=new IDirectDrawSurface7*[length];
	for (int i=0;i<=length-1;i++)
	{
		image[i]=imageLoader->load(filename[i],colorKey);
	}
	return load(image,length);
}

IDirectDrawSurface7* TextureLoader::load(char** filename,int length,
                                       DDCOLORKEY colorKey)
{
	IDirectDrawSurface7** image;

	image=new IDirectDrawSurface7*[length];
	for (int i=0;i<=length-1;i++)
	{
		image[i]=imageLoader->load(filename[i],colorKey);
	}
	return load(image,length);
}


/*   ---   ǂݍ   ---   */

IDirectDrawSurface7* TextureLoader::load(IDirectDrawSurface7* image0)
{
	D3DDEVICEDESC7 desc;
	DDSURFACEDESC2 ddsd0,ddsd;
	IDirectDrawSurface7* texture;
	IDirectDrawPalette* palette;
	D3DPRIMCAPS caps;

	desc.dpcLineCaps.dwSize=sizeof(D3DPRIMCAPS);
	desc.dpcTriCaps.dwSize=sizeof(D3DPRIMCAPS);
	device->GetCaps(&desc);
	caps=desc.dpcTriCaps;
	
	ddsd0.dwSize=sizeof(DDSURFACEDESC2);
	image0->GetSurfaceDesc(&ddsd0);
	
	ZeroMemory(&ddsd,sizeof(DDSURFACEDESC2));
	ddsd.dwSize=sizeof(DDSURFACEDESC2);
	ddsd.dwFlags=DDSD_CAPS | DDSD_WIDTH | DDSD_HEIGHT;
	ddsd.ddsCaps.dwCaps=DDSCAPS_TEXTURE;
	ddsd.ddpfPixelFormat.dwSize=sizeof(DDPIXELFORMAT);
	
	if (ddsd0.dwFlags & DDSD_CKSRCBLT)
	{
		ddsd.dwFlags|=DDSD_CKSRCBLT;
		ddsd.ddckCKSrcBlt=ddsd0.ddckCKSrcBlt;
	}
	
	if (caps.dwTextureCaps & D3DPTEXTURECAPS_POW2)
	{
		for (ddsd.dwWidth =1;ddsd.dwWidth <ddsd0.dwWidth ;ddsd.dwWidth <<=1);
		for (ddsd.dwHeight=1;ddsd.dwHeight<ddsd0.dwHeight;ddsd.dwHeight<<=1);
	}
	else
	{
		ddsd.dwWidth =ddsd0.dwWidth;
		ddsd.dwHeight=ddsd0.dwHeight;
	}
	if (caps.dwTextureCaps & D3DPTEXTURECAPS_SQUAREONLY)
	{
		if (ddsd.dwWidth>ddsd.dwHeight)
		{
			ddsd.dwHeight=ddsd.dwWidth;
		}
		else
		{
			ddsd.dwWidth=ddsd.dwHeight;
		}
	}
//	if (!format.bit[ddsd0.ddpfPixelFormat.dwRGBBitCount-1])
//	{
		memcpy(&(ddsd.ddpfPixelFormat),
		       &(format.validFormat),sizeof(DDPIXELFORMAT));
		ddsd.dwFlags|=DDSD_PIXELFORMAT;
		ddsd.ddpfPixelFormat.dwFlags=DDPF_RGB | DDPF_ALPHAPIXELS;
// TODO
//	}
	
	ddsd.ddsCaps.dwCaps=DDSCAPS_TEXTURE | DDSCAPS_VIDEOMEMORY;
	if (ddraw->CreateSurface(&ddsd,&texture,NULL)==DDERR_OUTOFVIDEOMEMORY)
	{
		ddsd.ddsCaps.dwCaps=DDSCAPS_TEXTURE | DDSCAPS_SYSTEMMEMORY;
		ddraw->CreateSurface(&ddsd,&texture,NULL);
	}
	
	if (image0->GetPalette(&palette)!=DDERR_NOPALETTEATTACHED)
	{
		texture->SetPalette(palette);
	}
	
	texture->Blt(NULL,image0,NULL,DDBLT_WAIT,NULL);
	
	image0->Release();
	
	return texture;
}

IDirectDrawSurface7* TextureLoader::load(IDirectDrawSurface7** image0,
                                         int length)
{
	D3DDEVICEDESC7 desc;
	DDSURFACEDESC2 ddsd0,ddsd;
	IDirectDrawSurface7* texture;
	IDirectDrawSurface7* imageAttached;
	IDirectDrawPalette* palette;
	D3DPRIMCAPS caps;
	int i;

	desc.dpcLineCaps.dwSize=sizeof(D3DPRIMCAPS);
	desc.dpcTriCaps.dwSize=sizeof(D3DPRIMCAPS);
	device->GetCaps(&desc);
	caps=desc.dpcTriCaps;
	if (!(caps.dwTextureFilterCaps & D3DPTFILTERCAPS_MIPFPOINT) &&
	    !(caps.dwTextureFilterCaps & D3DPTFILTERCAPS_MIPFLINEAR))
	{
		texture=load(image0[0]);
		for (i=1;i<=length-1;i++)
		{
			image0[i]->Release();
		}
		delete[] image0;
		return texture;
	}
	
	ddsd0.dwSize=sizeof(DDSURFACEDESC2);
	image0[0]->GetSurfaceDesc(&ddsd0);
	
	ZeroMemory(&ddsd,sizeof(DDSURFACEDESC2));
	ddsd.dwSize=sizeof(DDSURFACEDESC2);
	ddsd.dwFlags=DDSD_CAPS | DDSD_WIDTH | DDSD_HEIGHT | DDSD_MIPMAPCOUNT;
	ddsd.ddsCaps.dwCaps=DDSCAPS_TEXTURE | DDSCAPS_MIPMAP | DDSCAPS_COMPLEX;
	ddsd.dwWidth =ddsd0.dwWidth;
	ddsd.dwHeight=ddsd0.dwHeight;
	ddsd.dwMipMapCount=length; 
	ddsd.ddpfPixelFormat.dwSize=sizeof(DDPIXELFORMAT);
	
	if (ddsd0.dwFlags & DDSD_CKSRCBLT)
	{
		ddsd.dwFlags|=DDSD_CKSRCBLT;
		ddsd.ddckCKSrcBlt=ddsd0.ddckCKSrcBlt;
	}
	
	if (caps.dwTextureCaps & D3DPTEXTURECAPS_POW2)
	{
		for (ddsd.dwWidth =1;ddsd0.dwWidth >ddsd.dwWidth ;ddsd.dwWidth <<=1);
		for (ddsd.dwHeight=1;ddsd0.dwHeight>ddsd.dwHeight;ddsd.dwHeight<<=1);
	}
	if (caps.dwTextureCaps & D3DPTEXTURECAPS_SQUAREONLY)
	{
		if (ddsd.dwWidth>ddsd.dwHeight)
		{
			ddsd.dwHeight=ddsd.dwWidth;
		}
		else
		{
			ddsd.dwWidth=ddsd.dwHeight;
		}
	}
//	if (!format.bit[ddsd0.ddpfPixelFormat.dwRGBBitCount-1])
//	{
		memcpy(&(ddsd.ddpfPixelFormat),
		       &(format.validFormat),sizeof(DDPIXELFORMAT));
		ddsd.dwFlags|=DDSD_PIXELFORMAT;
		ddsd.ddpfPixelFormat.dwFlags=DDPF_RGB | DDPF_ALPHAPIXELS;
// TODO
//	}
	
	ddsd.ddsCaps.dwCaps=ddsd.ddsCaps.dwCaps | DDSCAPS_VIDEOMEMORY;
	if (ddraw->CreateSurface(&ddsd,&texture,NULL)==DDERR_OUTOFVIDEOMEMORY)
	{
		ddsd.ddsCaps.dwCaps=(ddsd.ddsCaps.dwCaps & !DDSCAPS_VIDEOMEMORY)
		                                         | DDSCAPS_SYSTEMMEMORY;
		ddraw->CreateSurface(&ddsd,&texture,NULL);
	}
	
	for (i=0,imageAttached=texture;i<=length-1;i++)
	{
		if (image0[i]->GetPalette(&palette)!=DDERR_NOPALETTEATTACHED)
		{
			imageAttached->SetPalette(palette);
		}
		
		imageAttached->Blt(NULL,image0[i],NULL,DDBLT_WAIT,NULL);
		imageAttached->GetAttachedSurface(&(ddsd.ddsCaps),&imageAttached);
	}
	
	for (i=0;i<=length-1;i++)
	{
		image0[i]->Release();
	}
	delete[] image0;
	
	return texture;
}


/*   ---   fXgN^   ---   */

TextureLoader::~TextureLoader(void)
{
	delete imageLoader;
}
