/*   X File ǂݍ  Ver 0.60   */
/*                               */
/*                               */
/*   by Atsushi     1999/10/20   */


#include "basic.h"
#include "rmxfguid.h"
#include "rmxftmpl.h"

#include "lib/math/Vector3.h"
#include "lib/math/Matrix4.h"
#include "lib/math/TransformMatrix.h"
#include "lib/util/Enumeration.h"
#include "lib/util/HashtableEnumeration.h"
#include "lib/util/Hashtable.h"
#include "lib/io/IOException.h"
#include "lib/io/FileNotFoundException.h"
#include "ImageLoader.h"
#include "TextureLoader.h"
#include "DirectXException.h"
#include "XFileObjectNode.h"
#include "XFileObjectMesh.h"
#include "XFileObjectFrame.h"
#include "XFileObject.h"
#include "XFileLoader.h"


/*   ---   RXgN^   ---   */

XFileLoader::XFileLoader(IDirectDraw7* ddraw,IDirect3D7* d3d,
                         IDirect3DDevice7* device,HINSTANCE hInstance)
{
	this->ddraw    =ddraw;
	this->d3d      =d3d;
	this->device   =device;
	this->hInstance=hInstance;
	
	textureLoader=new TextureLoader(ddraw,device,hInstance);
	textureCache=new Hashtable<String,IDirectDrawSurface7>();
	
	init();
}


/*   ---      ---   */

void XFileLoader::init(void)
{
	if (DirectXFileCreate(&xFile)!=DXFILE_OK)
	{
		throw new DirectXException();
	}
}


/*   ---   ǂݍ   ---   */

XFileObject* XFileLoader::load(WORD resource)
{
	IDirectXFileEnumObject* enumObject=NULL;
	DXFILELOADRESOURCE source;

	source.hModule=NULL;
	source.lpName=MAKEINTRESOURCE(resource);
	source.lpType="XFile";
	
	xFile->RegisterTemplates(D3DRM_XTEMPLATES,D3DRM_XTEMPLATE_BYTES);
	xFile->CreateEnumObject(&source,DXFILELOAD_FROMRESOURCE,&enumObject);
	
	if (enumObject==NULL)
	{
		throw new FileNotFoundException();
	}
	
	path=new String("");
	
	return load(enumObject);
}

XFileObject* XFileLoader::load(String* filename)
{
	IDirectXFileEnumObject* enumObject=NULL;

	xFile->RegisterTemplates(D3DRM_XTEMPLATES,D3DRM_XTEMPLATE_BYTES);
	xFile->CreateEnumObject(filename->toCharArray(),
	                        DXFILELOAD_FROMFILE,&enumObject);
	
	if (enumObject==NULL)
	{
		throw new FileNotFoundException(filename);
	}
	
	path=filename->substring(0,filename->lastIndexOf('/')+1);
	
	return load(enumObject);
}

XFileObject* XFileLoader::load(char* filename)
{
	IDirectXFileEnumObject* enumObject=NULL;
	String* filename1;

	xFile->RegisterTemplates(D3DRM_XTEMPLATES,D3DRM_XTEMPLATE_BYTES);
	xFile->CreateEnumObject(filename,DXFILELOAD_FROMFILE,&enumObject);
	
	if (enumObject==NULL)
	{
		throw new FileNotFoundException(new String(filename));
	}
	
	filename1=new String(filename);
	path=filename1->substring(0,filename1->lastIndexOf('/')+1);
	delete filename1;
	
	return load(enumObject);
}

XFileObject* XFileLoader::load(IDirectXFileEnumObject* enumObject)
{
	XFileObjectFrame* root;
	IDirectXFileData* fileData;
	const GUID* guid;

	root=new XFileObjectFrame(new String("root"));
	
	while (enumObject->GetNextDataObject(&fileData)==DXFILE_OK)
	{
		fileData->GetType(&guid);
		
		if (*guid==TID_D3DRMFrame)
		{
			parseFrame(fileData,root);
		}
		else if (*guid==TID_D3DRMMesh)
		{
			parseMesh(fileData,root);
		}
		
		fileData->Release();
	}
	
	enumObject->Release();
	
	delete path;
	
	return new XFileObject(root);
}


/*   ---      ---   */

void XFileLoader::parseFrame(IDirectXFileData* fileData,
                             XFileObjectFrame* parent)
{
	XFileObjectFrame* frame;
	IDirectXFileObject* child;
	IDirectXFileData* childData;
	void* data;
	char* name;
	DWORD nameLength,dataSize;
	const GUID* guid;

	name=new char[80];
	nameLength=80;
	fileData->GetName(name,&nameLength);
	frame=new XFileObjectFrame(new String(name,TRUE));
	
	while (fileData->GetNextObject(&child)==DXFILE_OK)
	{
		child->QueryInterface(IID_IDirectXFileData,(void**)(&childData));
		childData->GetType(&guid);
		
		if (*guid==TID_D3DRMFrame)
		{
			parseFrame(childData,frame);
		}
		else if (*guid==TID_D3DRMMesh)
		{
			parseMesh(childData,frame);
		}
		else if (*guid==TID_D3DRMFrameTransformMatrix)
		{
			childData->GetData(NULL,&dataSize,&data);
			
			if (dataSize==sizeof(D3DMATRIX))
			{
				frame->setMatrix(*((D3DMATRIX*)(data)));
			}
		}
		
		childData->Release();
		child->Release();
	}
	
	parent->addChild(frame);
}

void XFileLoader::parseMesh(IDirectXFileData* fileData,
                            XFileObjectFrame* parent)
{
	XFileObjectMesh* mesh;
	IDirectXFileObject* child;
	IDirectXFileData* childData;
	char* name;
	byte* data;
	DWORD nameLength,dataSize;
	D3DVECTOR* vertices;
	D3DVECTOR* normals;
	FLOAT* texCoords;
	VOID* face;
	int vertexLength,faceLength,normalLength,texCoordsLength,faceLength0;
	boolean hasNormals;
	const GUID* guid;

	name=new char[80];
	nameLength=80;
	fileData->GetName(name,&nameLength);
	fileData->GetData(NULL,&dataSize,(void**)(&data));
	
	vertexLength=*((DWORD*)(data));     data+=4;
	vertices    =((D3DVECTOR*)(data));  data+=12*vertexLength;
	faceLength  =*((DWORD*)(data));     data+=4;
	face        =(VOID*)(data);
	
	mesh=new XFileObjectMesh(new String(name,TRUE));
	parseMeshVertices(mesh,vertices,vertexLength,face,faceLength);
	
	faceLength0=faceLength;
	hasNormals=FALSE;
	
	while (fileData->GetNextObject(&child)==DXFILE_OK)
	{
		child->QueryInterface(IID_IDirectXFileData,(void**)(&childData));
		childData->GetType(&guid);
		
		childData->GetData(NULL,&dataSize,(void**)(&data));
		
		if (*guid==TID_D3DRMMeshMaterialList)
		{
			parseMeshMaterialList(childData,mesh);
		}
		else if (*guid==TID_D3DRMMeshNormals)
		{
			normalLength=*((DWORD*)(data));     data+=4;
			normals     =((D3DVECTOR*)(data));  data+=12*normalLength;
			faceLength  =*((DWORD*)(data));     data+=4;
			face        =(VOID*)(data);
			
			if (faceLength0==faceLength)
			{
				parseMeshNormals(mesh,normals,normalLength,face,faceLength);
				hasNormals=TRUE;
			}
		}
		else if (*guid==TID_D3DRMMeshTextureCoords)
		{
			texCoordsLength=*((DWORD*)(data));
			texCoords      =(FLOAT*)(data+4);
			
			if (texCoordsLength==vertexLength)
			{
				mesh->setTextureCoords(texCoords);
			}
		}
		
		childData->Release();
		child->Release();
	}
	
	if (!hasNormals) mesh->computeNormals();
	
	parent->addChild(mesh);
}

void XFileLoader::parseMeshVertices(XFileObjectMesh* mesh,
    D3DVECTOR* verticesData,int vertexDataLength,
    VOID* faces,int faceLength)
{
	D3DVERTEX* vertices;
	WORD* indices;
	int indexLength,k;
	byte* addr;
	DWORD verticesPerFace;
	WORD v1,v2,v3,vn;
	int i;

	vertices=new D3DVERTEX[vertexDataLength];
	
	for (i=0;i<=vertexDataLength-1;i++)
	{
		vertices[i].x=verticesData[i].x;
		vertices[i].y=verticesData[i].y;
		vertices[i].z=verticesData[i].z;
	}
	
	indexLength=0;
	addr=(byte*)(faces);
	for (i=0;i<=faceLength-1;i++)
	{
		verticesPerFace=*((DWORD*)(addr));
		
		addr+=4+4*verticesPerFace;
		indexLength+=3*(verticesPerFace-2);
	}
	
	indices=new WORD[indexLength];
	
	addr=(byte*)(faces);
	k=0;
	
	for (i=0;i<=faceLength-1;i++)
	{
		verticesPerFace=*((DWORD*)(addr));
		addr+=4;
		
		for (int j=0;j<=verticesPerFace-1;j++)
		{
			if (j==0)
			{
				v1=(WORD)(*((DWORD*)(addr)));
				indices[k++]=v1;
			}
			else if (j==1)
			{
				v2=(WORD)(*((DWORD*)(addr)));
				indices[k++]=v2;
			}
			else if (j==2)
			{
				v3=(WORD)(*((DWORD*)(addr)));
				indices[k++]=v3;
				vn=v3;
			}
			else
			{
				v3=(WORD)(*((DWORD*)(addr)));
				indices[k++]=v1;
				indices[k++]=vn;
				indices[k++]=v3;
				vn=v3;
			}
			
			addr+=4;
		}
	}
	
	mesh->setVertices(vertices,vertexDataLength,indices,indexLength);
}

void XFileLoader::parseMeshNormals(XFileObjectMesh* mesh,
    D3DVECTOR* normalData,int normalDataLength,
    VOID* faces,int faceLength)
{
	WORD* indices;
	int indexLength,k;
	byte* addr;
	DWORD normalsPerFace;
	WORD v1,v2,v3,vn;
	int i;

	indexLength=0;
	addr=(byte*)(faces);
	for (i=0;i<=faceLength-1;i++)
	{
		normalsPerFace=*((DWORD*)(addr));
		
		addr+=4+4*normalsPerFace;
		indexLength+=3*(normalsPerFace-2);
	}
	
	indices=new WORD[indexLength];
	
	addr=(byte*)(faces);
	k=0;
	
	for (i=0;i<=faceLength-1;i++)
	{
		normalsPerFace=*((DWORD*)(addr));
		addr+=4;
		
		for (int j=0;j<=normalsPerFace-1;j++)
		{
			if (j==0)
			{
				v1=(WORD)(*((DWORD*)(addr)));
				indices[k++]=v1;
			}
			else if (j==1)
			{
				v2=(WORD)(*((DWORD*)(addr)));
				indices[k++]=v2;
			}
			else if (j==2)
			{
				v3=(WORD)(*((DWORD*)(addr)));
				indices[k++]=v3;
				vn=v3;
			}
			else
			{
				v3=(WORD)(*((DWORD*)(addr)));
				indices[k++]=v1;
				indices[k++]=vn;
				indices[k++]=v3;
				vn=v3;
			}
			
			addr+=4;
		}
	}
	
	mesh->setNormals(normalData,normalDataLength,indices,indexLength);
}

void XFileLoader::parseMeshMaterialList(IDirectXFileData* fileData,
                                        XFileObjectMesh* mesh)
{
	IDirectXFileObject* child;
	IDirectXFileData* childData;
	IDirectXFileDataReference* childDataRef;
	const GUID* guid;

	while (fileData->GetNextObject(&child)==DXFILE_OK)
	{
		if (child->QueryInterface(IID_IDirectXFileData,
		                          (void**)(&childData))==S_OK)
		{
			childData->GetType(&guid);
			
			if (*guid==TID_D3DRMMaterial)
			{
				parseMeshMaterial(childData,mesh);
			}
			
			childData->Release();
		}
		
		if (child->QueryInterface(IID_IDirectXFileDataReference,
		                          (void**)(&childDataRef))==S_OK)
		{
			if (childDataRef->Resolve(&childData)==DXFILE_OK)
			{
				childData->GetType(&guid);
				
				if (*guid==TID_D3DRMMaterial)
				{
					parseMeshMaterial(childData,mesh);
				}
				
				childData->Release();
			}
			
			childDataRef->Release();
		}
		
		child->Release();
	}
}

void XFileLoader::parseMeshMaterial(IDirectXFileData* fileData,
                                    XFileObjectMesh* mesh)
{
	D3DMATERIAL7 material;
	IDirectXFileObject* child;
	IDirectXFileData* childData;
	IDirectDrawSurface7* texture;
	String* filename;
	byte* data;
	DWORD dataSize;
	CHAR** string;
	const GUID* guid;

	fileData->GetData(NULL,&dataSize,(void**)(&data));
	
	material.specular.a=1.0f;
	material.emissive.a=1.0f;
	memcpy(&material.diffuse ,(void*)(data+ 0),sizeof(FLOAT)*4);
	memcpy(&material.ambient ,(void*)(data+ 0),sizeof(FLOAT)*4);
	memcpy(&material.power   ,(void*)(data+16),sizeof(FLOAT)*1);
	memcpy(&material.specular,(void*)(data+20),sizeof(FLOAT)*3);
	memcpy(&material.emissive,(void*)(data+32),sizeof(FLOAT)*3);
	mesh->setMaterial(material);
	
	if (fileData->GetNextObject(&child)==DXFILE_OK)
	{
		if (child->QueryInterface(IID_IDirectXFileData,
		                          (void**)(&childData))==S_OK)
		{
			childData->GetType(&guid);
			
			if (*guid==TID_D3DRMTextureFilename)
			{
				childData->GetData(NULL,&dataSize,(void**)(&string));
				
				filename=String::format("%s%s",path->toCharArray(),*string);
				if (textureCache->containsKey(filename))
				{
					texture=textureCache->get(filename);
					texture->AddRef();
				}
				else
				{
					texture=textureLoader->load(filename);
					textureCache->put(new String(filename),texture);
				}
				mesh->setTexture(texture);
				delete filename;
			}
			
			childData->Release();
		}
		
		child->Release();
	}
}


/*   ---   fXgN^   ---   */

XFileLoader::~XFileLoader(void)
{
	for (HashtableEnumeration<String>*
	     enume=textureCache->getKeys();enume->hasMoreElements();)
	{
		delete enume->nextElement();
	}
	textureCache->clear();
	delete textureCache;
	delete textureLoader;
	
	xFile->Release();
}
