/*   X File ǂݍ  Ver 0.60   */
/*                               */
/*                               */
/*   by Atsushi     1999/10/20   */


#include "basic.h"
#include "dxfile.h"

#include "lib/math/Vector3.h"
#include "lib/math/Matrix4.h"
#include "lib/math/TransformMatrix.h"
#include "lib/math/Bounds.h"
#include "lib/math/BoundingSphere.h"
#include "lib/math/BoundingBox.h"
#include "ImageLoader.h"
#include "TextureLoader.h"
#include "XFileLoader.h"
#include "XFileObjectNode.h"
#include "XFileObjectMesh.h"
#include "XFileObjectFrame.h"
#include "XFileObject.h"


/*   ---   RXgN^   ---   */

XFileObject::XFileObject(XFileObjectNode* shapeTree)
{
	this->shapeTree=shapeTree;
}


/*   ---   `   ---   */

void XFileObject::paint(IDirect3DDevice7* device)
{
	shapeTree->paint(device);
}


/*   ---   m[h   ---   */

XFileObjectNode* XFileObject::getTree(void)
{
	return shapeTree;
}

XFileObjectNode* XFileObject::getNode(String* name)
{
	return shapeTree->getNode(name);
}

void XFileObject::detachNode(String* name)
{
	shapeTree->detachNode(name);
}


/*   ---      ---   */

BoundingSphere* XFileObject::getBoundingSphere(void)
{
	D3DVECTOR center;
	D3DVALUE radius;
	TransformMatrix* matrix;
	int pointLength;

	center.x=0.0f;
	center.y=0.0f;
	center.z=0.0f;
	radius=-1.0f;
	matrix=new TransformMatrix();
	
	shapeTree->computeBoundingSphereCenter(matrix,&center,&pointLength);
	center.x=center.x/pointLength;
	center.y=center.y/pointLength;
	center.z=center.z/pointLength;
	
	matrix->setIdentity();
	
	shapeTree->computeBoundingSphereRadius(matrix,&center,&radius);
	
	delete matrix;
	
	return new BoundingSphere(new Vector3(center),radius);
}

BoundingBox* XFileObject::getBoundingBox(void)
{
	TransformMatrix* transform;
	Vector3** vertices;
	float minX,minY,minZ,maxX,maxY,maxZ;

	transform=new TransformMatrix();
	vertices=shapeTree->getVertices(transform);
	delete transform;
	
	if (vertices[0]==NULL)
	{
		return new BoundingBox(0.0f,0.0f,0.0f,0.0f,0.0f,0.0f);
	}
	
	minX=vertices[0]->getX();
	minY=vertices[0]->getY();
	minZ=vertices[0]->getZ();
	maxX=vertices[0]->getX();
	maxY=vertices[0]->getY();
	maxZ=vertices[0]->getZ();
	for (int i=1;vertices[i]!=NULL;i++)
	{
		minX=min(minX,vertices[i]->getX());
		minY=min(minY,vertices[i]->getY());
		minZ=min(minZ,vertices[i]->getZ());
		maxX=max(maxX,vertices[i]->getX());
		maxY=max(maxY,vertices[i]->getY());
		maxZ=max(maxZ,vertices[i]->getZ());
		
		delete vertices[i];
	}
	delete[] vertices;
	
	return new BoundingBox(minX,minY,minZ,maxX,maxY,maxZ);
}

Vector3** XFileObject::getVertices(void)
{
	TransformMatrix* transform;
	Vector3** vertices;

	transform=new TransformMatrix();
	vertices=shapeTree->getVertices(transform);
	delete transform;
	
	return vertices;
}

TransformMatrix* XFileObject::getTransformMatrix(String* name)
{
	return shapeTree->getTransformMatrix(name);
}


/*   ---   ύX   ---   */

void XFileObject::setAlpha(int value)
{
// TODO
}


/*   ---   fXgN^   ---   */

XFileObject::~XFileObject(void)
{
	delete shapeTree;
}
