/*   X File ǂݍ  Ver 0.60   */
/*                               */
/*                               */
/*   by Atsushi     1999/10/20   */


#include "basic.h"
#include "dxfile.h"

#include "lib/math/Vector3.h"
#include "lib/math/Matrix4.h"
#include "lib/math/TransformMatrix.h"
#include "lib/math/Bounds.h"
#include "lib/math/BoundingSphere.h"
#include "lib/math/BoundingBox.h"
#include "lib/util/Vector.h"
#include "ImageLoader.h"
#include "TextureLoader.h"
#include "XFileLoader.h"
#include "XFileObject.h"
#include "XFileObjectNode.h"
#include "XFileObjectMesh.h"
#include "XFileObjectFrame.h"
#include "CompiledXFileObject.h"
#include "XFileObjectCompiler.h"


/*   ---   RXgN^   ---   */

XFileObjectCompiler::XFileObjectCompiler(void)
{
	meshes=new Vector<XFileObjectMesh>();
}


/*   ---   RpC   ---   */

XFileObject* XFileObjectCompiler::compile(XFileObject* object)
{
	XFileObjectMesh** shapeMeshes;

	compileObject(object);
	
	shapeMeshes=new XFileObjectMesh*[meshes->getSize()];
	for (int i=0;i<=meshes->getSize()-1;i++)
	{
		shapeMeshes[i]=meshes->getElement(i);
	}
	
	return new CompiledXFileObject(object->getTree(),
	                               shapeMeshes,meshes->getSize());
}

void XFileObjectCompiler::recompile(CompiledXFileObject* object)
{
	XFileObjectMesh** shapeMeshes;

	compileObject(object);
	
	shapeMeshes=new XFileObjectMesh*[meshes->getSize()];
	for (int i=0;i<=meshes->getSize()-1;i++)
	{
		shapeMeshes[i]=meshes->getElement(i);
	}
	
	object->setCompiledData(shapeMeshes,meshes->getSize());
}

void XFileObjectCompiler::compileObject(XFileObject* object)
{
	TransformMatrix* identity;
	XFileObjectMesh* mesh1;
	XFileObjectMesh* mesh2;

	meshes->clear();
	
	identity=new TransformMatrix();
	convertNode(object->getTree(),identity);
	delete identity;
	
	retry:
	for (int i=0;i<=meshes->getSize()-1;i++)
	{
		for (int j=i+1;j<=meshes->getSize()-1;j++)
		{
			mesh1=meshes->getElement(i);
			mesh2=meshes->getElement(j);
			if (canCombine(mesh1,mesh2))
			{
				meshes->addElement(combineMeshes(mesh1,mesh2));
				meshes->removeElement(mesh1);
				meshes->removeElement(mesh2);
				delete mesh1;
				delete mesh2;
				
				goto retry;
			}
		}
	}
}


/*   ---   m[h팸   ---   */

void XFileObjectCompiler::convertNode(XFileObjectNode* node,
                                      TransformMatrix* transform)
{
	if (node->instanceof(XFileObjectMesh))
	{
		convertNode((XFileObjectMesh*)(node),transform);
	}
	else if (node->instanceof(XFileObjectFrame))
	{
		convertNode((XFileObjectFrame*)(node),transform);
	}
}

void XFileObjectCompiler::convertNode(XFileObjectMesh* node,
                                      TransformMatrix* transform)
{
	XFileObjectMesh* newMesh;
	D3DVERTEX* newVertices;
	WORD* newIndices;
	TransformMatrix* rotation;
	Vector3 *v1,*v2;
	int i;

	rotation=new TransformMatrix();
	rotation->set(transform);
	rotation->setTranslation(0.0f,0.0f,0.0f);
	
	newVertices=new D3DVERTEX[node->getVertexLength()];
	newIndices=new WORD[node->getIndexLength()];
	for (i=0;i<=node->getVertexLength()-1;i++)
	{
		newVertices[i]=(node->getVertices())[i];
		v1=new Vector3(newVertices[i].x,newVertices[i].y,newVertices[i].z);
		v2=transform->transform(v1);
		newVertices[i].x=v2->getX();
		newVertices[i].y=v2->getY();
		newVertices[i].z=v2->getZ();
		delete v2;
		delete v1;
		v1=new Vector3(newVertices[i].nx,newVertices[i].ny,newVertices[i].nz);
		v2=rotation->transform(v1);
		newVertices[i].nx=v2->getX();
		newVertices[i].ny=v2->getY();
		newVertices[i].nz=v2->getZ();
		delete v2;
		delete v1;
	}
	for (i=0;i<=node->getIndexLength()-1;i++)
	{
		newIndices[i]=(node->getIndices())[i];
	}
	
	newMesh=new XFileObjectMesh(new String(node->getName()));
	newMesh->setVertices(newVertices,node->getVertexLength(),
	                     newIndices,node->getIndexLength());
	newMesh->setTexture(node->getTexture());
	newMesh->setMaterial(node->getMaterial());
	if (node->getTexture()!=NULL) node->getTexture()->AddRef();
	
	meshes->addElement(newMesh);
	
	delete rotation;
}

void XFileObjectCompiler::convertNode(XFileObjectFrame* node,
                                      TransformMatrix* transform)
{
	TransformMatrix* transform2;
	TransformMatrix* transform0;

	transform0=node->getTransformMatrix();
	transform2=new TransformMatrix();
	transform2->setMultiplication(transform0,transform);
	for (int i=0;node->getChild(i)!=NULL;i++)
	{
		convertNode(node->getChild(i),transform2);
	}
	delete transform2;
	delete transform0;
}


/*   ---   bV   ---   */

boolean XFileObjectCompiler::canCombine(XFileObjectMesh* mesh1,
                                        XFileObjectMesh* mesh2)
{
	D3DMATERIAL7 material1,material2;

	if (mesh1->getTexture()!=mesh2->getTexture()) return FALSE;
	
	material1=mesh1->getMaterial();
	material2=mesh2->getMaterial();
	if (material1.diffuse .r!=material2.diffuse .r) return FALSE;
	if (material1.diffuse .g!=material2.diffuse .g) return FALSE;
	if (material1.diffuse .b!=material2.diffuse .b) return FALSE;
	if (material1.ambient .r!=material2.ambient .r) return FALSE;
	if (material1.ambient .g!=material2.ambient .g) return FALSE;
	if (material1.ambient .b!=material2.ambient .b) return FALSE;
	if (material1.power     !=material2.power     ) return FALSE;
	if (material1.specular.r!=material2.specular.r) return FALSE;
	if (material1.specular.g!=material2.specular.g) return FALSE;
	if (material1.specular.b!=material2.specular.b) return FALSE;
	if (material1.emissive.r!=material2.emissive.r) return FALSE;
	if (material1.emissive.g!=material2.emissive.g) return FALSE;
	if (material1.emissive.b!=material2.emissive.b) return FALSE;
	
	return TRUE;
}

XFileObjectMesh* XFileObjectCompiler::combineMeshes(XFileObjectMesh* mesh1,
                                                    XFileObjectMesh* mesh2)
{
	XFileObjectMesh* newMesh;
	D3DVERTEX* newVertices;
	WORD* newIndices;
	int newVertexLength,newIndexLength;
	int i;

	newVertexLength=mesh1->getVertexLength()+mesh2->getVertexLength();
	newIndexLength =mesh1->getIndexLength() +mesh2->getIndexLength();
	newVertices=new D3DVERTEX[newVertexLength];
	newIndices=new WORD[newIndexLength];
	for (i=0;i<=mesh1->getVertexLength()-1;i++)
	{
		newVertices[i]=(mesh1->getVertices())[i];
	}
	for (i=0;i<=mesh2->getVertexLength()-1;i++)
	{
		newVertices[i+mesh1->getVertexLength()]=(mesh2->getVertices())[i];
	}
	for (i=0;i<=mesh1->getIndexLength()-1;i++)
	{
		newIndices[i]=(mesh1->getIndices())[i];
	}
	for (i=0;i<=mesh2->getIndexLength()-1;i++)
	{
		newIndices[i+mesh1->getIndexLength()]=(mesh2->getIndices())[i]+
		    mesh1->getVertexLength();
	}
	
	newMesh=new XFileObjectMesh(
	    String::format("%s+%s",mesh1->getName()->toCharArray(),
	                           mesh2->getName()->toCharArray()));
	newMesh->setVertices(newVertices,newVertexLength,
	                     newIndices,newIndexLength);
	newMesh->setTexture(mesh1->getTexture());
	newMesh->setMaterial(mesh1->getMaterial());
	if (mesh1->getTexture()!=NULL) mesh1->getTexture()->AddRef();
	
	return newMesh;
}


/*   ---   fXgN^   ---   */

XFileObjectCompiler::~XFileObjectCompiler(void)
{
	delete meshes;
}
