/*   X File ǂݍ  Ver 0.60   */
/*                               */
/*                               */
/*   by Atsushi     1999/10/20   */


#include "basic.h"

#include "lib/math/Vector3.h"
#include "lib/math/Matrix4.h"
#include "lib/math/TransformMatrix.h"
#include "XFileObjectNode.h"
#include "XFileObjectFrame.h"


/*   ---   RXgN^   ---   */

XFileObjectFrame::XFileObjectFrame(String* name) :XFileObjectNode(name)
{
	matrix._11=1.0f;matrix._12=0.0f;matrix._13=0.0f;matrix._14=0.0f;
	matrix._21=0.0f;matrix._22=1.0f;matrix._23=0.0f;matrix._24=0.0f;
	matrix._31=0.0f;matrix._32=0.0f;matrix._33=1.0f;matrix._34=0.0f;
	matrix._41=0.0f;matrix._42=0.0f;matrix._43=0.0f;matrix._44=1.0f;
}


/*   ---   `   ---   */

void XFileObjectFrame::paint(IDirect3DDevice7* device)
{
	D3DMATRIX oldMatrix;

	if (child!=NULL)
	{
		device->GetTransform(D3DTRANSFORMSTATE_WORLD,&oldMatrix);
		device->MultiplyTransform(D3DTRANSFORMSTATE_WORLD,&matrix);
		
		child->paint(device);
		
		device->SetTransform(D3DTRANSFORMSTATE_WORLD,&oldMatrix);
	}
	
	if (next!=NULL) next->paint(device);
}


/*   ---   s   ---   */

void XFileObjectFrame::setMatrix(D3DMATRIX matrix)
{
	this->matrix=matrix;
}

TransformMatrix* XFileObjectFrame::getTransformMatrix(void)
{
	return new TransformMatrix(matrix);
}

TransformMatrix* XFileObjectFrame::getTransformMatrix(String* name)
{
	TransformMatrix* transform0;
	TransformMatrix* transform1;
	TransformMatrix* transform;

	if (this->name->equals(name))
	{
		return new TransformMatrix(matrix);
	}
	if (child!=NULL)
	{
		transform0=child->getTransformMatrix(name);
		if (transform0!=NULL)
		{
			transform1=new TransformMatrix(matrix);
			transform=(*transform0)*(*transform1);
			delete transform1;
			delete transform0;
			
			return transform;
		}
	}
	if (next!=NULL)
	{
		return next->getTransformMatrix(name);
	}
	
	return NULL;
}


/*   ---   Ë   ---   */

void XFileObjectFrame::computeBoundingSphereCenter(
    TransformMatrix* matrix,D3DVECTOR* center,int* pointLength)
{
	TransformMatrix* matrix0;
	TransformMatrix* matrix2;

	matrix0=new TransformMatrix(this->matrix);
	matrix2=matrix0->multiply(matrix);
	
	if (child!=NULL)
	{
		child->computeBoundingSphereCenter(matrix2,center,pointLength);
	}
	if (next!=NULL)
	{
		next->computeBoundingSphereCenter(matrix,center,pointLength);
	}
	
	delete matrix2;
	delete matrix0;
}

void XFileObjectFrame::computeBoundingSphereRadius(
    TransformMatrix* matrix,D3DVECTOR* center,D3DVALUE* radius)
{
	TransformMatrix* matrix0;
	TransformMatrix* matrix2;

	matrix0=new TransformMatrix(this->matrix);
	matrix2=(*matrix0)*(*matrix);
	
	if (child!=NULL)
	{
		child->computeBoundingSphereRadius(matrix2,center,radius);
	}
	if (next!=NULL)
	{
		next->computeBoundingSphereRadius(matrix,center,radius);
	}
	
	delete matrix2;
	delete matrix0;
}


/*   ---      ---   */

XFileObjectNode* XFileObjectFrame::getChild(int index)
{
	XFileObjectNode* child;

	child=this->child;
	for (int i=0;i<=index-1;i++) child=child->getNext();
	
	return child;
}

Vector3** XFileObjectFrame::getVertices(TransformMatrix* transform)
{
	TransformMatrix* matrix;
	TransformMatrix* transform2;
	Vector3** vertices1;
	Vector3** vertices2;
	Vector3** vertices;
	int length1,length2;

	if (child!=NULL)
	{
		matrix=new TransformMatrix(this->matrix);
		transform2=(*matrix)*(*transform);
		vertices1=child->getVertices(transform2);
		delete matrix;
		delete transform2;
	}
	if (next!=NULL)
	{
		vertices2=next->getVertices(transform);
	}
	
	if (child!=NULL && next==NULL) return vertices1;
	if (child==NULL && next!=NULL) return vertices2;
	if (child==NULL && next==NULL)
	{
		vertices=new Vector3*[1];
		vertices[0]=NULL;
		
		return vertices;
	}
	
	for (length1=0;vertices1[length1]!=NULL;length1++);
	for (length2=0;vertices2[length2]!=NULL;length2++);
	vertices=new Vector3*[length1+length2+1];
	vertices[length1+length2]=NULL;
	for (int i=0;i<=length1-1;i++) vertices[        i]=vertices1[i];
	for (int j=0;j<=length2-1;j++) vertices[length1+j]=vertices2[j];
	
	delete[] vertices1;
	delete[] vertices2;
	
	return vertices;
}


/*   ---   fXgN^   ---   */

XFileObjectFrame::~XFileObjectFrame(void)
{
}
