/*   X File ǂݍ  Ver 0.60   */
/*                               */
/*                               */
/*   by Atsushi     1999/10/20   */


#include "basic.h"

#include "lib/math/Vector3.h"
#include "lib/math/Matrix4.h"
#include "lib/math/TransformMatrix.h"
#include "XFileObjectNode.h"
#include "XFileObjectMesh.h"


/*   ---   RXgN^   ---   */

XFileObjectMesh::XFileObjectMesh(String* name) :XFileObjectNode(name)
{
	texture=NULL;
	
	material.diffuse.r=1.0f;
	material.diffuse.g=1.0f;
	material.diffuse.b=1.0f;
	material.diffuse.a=1.0f;
	material.ambient.r=1.0f;
	material.ambient.g=1.0f;
	material.ambient.b=1.0f;
	material.ambient.a=1.0f;
	material.specular.r=0.0f;
	material.specular.g=0.0f;
	material.specular.b=0.0f;
	material.specular.a=0.0f;
	material.emissive.r=0.0f;
	material.emissive.g=0.0f;
	material.emissive.b=0.0f;
	material.emissive.a=0.0f;
	material.power=0.0f;
}


/*   ---   ݒ   ---   */

void XFileObjectMesh::setVertices(D3DVERTEX* vertices,int vertexLength,
                                  WORD* indices,int indexLength)
{
	this->vertices    =vertices;
	this->vertexLength=vertexLength;
	this->indices     =indices;
	this->indexLength =indexLength;
}

void XFileObjectMesh::setTexture(IDirectDrawSurface7* texture)
{
	this->texture=texture;
}

void XFileObjectMesh::setTextureCoords(FLOAT* texCoords)
{
	for (int i=0;i<=vertexLength-1;i++)
	{
		vertices[i].tu=texCoords[2*i  ];
		vertices[i].tv=texCoords[2*i+1];
	}
}

void XFileObjectMesh::setMaterial(D3DMATERIAL7 material)
{
	this->material=material;
}

void XFileObjectMesh::setNormals(D3DVECTOR* normals,int normalLength,
                                 WORD* indices,int indexLength)
{
	Vector3** normals0;
	Vector3* normal;
	int i;

	normals0=new Vector3*[vertexLength];
	for (i=0;i<=vertexLength-1;i++)
	{
		normals0[i]=new Vector3();
	}
	
	for (i=0;i<=indexLength-1;i++)
	{
		normal=new Vector3(normals[indices[i]]);
		normals0[this->indices[i]]->setAddition(normals0[this->indices[i]],
		                                        normal);
		delete normal;
	}
	
	for (i=0;i<=vertexLength-1;i++)
	{
		normals0[i]->setNormalize();
		if (normals0[i]->getLength()==0.0) normals0[i]->set(0.0,0.0,-1.0);
		
		vertices[i].nx=normals0[i]->getX();
		vertices[i].ny=normals0[i]->getY();
		vertices[i].nz=normals0[i]->getZ();
	}
	
	for (i=0;i<=vertexLength-1;i++)
	{
		delete normals0[i];
	}
	delete[] normals0;
}

void XFileObjectMesh::computeNormals(void)
{
	Vector3* *normals;
	Vector3 *v1,*v2,*v3,*v4,*v5,*vn;
	D3DVECTOR* v;
	WORD ix,iy,iz;
	int i;

	normals=new Vector3*[vertexLength];
	for (i=0;i<=vertexLength-1;i++)
	{
		normals[i]=new Vector3();
	}
	
	for (i=0;i<=indexLength-1;i+=3)
	{
		ix=indices[i  ];
		iy=indices[i+1];
		iz=indices[i+2];
		
		v=(D3DVECTOR*)(&vertices[ix]);v1=new Vector3(v->x,v->y,v->z);
		v=(D3DVECTOR*)(&vertices[iy]);v2=new Vector3(v->x,v->y,v->z);
		v=(D3DVECTOR*)(&vertices[iz]);v3=new Vector3(v->x,v->y,v->z);
		
		v4=(*v2)-(*v1);
		v5=(*v3)-(*v2);
		vn=v4->cross(v5);
		vn->setNormalize();
		
		normals[ix]->setAddition(normals[ix],vn);
		normals[iy]->setAddition(normals[iy],vn);
		normals[iz]->setAddition(normals[iz],vn);
		
		delete v1;
		delete v2;
		delete v3;
		delete v4;
		delete v5;
		delete vn;
	}
	
	for (i=0;i<=vertexLength-1;i++)
	{
		normals[i]->setNormalize();
		if (normals[i]->getLength()==0.0) normals[i]->set(0.0,0.0,-1.0);
		
		vertices[i].nx=normals[i]->getX();
		vertices[i].ny=normals[i]->getY();
		vertices[i].nz=normals[i]->getZ();
	}
	
	for (i=0;i<=vertexLength-1;i++)
	{
		delete normals[i];
	}
	delete[] normals;
}


/*   ---   `   ---   */

void XFileObjectMesh::paint(IDirect3DDevice7* device)
{
	device->SetTexture(0,texture);
	device->SetMaterial(&material);
	device->DrawIndexedPrimitive(D3DPT_TRIANGLELIST,D3DFVF_VERTEX,
	                             vertices,vertexLength,
	                             indices,indexLength,NULL);
	
	if (next!=NULL) next->paint(device);
}


/*   ---   Ë   ---   */

void XFileObjectMesh::computeBoundingSphereCenter(
    TransformMatrix* matrix,D3DVECTOR* center,int* pointLength)
{
	Vector3* vertex;
	Vector3* v;

	for (int i=0;i<=vertexLength-1;i++)
	{
		vertex=new Vector3(vertices[i].x,vertices[i].y,vertices[i].z);
		v=matrix->transform(vertex);
		
		center->x+=v->getX();
		center->y+=v->getY();
		center->z+=v->getZ();
		
		delete v;
		delete vertex;
	}
	
	(*pointLength)+=vertexLength;
	
	XFileObjectNode::computeBoundingSphereCenter(matrix,center,pointLength);
}

void XFileObjectMesh::computeBoundingSphereRadius(
    TransformMatrix* matrix,D3DVECTOR* center,D3DVALUE* radius)
{
	Vector3* vertex;
	Vector3 *v1,*v2,*vc;
	D3DVALUE dist;

	vc=new Vector3(center->x,center->y,center->z);
	
	for (int i=0;i<=vertexLength-1;i++)
	{
		vertex=new Vector3(vertices[i].x,vertices[i].y,vertices[i].z);
		
		v1=matrix->transform(vertex);
		v2=(*v1)-(*vc);
		
		dist=v2->getLength();
		
		if (*radius<dist) *radius=dist;
		
		delete v1;
		delete v2;
	}
	
	delete vc;
	
	XFileObjectNode::computeBoundingSphereRadius(matrix,center,radius);
}


/*   ---      ---   */

D3DVERTEX* XFileObjectMesh::getVertices(void)
{
	return vertices;
}

WORD* XFileObjectMesh::getIndices(void)
{
	return indices;
}

int XFileObjectMesh::getVertexLength(void)
{
	return vertexLength;
}

int XFileObjectMesh::getIndexLength(void)
{
	return indexLength;
}

IDirectDrawSurface7* XFileObjectMesh::getTexture(void)
{
	return texture;
}

D3DMATERIAL7 XFileObjectMesh::getMaterial(void)
{
	return material;
}


/*   ---      ---   */

Vector3** XFileObjectMesh::getVertices(TransformMatrix* transform)
{
	Vector3** vertices1;
	Vector3** vertices2;
	Vector3** vertices0;
	int length1,length2;
	D3DVECTOR* v;
	WORD ix,iy,iz;
	int i;

	vertices1=new Vector3*[indexLength+1];
	vertices1[indexLength]=NULL;
	for (i=0;i<=indexLength-1;i+=3)
	{
		ix=indices[i  ];
		iy=indices[i+1];
		iz=indices[i+2];
		
		v=(D3DVECTOR*)(&vertices[ix]);
		vertices1[i  ]=new Vector3(v->x,v->y,v->z);
		v=(D3DVECTOR*)(&vertices[iy]);
		vertices1[i+1]=new Vector3(v->x,v->y,v->z);
		v=(D3DVECTOR*)(&vertices[iz]);
		vertices1[i+2]=new Vector3(v->x,v->y,v->z);
	}
	
	if (next==NULL)
	{
		return vertices1;
	}
	
	vertices2=next->getVertices(transform);
	
	for (length1=0;vertices1[length1]!=NULL;length1++);
	for (length2=0;vertices2[length2]!=NULL;length2++);
	vertices0=new Vector3*[length1+length2+1];
	vertices0[length1+length2]=NULL;
	for (i=0;i<=length1-1;i++) vertices0[        i]=vertices1[i];
	for (i=0;i<=length2-1;i++) vertices0[length1+i]=vertices2[i];
	
	delete[] vertices1;
	delete[] vertices2;
	
	return vertices0;
}

TransformMatrix* XFileObjectMesh::getTransformMatrix(String* name)
{
	if (this->name->equals(name))
	{
		return new TransformMatrix();
	}
	else
	{
		return NULL;
	}
}


/*   ---   fXgN^   ---   */

XFileObjectMesh::~XFileObjectMesh(void)
{
	if (texture!=NULL) texture->Release();
	
	delete[] vertices;
	delete[] indices;
}
