/*   B-Spline Ȗ  Ver 0.00   */
/*                             */
/*                             */
/*   by Atsushi     1999/5/3   */


#include "basic.h"

#include "lib/util/PooledObject.h"
#include "lib/util/ObjectPool.h"
#include "Vector3.h"
#include "BSplineSurface.h"


/*   ---   RXgN^   ---   */

BSplineSurface::BSplineSurface(int controlN,int controlM,int degree)
{
	this->controlN=controlN;
	this->controlM=controlM;
	this->degree  =degree;
	
	control=new Vector3*[controlM*controlN];
	for (int i=0;i<=controlM*controlN-1;i++)
	{
		control[i]=new Vector3();
	}
}


/*   ---   _   ---   */

void BSplineSurface::setControlPoint(int x,int y,Vector3* point)
{
	control[x+y*controlN]->set(point);
}

Vector3* BSplineSurface::getControlPoint(int x,int y)
{
	return control[x+y*controlN];
}


/*   ---      ---   */

Vector3* BSplineSurface::getPosition(float u,float v)
{
	float x,y,z;
	float blendi,blendj;

	x=0.0f;
	y=0.0f;
	z=0.0f;
	for (int i=max((int)(u),0);i<=min((int)(u)+degree,controlN-1);i++)
	{
		blendi=getBlendValue(i,degree+1,u);
		for (int j=max((int)(v),0);j<=min((int)(v)+degree,controlM-1);j++)
		{
			blendj=getBlendValue(j,degree+1,v);
			
			x+=control[i+j*controlN]->getX()*blendi*blendj;
			y+=control[i+j*controlN]->getY()*blendi*blendj;
			z+=control[i+j*controlN]->getZ()*blendi*blendj;
		}
	}
	
	return new Vector3(x,y,z);
}

Vector3* BSplineSurface::getNormal(float u,float v)
{
	Vector3* normal;
	Vector3 *v1,*v2,*v3;
	Vector3 *vx,*vy;

	v1=getPosition(u-0.01,v     );
	v2=getPosition(u     ,v-0.02);
	v3=getPosition(u+0.01,v     );
	
	vx=(*v1)-(*v2);
	vy=(*v3)-(*v2);
	
	normal=vx->cross(vy);
	normal->setNormalize();
	
	delete vx;
	delete vy;
	delete v3;
	delete v2;
	delete v1;
	
	return normal;
}


/*   ---   ֐   ---   */

float BSplineSurface::getBlendValue(int i,int order,float u)
{
	if (order==1)
	{
		if (i-3.0f<=u && u<(i+1)-3.0f)
		{
			return 1.0;
		}
		else
		{
			return 0.0;
		}
	}
	else
	{
		return (u-(i-3.0f))*getBlendValue(i,order-1,u)/
		       (((i+order-1)-3.0f)-(i-3.0f))+
		       (((i+order)-3.0f)-u)*getBlendValue(i+1,order-1,u)/
		       (((i+order)-3.0f)-((i+1)-3.0f));
	}
}


/*   ---   fXgN^   ---   */

BSplineSurface::~BSplineSurface(void)
{
	for (int i=0;i<=controlM*controlN-1;i++)
	{
		delete control[i];
	}
	delete[] control;
}
