/*   xcBGȖ  Ver 0.00   */
/*                            */
/*                            */
/*   by Atsushi   1999/4/18   */


#include "basic.h"

#include "lib/util/PooledObject.h"
#include "lib/util/ObjectPool.h"
#include "Vector3.h"
#include "BezierSurface.h"


/*   ---   RXgN^   ---   */

BezierSurface::BezierSurface(int n,int m)
{
	this->n=n;
	this->m=m;
	
	control=new Vector3**[m+1];
	for (int i=0;i<=m+1-1;i++)
	{
		control[i]=new Vector3*[n+1];
	}
	
	computeCombination();
}


/*   ---   gݍ킹   ---   */

void BezierSurface::computeCombination(void)
{
	int i;

	combinationN=new int[n+1];
	for (i=0;i<=n;i++)
	{
		combinationN[i]=combination(n,i);
	}
	combinationM=new int[m+1];
	for (i=0;i<=m;i++)
	{
		combinationM[i]=combination(m,i);
	}
}


/*   ---   _   ---   */

void BezierSurface::setControlPoint(int x,int y,Vector3* control)
{
	this->control[x][y]=control;
}

Vector3* BezierSurface::getControlPoint(int x,int y)
{
	return control[x][y];
}


/*   ---      ---   */

Vector3* BezierSurface::getPosition(D3DVALUE u,D3DVALUE v)
{
	D3DVALUE x,y,z;
	D3DVALUE blendi,blendj;

	x=0.0f;
	y=0.0f;
	z=0.0f;
	for (int i=0;i<=n+1-1;i++)
	{
		blendi=combinationN[i]*pow(u,i)*pow(1-u,n-i);
		for (int j=0;j<=m+1-1;j++)
		{
			blendj=combinationM[j]*pow(v,j)*pow(1-v,m-j);
			
			x+=control[i][j]->getX()*blendi*blendj;
			y+=control[i][j]->getY()*blendi*blendj;
			z+=control[i][j]->getZ()*blendi*blendj;
		}
	}
	
	return new Vector3(x,y,z);
}

Vector3* BezierSurface::getNormal(D3DVALUE u,D3DVALUE v)
{
	Vector3* normal;
	Vector3 *v1,*v2,*v3;
	Vector3 *vx,*vy;

	v1=getPosition(u-0.01,v     );
	v2=getPosition(u     ,v-0.02);
	v3=getPosition(u+0.01,v     );
	
	vx=(*v1)-(*v2);
	vy=(*v3)-(*v2);
	
	normal=vx->cross(vy);
	normal->setNormalize();
	
	delete vx;
	delete vy;
	delete v3;
	delete v2;
	delete v1;
	
	return normal;
}


/*   ---   fXgN^   ---   */

BezierSurface::~BezierSurface(void)
{
	for (int i=0;i<=m+1-1;i++)
	{
		delete[] control;
	}
	delete[] control;
}
