/*   Ë ()  Ver 0.00   */
/*                             */
/*                             */
/*   by Atsushi     1999/8/3   */


#include "basic.h"

#include "lib/util/OutOfPoolException.h"
#include "lib/util/PooledObject.h"
#include "lib/util/ObjectPool.h"
#include "lib/math/Vector3.h"
#include "Bounds.h"
#include "BoundingSphere.h"
#include "BoundingBox.h"


/*   ---   RXgN^   ---   */

BoundingBox::BoundingBox(D3DVALUE minX,D3DVALUE minY,D3DVALUE minZ,
                         D3DVALUE maxX,D3DVALUE maxY,D3DVALUE maxZ)
{
	this->minX=minX;
	this->minY=minY;
	this->minZ=minZ;
	this->maxX=maxX;
	this->maxY=maxY;
	this->maxZ=maxZ;
}

BoundingBox::BoundingBox(BoundingBox* box)
{
	this->minX=box->minX;
	this->minY=box->minY;
	this->minZ=box->minZ;
	this->maxX=box->maxX;
	this->maxY=box->maxY;
	this->maxZ=box->maxZ;
}

BoundingBox::BoundingBox(void)
{
	this->minX=0.0f;
	this->minY=0.0f;
	this->minZ=0.0f;
	this->maxX=0.0f;
	this->maxY=0.0f;
	this->maxZ=0.0f;
}


/*   ---   v[   ---   */

void* BoundingBox::operator new(size_t size)
{
	return ObjectPool<BoundingBox>::newObject();
}

void BoundingBox::operator delete(void *object)
{
	ObjectPool<BoundingBox>::deleteObject((BoundingBox*)(object));
}


/*   ---   l   ---   */

void BoundingBox::set(D3DVALUE minX,D3DVALUE minY,D3DVALUE minZ,
                      D3DVALUE maxX,D3DVALUE maxY,D3DVALUE maxZ)
{
	this->minX=minX;
	this->minY=minY;
	this->minZ=minZ;
	this->maxX=maxX;
	this->maxY=maxY;
	this->maxZ=maxZ;
}

D3DVALUE BoundingBox::getMinX(void)
{
	return minX;
}

D3DVALUE BoundingBox::getMinY(void)
{
	return minY;
}

D3DVALUE BoundingBox::getMinZ(void)
{
	return minZ;
}

D3DVALUE BoundingBox::getMaxX(void)
{
	return maxX;
}

D3DVALUE BoundingBox::getMaxY(void)
{
	return maxY;
}

D3DVALUE BoundingBox::getMaxZ(void)
{
	return maxZ;
}


/*   ---      ---   */

boolean BoundingBox::intersects(BoundingBox* target)
{
	return !(maxX<target->minX || target->maxX<minX ||
	         maxY<target->minY || target->maxY<minY ||
	         maxZ<target->minZ || target->maxZ<minZ);
}

boolean BoundingBox::intersects(BoundingSphere* target)
{
	return (minX<=target->getCenter()->getX()+target->getRadius() &&
	        minY<=target->getCenter()->getY()+target->getRadius() &&
	        minZ<=target->getCenter()->getZ()+target->getRadius() &&
	        target->getCenter()->getX()-target->getRadius()<=maxX &&
	        target->getCenter()->getY()-target->getRadius()<=maxY &&
	        target->getCenter()->getZ()-target->getRadius()<=maxZ);
}

boolean BoundingBox::contains(Vector3* target)
{
	return (minX<=target->getX() && target->getX()<=maxX &&
	        minY<=target->getY() && target->getY()<=maxY &&
	        minZ<=target->getZ() && target->getZ()<=maxZ);
}


/*   ---   ړ   ---   */

void BoundingBox::translate(Vector3* delta)
{
	minX=minX+delta->getX();
	minY=minY+delta->getY();
	minZ=minZ+delta->getZ();
	maxX=maxX+delta->getX();
	maxY=maxY+delta->getY();
	maxZ=maxZ+delta->getZ();
}


/*   ---   fXgN^   ---   */

BoundingBox::~BoundingBox(void)
{
}
