/*   Ë ()  Ver 0.30   */
/*                             */
/*                             */
/*   by Atsushi    1999/8/22   */


#include "basic.h"

#include "lib/util/OutOfPoolException.h"
#include "lib/util/PooledObject.h"
#include "lib/util/ObjectPool.h"
#include "lib/math/Vector3.h"
#include "Bounds.h"
#include "BoundingSphere.h"


/*   ---   RXgN^   ---   */

BoundingSphere::BoundingSphere(Vector3* center,D3DVALUE radius)
{
	this->center=center;
	this->radius=radius;
}

BoundingSphere::BoundingSphere(BoundingSphere* bounds)
{
	this->center=new Vector3(bounds->center);
	this->radius=bounds->radius;
}

BoundingSphere::BoundingSphere(void)
{
	center=new Vector3();
	radius=1.0f;
}


/*   ---   v[   ---   */

void* BoundingSphere::operator new(size_t size)
{
	return ObjectPool<BoundingSphere>::newObject();
}

void BoundingSphere::operator delete(void *object)
{
	ObjectPool<BoundingSphere>::deleteObject((BoundingSphere*)(object));
}


/*   ---      ---   */

Vector3* BoundingSphere::getCenter(void)
{
	return center;
}

D3DVALUE BoundingSphere::getRadius(void)
{
	return radius;
}


/*   ---   ݒ   ---   */

void BoundingSphere::set(BoundingSphere* bounds)
{
	this->center->set(bounds->center);
	this->radius=bounds->radius;
}

void BoundingSphere::setCenter(Vector3* center)
{
	this->center=center;
}

void BoundingSphere::setRadius(D3DVALUE radius)
{
	this->radius=radius;
}


/*   ---      ---   */

boolean BoundingSphere::intersects(BoundingSphere* target)
{
	Vector3* delta;
	boolean intersect;

	delta=(*(target->center))-(*(center));
	
	intersect=(delta->getLength2()<
	           (target->radius+radius)*(target->radius+radius));
	
	delete delta;
	
	return intersect;
}

boolean BoundingSphere::contains(Vector3* target)
{
	Vector3* delta;
	boolean contain;

	delta=(*(target))-(*(center));
	
	contain=(delta->getLength2()<radius*radius);
	
	delete delta;
	
	return contain;
}


/*   ---   ړ   ---   */

void BoundingSphere::translate(Vector3* delta)
{
	center->setAddition(center,delta);
}


/*   ---   fXgN^   ---   */

BoundingSphere::~BoundingSphere(void)
{
	delete center;
}
