/*   4 s 4 s  Ver 0.30   */
/*                             */
/*                             */
/*   by Atsushi   1999/10/21   */


#include "basic.h"

#include "lib/util/OutOfPoolException.h"
#include "lib/util/PooledObject.h"
#include "lib/util/ObjectPool.h"
#include "Vector3.h"
#include "Quaternion.h"
#include "Matrix4.h"


/*   ---   RXgN^   ---   */

Matrix4::Matrix4(D3DMATRIX matrix)
{
	this->matrix=matrix;
}

Matrix4::Matrix4(Matrix4* matrix)
{
	this->matrix=matrix->matrix;
}

Matrix4::Matrix4(void)
{
	setIdentity();
}


/*   ---   v[   ---   */

void* Matrix4::operator new(size_t size)
{
	return ObjectPool<Matrix4>::newObject();
}

void Matrix4::operator delete(void *object)
{
	ObjectPool<Matrix4>::deleteObject((Matrix4*)(object));
}


/*   ---   s   ---   */

D3DMATRIX* Matrix4::getMatrix(void)
{
	return (D3DMATRIX*)(&matrix);
}

D3DXMATRIX* Matrix4::getDXMatrix(void)
{
	return &matrix;
}

float Matrix4::get(int y,int x)
{
	return matrix.m[y][x];
}

boolean Matrix4::isIdentity(void)
{
	return D3DXMatrixIsIdentity(&matrix);
}


/*   ---   ݒ   ---   */

void Matrix4::set(Matrix4* matrix)
{
	memcpy(&this->matrix,&matrix->matrix,sizeof(D3DXMATRIX));
}

void Matrix4::set(int y,int x,float value)
{
	matrix.m[y][x]=value;
}

void Matrix4::setZero(void)
{
	ZeroMemory(&matrix,sizeof(D3DXMATRIX));
}

void Matrix4::setIdentity(void)
{
	D3DXMatrixIdentity(&matrix);
}


/*   ---   vfݒ   ---   */

void Matrix4::setInverse(Matrix4* value)
{
	float det;

	D3DXMatrixInverse(&matrix,&det,&value->matrix);
}


/*   ---   Z   ---   */

void Matrix4::setMultiplication(Matrix4* multiplicand,
                                Matrix4* multiplier)
{
	D3DXMatrixMultiply(&matrix,&multiplicand->matrix,&multiplier->matrix);
}

Matrix4* Matrix4::multiply(Matrix4* multiplier)
{
	Matrix4* answer;

	answer=new Matrix4();
	answer->setMultiplication(this,multiplier);
	
	return answer;
}

Matrix4* Matrix4::operator*(Matrix4 multiplier)
{
	return multiply(&multiplier);
}

Matrix4* Matrix4::inverse(void)
{
	Matrix4* answer;

	answer=new Matrix4();
	answer->setInverse(this);
	
	return answer;
}


/*   ---   fXgN^   ---   */

Matrix4::~Matrix4(void)
{
}
