/*   NH[^jI  Ver 0.20   */
/*                              */
/*                              */
/*   by Atsushi    1999/10/21   */


#include "basic.h"

#include "lib/util/OutOfPoolException.h"
#include "lib/util/PooledObject.h"
#include "lib/util/ObjectPool.h"
#include "Vector3.h"
#include "Matrix4.h"
#include "Quaternion.h"


/*   ---   RXgN^   ---   */

Quaternion::Quaternion(float w,float x,float y,float z)
{
	this->quaternion.w=w;
	this->quaternion.x=x;
	this->quaternion.y=y;
	this->quaternion.z=z;
}

Quaternion::Quaternion(Quaternion* quaternion)
{
	this->quaternion=quaternion->quaternion;
}

Quaternion::Quaternion(void)
{
	setZero();
}


/*   ---   v[   ---   */

void* Quaternion::operator new(size_t size)
{
	return ObjectPool<Quaternion>::newObject();
}

void Quaternion::operator delete(void *object)
{
	ObjectPool<Quaternion>::deleteObject((Quaternion*)(object));
}


/*   ---   l   ---   */

D3DXQUATERNION* Quaternion::getDXQuaternion(void)
{
	return &quaternion;
}

float Quaternion::getX(void)
{
	return quaternion.x;
}

float Quaternion::getY(void)
{
	return quaternion.y;
}

float Quaternion::getZ(void)
{
	return quaternion.z;
}

float Quaternion::getW(void)
{
	return quaternion.w;
}

boolean Quaternion::isIdentity(void)
{
	return D3DXQuaternionIsIdentity(&quaternion);
}


/*   ---   ݒ   ---   */

void Quaternion::set(Quaternion* quaternion)
{
	this->quaternion=quaternion->quaternion;
}

void Quaternion::set(float w,float x,float y,float z)
{
	this->quaternion.w=w;
	this->quaternion.x=x;
	this->quaternion.y=y;
	this->quaternion.z=z;
}

void Quaternion::setZero(void)
{
	this->quaternion.w=0.0f;
	this->quaternion.x=0.0f;
	this->quaternion.y=0.0f;
	this->quaternion.z=0.0f;
}

void Quaternion::setIdentity(void)
{
	D3DXQuaternionIdentity(&quaternion);
}


/*   ---   vfݒ   ---   */

void Quaternion::setRotation(Vector3* axis,double angle)
{
	D3DXQuaternionRotationAxis(&quaternion,axis->getDXVector(),angle);
}

void Quaternion::setRotation(Matrix4* matrix)
{
	D3DXQuaternionRotationMatrix(&quaternion,matrix->getDXMatrix());
}


/*   ---   vfݒ   ---   */

void Quaternion::setAddition(Quaternion* value,Quaternion* addend)
{
	quaternion.w=(value->quaternion.w)+(addend->quaternion.w);
	quaternion.x=(value->quaternion.x)+(addend->quaternion.x);
	quaternion.y=(value->quaternion.y)+(addend->quaternion.y);
	quaternion.z=(value->quaternion.z)+(addend->quaternion.z);
}

void Quaternion::setSubtraction(Quaternion* value0,Quaternion* value1)
{
	quaternion.w=(value0->quaternion.w)-(value1->quaternion.w);
	quaternion.x=(value0->quaternion.x)-(value1->quaternion.x);
	quaternion.y=(value0->quaternion.y)-(value1->quaternion.y);
	quaternion.z=(value0->quaternion.z)-(value1->quaternion.z);
}

void Quaternion::setMultiplication(Quaternion* multiplicand,
                                   Quaternion* multiplier)
{
	D3DXQuaternionMultiply(&quaternion,&multiplicand->quaternion,
	                                   &multiplier->quaternion);
}

void Quaternion::setMultiplication(float multiplier)
{
	quaternion.w=quaternion.w*multiplier;
	quaternion.x=quaternion.x*multiplier;
	quaternion.y=quaternion.y*multiplier;
	quaternion.z=quaternion.z*multiplier;
}

void Quaternion::setConjugate(Quaternion* value)
{
	D3DXQuaternionConjugate(&quaternion,&value->quaternion);
}

void Quaternion::setInverse(Quaternion* value)
{
	D3DXQuaternionInverse(&quaternion,&value->quaternion);
}

void Quaternion::setSlerp(Quaternion* value0,Quaternion* value1,double alpha)
{
	D3DXQuaternionSlerp(&quaternion,&value0->quaternion,
	                    &value1->quaternion,alpha);
}

void Quaternion::setNormalize(void)
{
	D3DXQUATERNION quaternion0;

	quaternion0=quaternion;
	D3DXQuaternionNormalize(&quaternion,&quaternion0);
}


/*   ---   擾   ---   */

float Quaternion::getLength(void)
{
	return D3DXQuaternionLength(&quaternion);
}

float Quaternion::getLength2(void)
{
	return D3DXQuaternionLengthSq(&quaternion);
}


/*   ---   Z   ---   */

Quaternion* Quaternion::add(Quaternion* addend)
{
	Quaternion* answer;

	answer=new Quaternion();
	answer->setAddition(this,addend);
	
	return answer;
}

Quaternion* Quaternion::operator+(Quaternion addend)
{
	return add(&addend);
}

Quaternion* Quaternion::subtract(Quaternion* value1)
{
	Quaternion* answer;

	answer=new Quaternion();
	answer->setSubtraction(this,value1);
	
	return answer;
}

Quaternion* Quaternion::operator-(Quaternion value1)
{
	return subtract(&value1);
}

Quaternion* Quaternion::multiply(Quaternion* multiplier)
{
	Quaternion* answer;

	answer=new Quaternion();
	answer->setMultiplication(this,multiplier);
	
	return answer;
}

Quaternion* Quaternion::operator*(Quaternion multiplier)
{
	return multiply(&multiplier);
}

Quaternion* Quaternion::multiply(float multiplier)
{
	Quaternion* answer;

	answer=new Quaternion();
	answer->quaternion.w=quaternion.w;
	answer->quaternion.x=quaternion.x;
	answer->quaternion.y=quaternion.y;
	answer->quaternion.z=quaternion.z;
	answer->setMultiplication(multiplier);
	
	return answer;
}

Quaternion* Quaternion::operator*(float multiplier)
{
	return multiply(multiplier);
}

Quaternion* Quaternion::conjugate(void)
{
	Quaternion* answer;

	answer=new Quaternion();
	D3DXQuaternionConjugate(&answer->quaternion,&quaternion);
	
	return answer;
}

float Quaternion::dot(Quaternion* multiplier)
{
	return D3DXQuaternionDot(&quaternion,&multiplier->quaternion);
}

Quaternion* Quaternion::inverse(void)
{
	Quaternion* answer;

	answer=new Quaternion();
	D3DXQuaternionInverse(&answer->quaternion,&quaternion);
	
	return answer;
}

Quaternion* Quaternion::slerp(Quaternion* value1,double alpha)
{
	Quaternion* answer;

	answer=new Quaternion();
	answer->setSlerp(this,value1,alpha);
	
	return answer;
}

Quaternion* Quaternion::normalize(void)
{
	Quaternion* answer;

	answer=new Quaternion();
	D3DXQuaternionNormalize(&answer->quaternion,&quaternion);
	
	return answer;
}


/*   ---   fXgN^   ---   */

Quaternion::~Quaternion(void)
{
}
