/*   ϊs      Ver 0.40   */
/*                            */
/*                            */
/*   by Atsushi  1999/10/21   */


#include "basic.h"

#include "lib/util/OutOfPoolException.h"
#include "lib/util/PooledObject.h"
#include "lib/util/ObjectPool.h"
#include "Vector3.h"
#include "Vector4.h"
#include "Quaternion.h"
#include "Matrix4.h"
#include "TransformMatrix.h"
#include "TranslationMatrix.h"
#include "RotationMatrix.h"


/*   ---   RXgN^   ---   */

TransformMatrix::TransformMatrix(D3DMATRIX matrix) :Matrix4(matrix)
{
}

TransformMatrix::TransformMatrix(TransformMatrix* matrix) :Matrix4(matrix)
{
}

TransformMatrix::TransformMatrix(void) :Matrix4()
{
}


/*   ---   v[   ---   */

void* TransformMatrix::operator new(size_t size)
{
	return ObjectPool<TransformMatrix>::newObject();
}

void TransformMatrix::operator delete(void *object)
{
	ObjectPool<TransformMatrix>::deleteObject((TransformMatrix*)(object));
}


/*   ---   ړ   ---   */

void TransformMatrix::setTranslation(float vx,float vy,float vz)
{
	D3DXMatrixTranslation(&matrix,vx,vy,vz);
}

void TransformMatrix::setTranslation(Vector3* delta)
{
	D3DXMatrixTranslation(&matrix,delta->getX(),delta->getY(),delta->getZ());
}


/*   ---   ]   ---   */

void TransformMatrix::setRotationX(double angle)
{
	D3DXMatrixRotationX(&matrix,angle);
}

void TransformMatrix::setRotationY(double angle)
{
	D3DXMatrixRotationY(&matrix,angle);
}

void TransformMatrix::setRotationZ(double angle)
{
	D3DXMatrixRotationZ(&matrix,angle);
}

void TransformMatrix::setRotation(Vector3* i,Vector3* j,Vector3* k)
{
	matrix.m[0][0]=i->getX();matrix.m[0][1]=i->getY();matrix.m[0][2]=i->getZ();
	matrix.m[1][0]=j->getX();matrix.m[1][1]=j->getY();matrix.m[1][2]=j->getZ();
	matrix.m[2][0]=k->getX();matrix.m[2][1]=k->getY();matrix.m[2][2]=k->getZ();
}

void TransformMatrix::setRotation(Vector3* axis,double angle)
{
	D3DXMatrixRotationAxis(&matrix,axis->getDXVector(),angle);
}

void TransformMatrix::setRotation(Quaternion* rotation)
{
	D3DXMatrixRotationQuaternion(&matrix,rotation->getDXQuaternion());
}

void TransformMatrix::setRotationInverse(Vector3* i,Vector3* j,Vector3* k)
{
	matrix.m[0][0]=i->getX();matrix.m[0][1]=j->getX();matrix.m[0][2]=k->getX();
	matrix.m[1][0]=i->getY();matrix.m[1][1]=j->getY();matrix.m[1][2]=k->getY();
	matrix.m[2][0]=i->getZ();matrix.m[2][1]=j->getZ();matrix.m[2][2]=k->getZ();
}

void TransformMatrix::setRotationInverse(TransformMatrix* rotation)
{
	D3DMATRIX* matrix0;

	matrix0=rotation->getMatrix();
	
	matrix.m[0][0]=matrix0->_11;
	matrix.m[0][1]=matrix0->_21;
	matrix.m[0][2]=matrix0->_31;
	matrix.m[1][0]=matrix0->_12;
	matrix.m[1][1]=matrix0->_22;
	matrix.m[1][2]=matrix0->_32;
	matrix.m[2][0]=matrix0->_13;
	matrix.m[2][1]=matrix0->_23;matrix.m[2][2]=matrix0->_33;
}


/*   ---   XP[   ---   */

void TransformMatrix::setScale(float scale)
{
	matrix.m[3][3]=scale;
}

void TransformMatrix::setScale(float x,float y,float z)
{
	D3DXMatrixScaling(&matrix,x,y,z);
}

void TransformMatrix::setScaleMultiplication(float scale)
{
	matrix.m[3][3]*=scale;
}

void TransformMatrix::setScaleMultiplication(float x,float y,float z)
{
	matrix.m[0][0]*=x;
	matrix.m[1][1]*=y;
	matrix.m[2][2]*=z;
}


/*   ---   Be   ---   */

void TransformMatrix::setLookAt(Vector3* position,Vector3* target,Vector3* up)
{
	D3DXMatrixLookAtLH(&matrix,position->getDXVector(),
	                   target->getDXVector(),up->getDXVector());
}


/*   ---   eݒ   ---   */

void TransformMatrix::setFrustumProjection(double nearClip,double farClip,
                                           double hFOV,double vFOV)
{
	D3DXMatrixPerspectiveFovLH(&matrix,hFOV,vFOV/hFOV,nearClip,farClip);
}

void TransformMatrix::setParallelProjection(double nearClip,double farClip,
                                            double width,double height)
{
	D3DXMatrixOrthoLH(&matrix,width,height,nearClip,farClip);
}


/*   ---   K   ---   */

void TransformMatrix::setNormalizeRotation(void)
{
	float length;

	length=_hypot(matrix.m[0][0],matrix.m[0][1],matrix.m[0][2]);
	matrix.m[0][0]/=length;matrix.m[0][1]/=length;matrix.m[0][2]/=length;
	length=_hypot(matrix.m[1][0],matrix.m[1][1],matrix.m[1][2]);
	matrix.m[1][0]/=length;matrix.m[1][1]/=length;matrix.m[1][2]/=length;
	length=_hypot(matrix.m[2][0],matrix.m[2][1],matrix.m[2][2]);
	matrix.m[2][0]/=length;matrix.m[2][1]/=length;matrix.m[2][2]/=length;
}


/*   ---   ϊ   ---   */

Vector3* TransformMatrix::transform(Vector3* vector)
{
	D3DXVECTOR4 answer;

	D3DXVec3Transform(&answer,vector->getDXVector(),&matrix);
	
	return new Vector3(answer.x/answer.w,answer.y/answer.w,answer.z/answer.w);
}

Vector4* TransformMatrix::transform(Vector4* vector)
{
	Vector4* answer;

	answer=new Vector4();
	D3DXVec4Transform(answer->getDXVector(),vector->getDXVector(),&matrix);
	
	return answer;
}


/*   ---   Z   ---   */

TransformMatrix* TransformMatrix::multiply(TransformMatrix* multiplier)
{
	TransformMatrix* answer;

	answer=new TransformMatrix();
	answer->setMultiplication(this,multiplier);
	
	return answer;
}

TransformMatrix* TransformMatrix::operator*(TransformMatrix multiplier)
{
	return multiply(&multiplier);
}


/*   ---   ώZ   ---   */

void TransformMatrix::setMultiplication(TransformMatrix* multiplicand,
                                        TransformMatrix* multiplier)
{
	Matrix4::setMultiplication(multiplicand,multiplier);
}


/*   ---   fXgN^   ---   */

TransformMatrix::~TransformMatrix(void)
{
}
