/*   2 xNg  Ver 0.20   */
/*                              */
/*                              */
/*   by Atsushi    1999/12/20   */


#include "basic.h"

#include "lib/util/OutOfPoolException.h"
#include "lib/util/PooledObject.h"
#include "lib/util/ObjectPool.h"
#include "Vector2.h"


/*   ---   RXgN^   ---   */

Vector2::Vector2(float x,float y)
{
	this->vector.x=x;
	this->vector.y=y;
}

Vector2::Vector2(Vector2* vector)
{
	this->vector=vector->vector;
}

Vector2::Vector2(void)
{
	setZero();
}


/*   ---   v[   ---   */

void* Vector2::operator new(size_t size)
{
	return ObjectPool<Vector2>::newObject();
}

void Vector2::operator delete(void *object)
{
	ObjectPool<Vector2>::deleteObject((Vector2*)(object));
}


/*   ---   xNgl   ---   */

D3DXVECTOR2* Vector2::getDXVector(void)
{
	return &vector;
}

float Vector2::getX(void)
{
	return vector.x;
}

float Vector2::getY(void)
{
	return vector.y;
}

boolean Vector2::isZero(void)
{
	return (vector.x==0.0 && vector.y==0.0);
}


/*   ---   ݒ   ---   */

void Vector2::set(Vector2* vector)
{
	this->vector=vector->vector;
}

void Vector2::set(float x,float y)
{
	this->vector.x=x;
	this->vector.y=y;
}

void Vector2::setZero(void)
{
	this->vector.x=0.0f;
	this->vector.y=0.0f;
}


/*   ---   vfݒ   ---   */

void Vector2::setAddition(Vector2* value,Vector2* addend)
{
	D3DXVec2Add(&vector,&value->vector,&addend->vector);
}

void Vector2::setSubtraction(Vector2* value0,Vector2* value1)
{
	D3DXVec2Subtract(&vector,&value0->vector,&value1->vector);
}

void Vector2::setMultiplication(float multiplier)
{
	D3DXVECTOR2 vector0;

	vector0=vector;
	D3DXVec2Scale(&vector,&vector0,multiplier);
}

void Vector2::setNormalize(void)
{
	D3DXVECTOR2 vector0;

	vector0=vector;
	D3DXVec2Normalize(&vector,&vector0);
}

void Vector2::setInterpolation(Vector2* v1,Vector2* v2,double blend)
{
	D3DXVec2Lerp(&vector,&v1->vector,&v2->vector,blend);
}

void Vector2::setBaryCentric(Vector2* v1,Vector2* v2,Vector2* v3,
                             float weight2,float weight3)
{
	D3DXVec2BaryCentric(&vector,&v1->vector,&v2->vector,&v3->vector,
	                    weight2,weight3);
}


/*   ---   擾   ---   */

float Vector2::getLength(void)
{
	return D3DXVec2Length(&vector);
}

float Vector2::getLength2(void)
{
	return D3DXVec2LengthSq(&vector);
}

double Vector2::getAngle(Vector2* target)
{
	float mul,len1,len2;

	mul=this->dot(target);
	len1=this->getLength();
	len2=target->getLength();
	
	if (len1==0.0 || len2==0.0) return 0.0;
	
	return (acos(mul/(len1*len2)));
}


/*   ---   Z   ---   */

Vector2* Vector2::add(Vector2* addend)
{
	Vector2* answer;

	answer=new Vector2();
	answer->setAddition(this,addend);
	
	return answer;
}

Vector2* Vector2::operator+(Vector2 addend)
{
	return add(&addend);
}

Vector2* Vector2::subtract(Vector2* value1)
{
	Vector2* answer;

	answer=new Vector2();
	answer->setSubtraction(this,value1);
	
	return answer;
}

Vector2* Vector2::operator-(Vector2 value1)
{
	return subtract(&value1);
}

float Vector2::dot(Vector2* multiplier)
{
	return D3DXVec2Dot(&vector,&multiplier->vector);
}

Vector2* Vector2::multiply(float multiplier)
{
	Vector2* answer;

	answer=new Vector2();
	D3DXVec2Scale(&answer->vector,&vector,multiplier);
	
	return answer;
}

Vector2* Vector2::operator*(float multiplier)
{
	return multiply(multiplier);
}

Vector2* Vector2::normalize(void)
{
	Vector2* answer;

	answer=new Vector2();
	D3DXVec2Normalize(&answer->vector,&vector);
	
	return answer;
}


/*   ---   fXgN^   ---   */

Vector2::~Vector2(void)
{
}
