/*   3 xNg  Ver 0.60   */
/*                              */
/*                              */
/*   by Atsushi    1999/12/20   */


#include "basic.h"

#include "lib/util/OutOfPoolException.h"
#include "lib/util/PooledObject.h"
#include "lib/util/ObjectPool.h"
#include "Vector3.h"


/*   ---   RXgN^   ---   */

Vector3::Vector3(D3DVECTOR vector)
{
	this->vector=vector;
}

Vector3::Vector3(float x,float y,float z)
{
	this->vector.x=x;
	this->vector.y=y;
	this->vector.z=z;
}

Vector3::Vector3(Vector3* vector)
{
	this->vector=vector->vector;
}

Vector3::Vector3(void)
{
	setZero();
}


/*   ---   v[   ---   */

void* Vector3::operator new(size_t size)
{
	return ObjectPool<Vector3>::newObject();
}

void Vector3::operator delete(void *object)
{
	ObjectPool<Vector3>::deleteObject((Vector3*)(object));
}


/*   ---   xNgl   ---   */

D3DVECTOR* Vector3::getVector(void)
{
	return (D3DVECTOR*)(&vector);
}

D3DXVECTOR3* Vector3::getDXVector(void)
{
	return &vector;
}

float Vector3::getX(void)
{
	return vector.x;
}

float Vector3::getY(void)
{
	return vector.y;
}

float Vector3::getZ(void)
{
	return vector.z;
}

boolean Vector3::isZero(void)
{
	return (vector.x==0.0 && vector.y==0.0 && vector.z==0.0);
}


/*   ---   ݒ   ---   */

void Vector3::set(Vector3* vector)
{
	this->vector=vector->vector;
}

void Vector3::set(float x,float y,float z)
{
	this->vector.x=x;
	this->vector.y=y;
	this->vector.z=z;
}

void Vector3::setZero(void)
{
	this->vector.x=0.0f;
	this->vector.y=0.0f;
	this->vector.z=0.0f;
}


/*   ---   vfݒ   ---   */

void Vector3::setAddition(Vector3* value,Vector3* addend)
{
	D3DXVec3Add(&vector,&value->vector,&addend->vector);
}

void Vector3::setSubtraction(Vector3* value0,Vector3* value1)
{
	D3DXVec3Subtract(&vector,&value0->vector,&value1->vector);
}

void Vector3::setCrossProduct(Vector3* multiplicand,Vector3* multiplier)
{
	D3DXVec3Cross(&vector,&multiplicand->vector,&multiplier->vector);
}

void Vector3::setMultiplication(float multiplier)
{
	D3DXVECTOR3 vector0;

	vector0=vector;
	D3DXVec3Scale(&vector,&vector0,multiplier);
}

void Vector3::setNormalize(void)
{
	D3DXVECTOR3 vector0;

	vector0=vector;
	D3DXVec3Normalize(&vector,&vector0);
}

void Vector3::setInterpolation(Vector3* v1,Vector3* v2,double blend)
{
	D3DXVec3Lerp(&vector,&v1->vector,&v2->vector,blend);
}

void Vector3::setBaryCentric(Vector3* v1,Vector3* v2,Vector3* v3,
                             float weight2,float weight3)
{
	D3DXVec3BaryCentric(&vector,&v1->vector,&v2->vector,&v3->vector,
	                    weight2,weight3);
}


/*   ---   擾   ---   */

float Vector3::getLength(void)
{
	return D3DXVec3Length(&vector);
}

float Vector3::getLength2(void)
{
	return D3DXVec3LengthSq(&vector);
}

double Vector3::getAngle(Vector3* target)
{
	float len1,len2;

	len1=this  ->getLength();
	len2=target->getLength();
	
	if (len1==0.0 || len2==0.0) return 0.0;
	
	return (acos((vector.x/len1)*(target->vector.x/len2)+
	             (vector.y/len1)*(target->vector.y/len2)+
	             (vector.z/len1)*(target->vector.z/len2)));
}


/*   ---   Z   ---   */

Vector3* Vector3::add(Vector3* addend)
{
	Vector3* answer;

	answer=new Vector3();
	answer->setAddition(this,addend);
	
	return answer;
}

Vector3* Vector3::operator+(Vector3 addend)
{
	return add(&addend);
}

Vector3* Vector3::subtract(Vector3* value1)
{
	Vector3* answer;

	answer=new Vector3();
	answer->setSubtraction(this,value1);
	
	return answer;
}

Vector3* Vector3::operator-(Vector3 value1)
{
	return subtract(&value1);
}

float Vector3::dot(Vector3* multiplier)
{
	return D3DXVec3Dot(&vector,&multiplier->vector);
}

Vector3* Vector3::cross(Vector3* multiplier)
{
	Vector3* answer;

	answer=new Vector3();
	answer->setCrossProduct(this,multiplier);
	
	return answer;
}

Vector3* Vector3::multiply(float multiplier)
{
	Vector3* answer;

	answer=new Vector3();
	D3DXVec3Scale(&answer->vector,&vector,multiplier);
	
	return answer;
}

Vector3* Vector3::operator*(float multiplier)
{
	return multiply(multiplier);
}

Vector3* Vector3::normalize(void)
{
	Vector3* answer;

	answer=new Vector3();
	D3DXVec3Normalize(&answer->vector,&vector);
	
	return answer;
}


/*   ---   fXgN^   ---   */

Vector3::~Vector3(void)
{
}
