/*   4 xNg  Ver 0.20   */
/*                              */
/*                              */
/*   by Atsushi    1999/12/20   */


#include "basic.h"

#include "lib/util/OutOfPoolException.h"
#include "lib/util/PooledObject.h"
#include "lib/util/ObjectPool.h"
#include "Vector3.h"
#include "Vector4.h"


/*   ---   RXgN^   ---   */

Vector4::Vector4(float x,float y,float z)
{
	this->vector.x=x;
	this->vector.y=y;
	this->vector.z=z;
	this->vector.w=1.0f;
}

Vector4::Vector4(float x,float y,float z,float w)
{
	this->vector.x=x;
	this->vector.y=y;
	this->vector.z=z;
	this->vector.w=w;
}

Vector4::Vector4(Vector4* vector)
{
	this->vector=vector->vector;
}

Vector4::Vector4(Vector3* vector)
{
	this->vector.x=vector->getX();
	this->vector.y=vector->getY();
	this->vector.z=vector->getZ();
	this->vector.w=1.0f;
}

Vector4::Vector4(void)
{
	setZero();
}


/*   ---   v[   ---   */

void* Vector4::operator new(size_t size)
{
	return ObjectPool<Vector4>::newObject();
}

void Vector4::operator delete(void *object)
{
	ObjectPool<Vector4>::deleteObject((Vector4*)(object));
}


/*   ---   xNgl   ---   */

D3DXVECTOR4* Vector4::getDXVector(void)
{
	return &vector;
}

float Vector4::getX(void)
{
	return vector.x;
}

float Vector4::getY(void)
{
	return vector.y;
}

float Vector4::getZ(void)
{
	return vector.z;
}

float Vector4::getW(void)
{
	return vector.w;
}

boolean Vector4::isZero(void)
{
	return (vector.x==0.0 && vector.y==0.0 && vector.z==0.0);
}


/*   ---   ݒ   ---   */

void Vector4::set(Vector4* vector)
{
	this->vector=vector->vector;
}

void Vector4::set(float x,float y,float z)
{
	this->vector.x=x;
	this->vector.y=y;
	this->vector.z=z;
	this->vector.w=1.0f;
}

void Vector4::setZero(void)
{
	this->vector.x=0.0f;
	this->vector.y=0.0f;
	this->vector.z=0.0f;
	this->vector.w=1.0f;
}


/*   ---   vfݒ   ---   */

void Vector4::setAddition(Vector4* value,Vector4* addend)
{
	D3DXVec4Add(&vector,&value->vector,&addend->vector);
}

void Vector4::setSubtraction(Vector4* value0,Vector4* value1)
{
	D3DXVec4Subtract(&vector,&value0->vector,&value1->vector);
}

void Vector4::setCrossProduct(Vector4* multiplicand,Vector4* multiplier1,
                                                    Vector4* multiplier2)
{
	D3DXVec4Cross(&vector,&multiplicand->vector,
	              &multiplier1->vector,&multiplier2->vector);
}

void Vector4::setMultiplication(float multiplier)
{
	D3DXVECTOR4 vector0;

	vector0=vector;
	D3DXVec4Scale(&vector,&vector0,multiplier);
}

void Vector4::setNormalize(void)
{
	D3DXVECTOR4 vector0;

	vector0=vector;
	D3DXVec4Normalize(&vector,&vector0);
}

void Vector4::setInterpolation(Vector4* v1,Vector4* v2,double blend)
{
	D3DXVec4Lerp(&vector,&v1->vector,&v2->vector,blend);
}

void Vector4::setBaryCentric(Vector4* v1,Vector4* v2,Vector4* v3,
                             float weight2,float weight3)
{
	D3DXVec4BaryCentric(&vector,&v1->vector,&v2->vector,&v3->vector,
	                    weight2,weight3);
}


/*   ---   擾   ---   */

float Vector4::getLength(void)
{
	return D3DXVec4Length(&vector);
}

float Vector4::getLength2(void)
{
	return D3DXVec4LengthSq(&vector);
}

double Vector4::getAngle(Vector4* target)
{
	float mul,len1,len2;

	mul=this->dot(target);
	len1=this->getLength();
	len2=target->getLength();
	
	return (acos(mul/(len1*len2)));
}


/*   ---   Z   ---   */

Vector4* Vector4::add(Vector4* addend)
{
	Vector4* answer;

	answer=new Vector4();
	answer->setAddition(this,addend);
	
	return answer;
}

Vector4* Vector4::operator+(Vector4 addend)
{
	return add(&addend);
}

Vector4* Vector4::subtract(Vector4* value1)
{
	Vector4* answer;

	answer=new Vector4();
	answer->setSubtraction(this,value1);
	
	return answer;
}

Vector4* Vector4::operator-(Vector4 value1)
{
	return subtract(&value1);
}

float Vector4::dot(Vector4* multiplier)
{
	return D3DXVec4Dot(&vector,&multiplier->vector);
}

Vector4* Vector4::multiply(float multiplier)
{
	Vector4* answer;

	answer=new Vector4();
	D3DXVec4Scale(&answer->vector,&vector,multiplier);
	
	return answer;
}

Vector4* Vector4::operator*(float multiplier)
{
	return multiply(multiplier);
}

Vector4* Vector4::normalize(void)
{
	Vector4* answer;

	answer=new Vector4();
	D3DXVec4Normalize(&answer->vector,&vector);
	
	return answer;
}


/*   ---   fXgN^   ---   */

Vector4::~Vector4(void)
{
}
