/*   DatagramSocket NX  Ver 0.10   */
/*                                     */
/*                                     */
/*   by toyo    1999/4/29              */

#include "basic.h"
#include "../io/IOException.h"
#include "WinSockManager.h"
#include "NetworkException.h"
#include "InetAddress.h"
#include "DatagramPacket.h"
#include "DatagramSocket.h"

DatagramSocket::DatagramSocket() throw (NetworkException*){
	this->localAddress=InetAddress::getLocalHost();
	this->lpSocketAddress=localAddress->lpSocketAddress;
	this->lpSocketAddress->sin_port=0;
	initInstance();
}

DatagramSocket::DatagramSocket(int localPort) throw (NetworkException*){
	this->localAddress=InetAddress::getLocalHost();
	this->lpSocketAddress=localAddress->lpSocketAddress;
	this->lpSocketAddress->sin_port=htons(localPort);
	//this->lpSocketAddress->sin_port=localPort;
	initInstance();
}

DatagramSocket::DatagramSocket(int localPort,InetAddress *localAddress) throw (NetworkException*){
	this->localAddress=localAddress;
	this->lpSocketAddress=localAddress->lpSocketAddress;
	this->lpSocketAddress->sin_port=htons(localPort);
	initInstance();
}

void DatagramSocket::initInstance() throw (NetworkException*){
	sock=WSASocket(AF_INET,SOCK_DGRAM,IPPROTO_UDP,NULL,NULL,WSA_FLAG_OVERLAPPED);
	
	if(sock==INVALID_SOCKET){
		throw new NetworkException(new String("INVALID_SOCKET"));
	}
	
	this->overlapped.hEvent=WSACreateEvent();
	if(overlapped.hEvent==WSA_INVALID_EVENT){
		throw new NetworkException(new String("WSA_INVALID_EVENT"));
	}

	int ret=bind(sock,(LPSOCKADDR)lpSocketAddress,sizeof(SOCKADDR_IN));
	if(ret==SOCKET_ERROR){
		Debugger::println("Bind Error");
		throw new NetworkException(WinSockManager::getErrorMessage(WSAGetLastError()));
	}

	unreadDatagramPacket=NULL;
}

DatagramSocket::~DatagramSocket(){
	try{
		close();
	}catch(NetworkException *e){
		Debugger::println("DatagramSocket::~DatagramSocket()");
		Debugger::println(e->getMessage());
	}
	delete localAddress;
}

void DatagramSocket::close() throw (NetworkException*){
	int ret=closesocket(sock);
	if(ret==SOCKET_ERROR){
		//O𓊂
		throw new NetworkException(WinSockManager::getErrorMessage(WSAGetLastError()));
	}
}

InetAddress* DatagramSocket::getLocalAddress(){
	return localAddress;
}

int DatagramSocket::getLocalPort(){
	return ntohs(lpSocketAddress->sin_port);
	//return lpSocketAddress->sin_port;
}

void DatagramSocket::send(DatagramPacket *packet) throw (IOException*){
	int ret=sendto(
		sock,(char *)packet->data,packet->length,0,
		(LPSOCKADDR)packet->inetAddress->lpSocketAddress,sizeof(SOCKADDR)
	);
	
	if(ret==SOCKET_ERROR){
		throw new IOException(WinSockManager::getErrorMessage(WSAGetLastError()));
	}
}

boolean DatagramSocket::receive(DatagramPacket *packet) throw (IOException*){
	wsabuf.buf=(char *)(packet->data);
	wsabuf.len=packet->dataSize;
	
	//Debugger::println("DatagramSocket.receive()\n");

	int fromlen=sizeof(SOCKADDR);
	dwFlags=0;
	
	/*
	this->overlapped.hEvent=WSACreateEvent();
	if(overlapped.hEvent==WSA_INVALID_EVENT){
		throw new NetworkException(new String("WSA_INVALID_EVENT"));
	}
	*/
	
	int nRet=WSARecvFrom(
		sock,&wsabuf,1,&dwRecvd,
		&dwFlags,NULL,
		NULL,&overlapped,NULL
	);
	
	if(nRet==0){
		packet->setLength((int)dwRecvd);
		return true;
	}else if(nRet==SOCKET_ERROR){
		if(WSAGetLastError()==WSA_IO_PENDING){
			this->unreadDatagramPacket=packet;
			return false;
		}else{
			throw new IOException(WinSockManager::getErrorMessage(WSAGetLastError()));
		}
	}else{
		this->unreadDatagramPacket=packet;
		return false;
	}
}

//receive(DatagramPacket *packet) Ăяoɗ܂f[^ true Ԃ
boolean DatagramSocket::isReceived() throw (IOException*){
	boolean isFinished=WSAGetOverlappedResult(sock,&overlapped,&dwRecvd,FALSE,&dwFlags);
	
	if(!isFinished){
		if(WSAGetLastError()==WSA_IO_INCOMPLETE){
			return false;
		}else{
			throw new IOException(WinSockManager::getErrorMessage(WSAGetLastError()));
		}
	}else{
		if(unreadDatagramPacket!=NULL){
			//Debugger::println(unreadDatagramPacket->data);

			//memmove(unreadDatagramPacket->data,wsabuf.buf,wsabuf.len);
			
			/*
			Debugger::println("wsabuf.len");
			Debugger::println(wsabuf.len);
			*/
			
			unreadDatagramPacket->setLength((int)dwRecvd);
			unreadDatagramPacket=NULL;
			//WSACloseEvent(this->overlapped.hEvent);
			return true;
		}else{
			return false;
		}
	}
}


/*
int DatagramSocket::getSoTimeout(){
}

void DatagramSocket::setSoTimeout(){
}
*/
