/*   NetSpecAnalyzer NX  Ver 0.00  */
/*                                     */
/*                                     */
/*   by toyo    1999/11/11             */

#include "basic.h"
#include "lib/io/IOException.h"
#include "lib/io/InputStream.h"
#include "lib/io/OutputStream.h"
#include "Socket.h"
#include "NetSpec.h"
#include "NetSpecAnalyzer.h"

const int NetSpecAnalyzer::TCP_BANDWIDTH_CHECK_SIZE=1024*64;//64k
const int NetSpecAnalyzer::TCP_RESPONSE_CHECK_SIZE=32;
const int NetSpecAnalyzer::TCP_RESPONSE_CHECK_TIMES=5;
	
NetSpecAnalyzer::NetSpecAnalyzer(Socket *socket){
	this->socket=socket;
	this->in=socket->getInputStream();
	this->out=socket->getOutputStream();
}

void NetSpecAnalyzer::serveTCPResponse() throw (IOException*){
	byte buffer[TCP_RESPONSE_CHECK_SIZE];
	
	for(int i=0;i<TCP_RESPONSE_CHECK_TIMES;i++){
		write(buffer,0,TCP_RESPONSE_CHECK_SIZE);
		read(buffer,0,TCP_RESPONSE_CHECK_SIZE);
	}
}
	
//average response time
double NetSpecAnalyzer::getTCPResponse() throw (IOException*){
	clock_t sum=0;
	clock_t before;
	byte buffer[TCP_RESPONSE_CHECK_SIZE];
	
	for(int i=0;i<TCP_RESPONSE_CHECK_TIMES;i++){
		before=clock();
		write(buffer,0,TCP_RESPONSE_CHECK_SIZE);
		read(buffer,0,TCP_RESPONSE_CHECK_SIZE);
		sum+=clock()-before;
	}
	
	double averageResponseTime=((double)(sum/CLOCKS_PER_SEC))/5;
	return averageResponseTime;
}

void NetSpecAnalyzer::serveTCPBandWidth() throw (IOException*){
	byte buffer[TCP_BANDWIDTH_CHECK_SIZE];
	read(buffer,0,TCP_BANDWIDTH_CHECK_SIZE);
	write(buffer,0,TCP_BANDWIDTH_CHECK_SIZE);
}

//bytes per sec
double NetSpecAnalyzer::getTCPBandWidth() throw (IOException*){
	clock_t before;
	clock_t after;
	byte buffer[TCP_BANDWIDTH_CHECK_SIZE];

	before=clock();
	write(buffer,0,TCP_BANDWIDTH_CHECK_SIZE);
	read(buffer,0,TCP_BANDWIDTH_CHECK_SIZE);
	after=clock();
	
	double bytesPerSec=TCP_BANDWIDTH_CHECK_SIZE*2/((double)(after-before)/CLOCKS_PER_SEC);
	return bytesPerSec;
}

NetSpec *NetSpecAnalyzer::getNetSpec() throw (IOException*){
	double bytesPerSec=getTCPBandWidth();
	double responseTime=getTCPResponse();
	return new NetSpec(bytesPerSec,responseTime);
}

void NetSpecAnalyzer::serveNetSpec() throw (IOException*){
	serveTCPBandWidth();
	serveTCPResponse();
}

void NetSpecAnalyzer::read(byte *buffer,int off,int len) throw (IOException*){
	int index=0;
	while(true){
		int returnValue=in->read(buffer,off+index,len-index);
		index+=returnValue;
		if(index==len)break;
	}
}

void NetSpecAnalyzer::write(byte *buffer,int off,int len) throw (IOException*){
	out->write(buffer,off,len);
	while(!socket->isSendBufferEmpty()){
		out->flush();
		Sleep(50);
	}
}
