/*   Socket NX       Ver 0.10   */
/*                                  */
/*                                  */
/*   by toyo    1999/4/29           */

#include "basic.h"
#include "../io/IOException.h"
#include "../io/OutputStream.h"
#include "../io/InputStream.h"
#include "WinSockManager.h"
#include "NetworkException.h"
#include "InetAddress.h"
#include "TCPInputStream.h"
#include "TCPOutputStream.h"
#include "Socket.h"

#include<stdio.h>

void Socket::initInstance(InetAddress *inetAddress,int portNum) throw (NetworkException*){
	this->inetAddress=inetAddress;
	this->portNum=portNum;

	sock=WSASocket(AF_INET,SOCK_STREAM,IPPROTO_TCP,NULL,NULL,WSA_FLAG_OVERLAPPED);

	if(sock==INVALID_SOCKET){
		throw new NetworkException(WinSockManager::getErrorMessage(WSAGetLastError()));
	}

	localAddress=InetAddress::getLocalHost();
	this->tcpInputStream=NULL;
	this->tcpOutputStream=NULL;

	/*
	int sndBuf=0;
	int optlen=4;
	getsockopt(sock,SOL_SOCKET,SO_SNDBUF,(char*)(&sndBuf),&optlen);
	Debugger::println("sendBuf");
	Debugger::println(sndBuf);

	sndBuf=4096*8;
	setsockopt(sock,SOL_SOCKET,SO_SNDBUF,(char*)(&sndBuf),optlen);
	*/

	boolean tcpNodelay=true;
	int optlen=sizeof(boolean);

	setsockopt(sock,IPPROTO_TCP,TCP_NODELAY,(char*)(&tcpNodelay),optlen);
	/*
	getsockopt(sock,IPPROTO_TCP,TCP_NODELAY,(char*)(&tcpNodelay),&optlen);
	Debugger::println("TCP_NODELAY = ");
	Debugger::println(tcpNodelay);
	*/
 }

Socket::Socket(InetAddress *inetAddress,int portNum) throw (NetworkException*){
	initInstance(inetAddress,portNum);
}

/*
Socket::Socket(String *hostName,int portNum) throw (NetworkException){
	inetAddress=InetAddress::getByName(hostName);
	initInstance(inetAddress,portNum);
}
*/

Socket::~Socket(){
	try{
		close();
	}catch(NetworkException *e){
		Debugger::println("Socket::~Socket()");
		Debugger::println(e->getMessage());
	}
	delete localAddress;
	delete inetAddress;
	delete tcpInputStream;
	delete tcpOutputStream;
}

void Socket::close() throw (NetworkException*){
  int ret=closesocket(sock);
  if(ret==SOCKET_ERROR){
    //O𓊂
	  throw new NetworkException(WinSockManager::getErrorMessage(WSAGetLastError()));
  }
}

void Socket::initLocalPort(void) throw (NetworkException*){
	int namelen;

	namelen=sizeof(SOCKADDR_IN);
	memset(localAddress->lpSocketAddress,0,sizeof(SOCKADDR_IN));
	int ret=getsockname(this->sock,(LPSOCKADDR)localAddress->lpSocketAddress,&namelen);
	this->localPortNum=ntohs(localAddress->lpSocketAddress->sin_port);

	if(ret==SOCKET_ERROR){
		throw new NetworkException(WinSockManager::getErrorMessage(WSAGetLastError()));
	}
}

void Socket::makeConnection() throw (NetworkException*){
	this->hEventObject=WSACreateEvent();
	if(hEventObject==WSA_INVALID_EVENT){
		throw new NetworkException(WinSockManager::getErrorMessage(WSAGetLastError()));
	}

	int ret=WSAEventSelect(sock,hEventObject,FD_CONNECT);
	if(ret==SOCKET_ERROR){
		throw new NetworkException(WinSockManager::getErrorMessage(WSAGetLastError()));
	}

	SOCKADDR_IN cl_addr;

	memmove(&cl_addr,(char *)inetAddress->lpSocketAddress,sizeof(SOCKADDR_IN));
	cl_addr.sin_port=htons(portNum);

	int returnValue=connect(sock,(LPSOCKADDR)&cl_addr,sizeof(cl_addr));

	if(returnValue==SOCKET_ERROR){
		if(WSAGetLastError()==WSAEWOULDBLOCK){
			this->isConnectionComplete=false;
		}else{
			this->isConnectionComplete=false;
			throw new NetworkException(WinSockManager::getErrorMessage(WSAGetLastError()));
		}
	}else{
		initLocalPort();
		this->isConnectionComplete=true;
	}
}

boolean Socket::isConnected() throw (NetworkException*){
	if(isConnectionComplete){
		return true;
	}else{
		int returnValue=WSAWaitForMultipleEvents(1,&hEventObject,false,10,false);

		if(returnValue==WSA_WAIT_FAILED){
			throw new NetworkException(WinSockManager::getErrorMessage(WSAGetLastError()));
		}else if(returnValue==WSA_WAIT_TIMEOUT){
			return false;
		}

//printf("WaitForMultipleEvents returnValue: %d\n",returnValue);

		WSANETWORKEVENTS wsaNetworkEvents;

		int ret=WSAEnumNetworkEvents(sock,hEventObject,&wsaNetworkEvents);
		if(ret==SOCKET_ERROR){
			throw new NetworkException(WinSockManager::getErrorMessage(WSAGetLastError()));
		}

//printf("WSAEnumNetworkEvents %d\n",ret);

		if(wsaNetworkEvents.lNetworkEvents & FD_CONNECT){
			this->isConnectionComplete=true;
			initLocalPort();
			return true;
		}else{
			return false;
		}
	}
}

InetAddress* Socket::getInetAddress(){
	return inetAddress;
}

InputStream* Socket::getInputStream() throw (IOException*){
	if(tcpInputStream!=NULL){
		return tcpInputStream;
	}else{
		this->tcpInputStream=new TCPInputStream(sock,4096*8);
		return tcpInputStream;
	}
}

InetAddress* Socket::getLocalAddress(){
  return localAddress;
}

int Socket::getLocalPort(){
  return localPortNum;
}

OutputStream* Socket::getOutputStream() throw (IOException*){
	if(tcpOutputStream!=NULL){
		return tcpOutputStream;
	}else{
		this->tcpOutputStream=new TCPOutputStream(sock,4096*8);
		return tcpOutputStream;
	}
}

int Socket::getPort(){
  return portNum;
}


boolean Socket::isSendBufferEmpty(){
	return tcpOutputStream->isSendBufferEmpty();
}

