/*	 TCPOutputStream NX Ver 0.10	*/
/*		        						 */
/*										 */
/*	 by toyo		1999/4/29			 */

#include "basic.h"
#include "../io/IOException.h"
#include "../io/OutputStream.h"
#include "../io/InputStream.h"
#include "WinSockManager.h"
#include "TCPOutputStream.h"

const int TCPOutputStream::SLEEP_TIME=300;
const int TCPOutputStream::BLOCK_SIZE=8192;

TCPOutputStream::TCPOutputStream(SOCKET sock,DWORD bufferSize) throw (IOException*){
	initInstance(sock,bufferSize);
}

TCPOutputStream::TCPOutputStream(SOCKET sock) throw (IOException*){
	initInstance(sock,4096);
}

void TCPOutputStream::initInstance(SOCKET sock,DWORD bufferSize) throw (IOException*){
	this->sock=sock;
	this->buffer=new BYTE[BLOCK_SIZE];
	this->wsabuf.buf=(char *)buffer;
	this->wsabuf.len=0;

	this->bufferSize=bufferSize;
	this->initialBufferSize=bufferSize;

	this->subBuffer=new BYTE[bufferSize];
	this->subBufferIndex=0;
	this->subBufferOffset=0;

	this->overlapped.hEvent=WSACreateEvent();
	if(overlapped.hEvent==WSA_INVALID_EVENT){
		throw IOException(new String("WSA_INVALID_EVENT"));
		return;
	}

	this->dwFlags=0;
	this->dwSent=0;
	this->isCalledWSASend=false;
}

TCPOutputStream::~TCPOutputStream(){
	delete[] subBuffer;
	delete[] buffer;
}

void TCPOutputStream::write(BYTE b) throw (IOException*){
	buffer[subBufferIndex]=(char)b;
	if(subBufferIndex>=bufferSize){
		flush();
	}else{
		subBufferIndex++;
	}
}

void TCPOutputStream::changeSubBufferSize(DWORD newBufferSize){
	byte *newBuffer=new byte[newBufferSize];
	memmove(newBuffer,subBuffer,subBufferIndex);

	if(this->subBuffer!=NULL)delete[] this->subBuffer;
	this->subBuffer=newBuffer;
}

void TCPOutputStream::write(BYTE string[],int off,int length) throw (IOException*){
	int bufferRemainder=bufferSize-subBufferIndex;

	if(bufferRemainder>length){
		memmove(subBuffer+subBufferIndex,string+off,length);
		subBufferIndex+=length;
	}else if(bufferRemainder==length){
		memmove(subBuffer+subBufferIndex,string+off,length);
		subBufferIndex+=length;
		flush();
	}else{
		DWORD newBufferSize=bufferSize+length-bufferRemainder+BLOCK_SIZE;

		//Debugger::println("newBufferSize");
		//Debugger::println((int)newBufferSize);

		changeSubBufferSize(newBufferSize);
		memmove(subBuffer+subBufferIndex,string+off,length);
		subBufferIndex+=length;
		
		//throw new IOException(new String("obt@̃TCY𒴂đM悤ƂĂ܂"));

		/*
		memmove(subBuffer+subBufferIndex,string+off,bufferRemainder);
		flush();
		write(string,off+bufferRemainder,len-bufferRemainder);
		*/
	}
}

/*
void TCPOutputStream::write(BYTE string[],int off,int length) throw (IOException*){
	if(length>=BLOCK_SIZE){
		for(int i=0;i<length/BLOCK_SIZE;i++){
			writeSub(string,off+i*BLOCK_SIZE,BLOCK_SIZE);
			flush();
			Sleep(SLEEP_TIME);
		}

		if(length%BLOCK_SIZE!=0){
			writeSub(string,off+((int)(length/BLOCK_SIZE))*BLOCK_SIZE,length%BLOCK_SIZE);
		}
	}else{
		writeSub(string,off,length);
	}
}

void TCPOutputStream::writeSub(BYTE string[],int off,int len) throw (IOException*){
	int bufferRemainder=bufferSize-wsabuf.len;

	if(bufferRemainder>len){
		memmove(buffer+wsabuf.len,string+off,len);
		wsabuf.len+=len;
	}else if(bufferRemainder==len){
		memmove(buffer+wsabuf.len,string+off,len);
		flush();
	}else{
		memmove(buffer+wsabuf.len,string+off,bufferRemainder);
		flush();
		writeSub(string,off+bufferRemainder,len-bufferRemainder);
	}
}
*/


void TCPOutputStream::flush() throw (IOException*){
	if(isCalledWSASend){
		if(isSent()){
			isCalledWSASend=false;

			//Debugger::println("flush(): isSent() is true");
			//Debugger::println("subBufferOffset:");
			//Debugger::println(subBufferOffset);

			if(subBufferIndex>0){
				if(subBufferIndex>BLOCK_SIZE){

					//Debugger::println("subBufferIndex");
					//Debugger::println(subBufferIndex);
					//Debugger::println("subBufferOffset");
					//Debugger::println(subBufferOffset);
					
					

					if(subBufferIndex-subBufferOffset>BLOCK_SIZE){
						memmove(buffer,subBuffer+subBufferOffset,BLOCK_SIZE);
						wsabuf.len=BLOCK_SIZE;
						send();
						subBufferOffset+=BLOCK_SIZE;
					}else{
						memmove(buffer,subBuffer+subBufferOffset,subBufferIndex-subBufferOffset);
						wsabuf.len=subBufferIndex-subBufferOffset;
						send();
						subBufferOffset=0;
						subBufferIndex=0;
					}
				}else{
					memmove(buffer,subBuffer+subBufferOffset,subBufferIndex);
					wsabuf.len=subBufferIndex;
					send();
					subBufferIndex=0;
				}
			}
		}else{
			//Debugger::println("flush(): isSent() is false");
		}
	}else{
		if(subBufferIndex>0){
			if(subBufferIndex>BLOCK_SIZE){
				//Debugger::println("subBufferIndex");
				//Debugger::println(subBufferIndex);
				//Debugger::println("subBufferOffset");
				//Debugger::println(subBufferOffset);
			
				if(subBufferIndex-subBufferOffset>BLOCK_SIZE){
					memmove(buffer,subBuffer+subBufferOffset,BLOCK_SIZE);
					wsabuf.len=BLOCK_SIZE;
					send();
					subBufferOffset+=BLOCK_SIZE;
				}else{
					memmove(buffer,subBuffer+subBufferOffset,subBufferIndex-subBufferOffset);
					wsabuf.len=subBufferIndex-subBufferOffset;
					send();
					subBufferOffset=0;
					subBufferIndex=0;
				}
			}else{
				memmove(buffer,subBuffer+subBufferOffset,subBufferIndex);
				wsabuf.len=subBufferIndex;
				send();
				subBufferIndex=0;
			}
		}
	}
}

/*
boolean TCPOutputStream::send() throw (IOException){
	dwFlags=0;

	int returnValue=::send(sock,(char *)(buffer+currentBlock*BLOCK_SIZE),(int)wsabuf.len,(int)dwFlags);

	Sleep(10);

	currentBlock=(currentBlock+1)%BLOCK_NUM;

	Debugger::println("send() returnValue= ");
	Debugger::println(returnValue);

	isCalledWSASend=false;
	return true;
}
*/

boolean TCPOutputStream::send() throw (IOException){
	dwFlags=0;

	wsabuf.buf=(char *)buffer;
	int returnValue=WSASend(sock,&wsabuf,1,&dwSent,dwFlags,&overlapped,NULL);
	//int returnValue=WSASend(sock,&wsabuf,1,&dwSent,dwFlags,NULL,NULL);

	isCalledWSASend=true;
	
	if(returnValue==0){
		isCalledWSASend=false;
		
		return true;
	}else if(returnValue==SOCKET_ERROR){
		if(WSAGetLastError()==WSA_IO_PENDING){
			return false;
		}else{
			//O𓊂鏈
			throw IOException(WinSockManager::getErrorMessage(WSAGetLastError()));
		}
	}else{
		return false;
	}
}


boolean TCPOutputStream::isSent() throw (IOException){
	boolean isFinished=WSAGetOverlappedResult(sock,&overlapped,&dwSent,FALSE,&dwFlags);
	if(!isFinished){
		if(WSAGetLastError()==WSA_IO_INCOMPLETE){
			return false;
		}else{
			//O𓊂鏈
			throw IOException(WinSockManager::getErrorMessage(WSAGetLastError()));
		}
	}else if(dwSent==0){
		return false;
	}else{
		return true;
	}
}


boolean TCPOutputStream::isSendBufferEmpty(){
	if(subBufferIndex-subBufferOffset>0){
		return false;
	}else{
		return true;
	}
}

