/*   DirectSoundManager NX   Ver 0.00 */
/*                                        */
/*                                        */
/*   by Toyonori    1999/4/4              */

#include "basic.h"

#include "lib/io/InputStream.h"
#include "lib/io/OutputStream.h"
#include "lib/io/ObjectInputStream.h"
#include "lib/io/ObjectOutputStream.h"
#include "lib/io/Serializable.h"

#include "WaveFormat.h"
#include "SoundBuffer.h"
#include "SoundData.h"
#include "SoundData3D.h"
#include "SoundListener3D.h"

#include "SoundRecorder.h"
#include "DirectSoundRecorder.h"
#include "WaveformAudioRecorder.h"
#include "StreamSoundPlayer.h"
#include "DirectSoundManager.h"

DirectSoundManager::DirectSoundManager(){
  lpDirectSound=NULL;
  lpDirectSoundCapture=NULL;
  lpDirectMusic=NULL;
}

DirectSoundManager::~DirectSoundManager(){
	if(lpDirectSound!=NULL){
		lpDirectSound->Release();
	}
	
	if(lpDirectSoundCapture!=NULL){
		lpDirectSoundCapture->Release();
	}
}

boolean DirectSoundManager::initDirectSound(HWND hWnd){
	HRESULT hr;
	
	WaveFormat *waveFormat=new WaveFormat(1,8000,1,8);
	initWaveformAudioRecorder(waveFormat);
	
	memset(&dsCaps,0,sizeof(DSCAPS));
	dsCaps.dwSize=sizeof(DSCAPS);
	
	hr=DirectSoundEnumerate((LPDSENUMCALLBACK)DSEnumProc,this);
	if(hr!=DS_OK){
		Debugger::printErrorDSound(hr);
		return false;
	}
	
	hr=lpDirectSound->SetCooperativeLevel(hWnd,DSSCL_PRIORITY);
	//hr=lpDirectSound->SetCooperativeLevel(hWnd,DSSCL_WRITEPRIMARY);
	if(hr!=DS_OK){
		Debugger::printErrorDSound(hr);
		return false;
	}
	
	return true;
}

boolean CALLBACK DirectSoundManager::DSEnumProc(
	LPGUID lpGUID,
	LPCTSTR lpszDesc,
	LPCTSTR lpszDrvName,
	LPVOID lpContext){
	
	LPGUID lpTemp = NULL;
	
	HRESULT hr;
	
	DirectSoundManager *directSoundManager=(DirectSoundManager*)lpContext;
	
	LPDIRECTSOUND lpTempDirectSound;
	DSCAPS tempDSCAPS;
	
    if(lpGUID!=NULL){
		if((lpTemp=(LPGUID)LocalAlloc(LPTR,sizeof(GUID)))==NULL)return true;
		memcpy(lpTemp,lpGUID,sizeof(GUID));
		
		hr=DirectSoundCreate(lpTemp,&lpTempDirectSound,NULL);
		if(hr!=DS_OK){
			Debugger::printErrorDSound(hr);
			return false;
		}
		
		memset(&tempDSCAPS,0,sizeof(DSCAPS));
		tempDSCAPS.dwSize=sizeof(DSCAPS);
		
		hr=lpTempDirectSound->GetCaps(&tempDSCAPS);
		if(hr!=DS_OK){
			Debugger::printErrorDSound(hr);
			return false;
		}
		
		if(directSoundManager->lpDirectSound!=NULL){
			if(evaluateDSCAPS(tempDSCAPS)>evaluateDSCAPS(directSoundManager->dsCaps)){
				directSoundManager->lpDirectSound->Release();
				directSoundManager->lpDirectSound=NULL;
				directSoundManager->lpDirectSound=lpTempDirectSound;
				directSoundManager->dsCaps=tempDSCAPS;
			}
		}else{
			evaluateDSCAPS(tempDSCAPS);
			directSoundManager->lpDirectSound=lpTempDirectSound;
			directSoundManager->dsCaps=tempDSCAPS;
		}
	}
	
	/*
	Debugger::println("Driver Name=");
	Debugger::println((char *)lpszDrvName);
	Debugger::println("\n");
	
	Debugger::println("Description=");
	Debugger::println((char *)lpszDesc);
	Debugger::println("\n");
	*/
	
	return true;
}

DWORD DirectSoundManager::evaluateDSCAPS(DSCAPS caps){
	DWORD point=0;
	
	//point+=caps.dwFreeHw3DAllBuffers;
	//point+=caps.dwFreeHw3DStaticBuffers;
	//point+=caps.dwFreeHw3DStreamingBuffers;
	//point+=caps.dwFreeHwMemBytes;
	//point+=caps.dwFreeHwMixingAllBuffers;
	//point+=caps.dwFreeHwMixingStaticBuffers;
	//point+=caps.dwFreeHwMixingStreamingBuffers;
	//point+=caps.dwMaxContigFreeHwMemBytes;
	point+=caps.dwMaxHw3DAllBuffers*1000*500;
	
	/*
	Debugger::println("dwMaxHw3DAllBuffers");
	Debugger::println((int)caps.dwMaxHw3DAllBuffers);
	Debugger::println("\n");
	*/
	
	//point+=caps.dwMaxHw3DStaticBuffers;
	//point+=caps.dwMaxHwMixingStreamingBuffers;
	//point+=caps.dwMaxSecondarySampleRate;
	//point+=caps.dwMinSecondarySampleRate;
	//point+=caps.dwPlayCpuOverheadSwBuffers;
	//point+=caps.dwPrimaryBuffers;
	point+=caps.dwTotalHwMemBytes;
	
	/*
	Debugger::println("dwTotalHwMemBytes");
	Debugger::println((int)caps.dwTotalHwMemBytes);
	Debugger::println("\n");
	*/
	
	//point+=caps.dwUnlockTransferRateHwBuffers;

	return point;
}

boolean CALLBACK DirectSoundManager::DSCaptureEnumProc(
	LPGUID lpGUID,
	LPCTSTR lpszDesc,
	LPCTSTR lpszDrvName,
	LPVOID lpContext){
	
	LPGUID lpTemp = NULL;
	
	HRESULT hr;
	
	DirectSoundManager *directSoundManager=(DirectSoundManager*)lpContext;
	
	LPDIRECTSOUNDCAPTURE lpTempDirectSoundCapture;
	DSCCAPS tempDSCaptureCAPS;
	
    if(lpGUID!=NULL){
		if((lpTemp=(LPGUID)LocalAlloc(LPTR,sizeof(GUID)))==NULL)return true;
		memcpy(lpTemp,lpGUID,sizeof(GUID));
		
		hr=DirectSoundCaptureCreate(lpTemp,&lpTempDirectSoundCapture,NULL);
		if(hr!=DS_OK){
			Debugger::printErrorDSound(hr);
			return false;
		}
		
		memset(&tempDSCaptureCAPS,0,sizeof(DSCCAPS));
		tempDSCaptureCAPS.dwSize=sizeof(DSCCAPS);
		
		hr=lpTempDirectSoundCapture->GetCaps(&tempDSCaptureCAPS);
		if(hr!=DS_OK){
			Debugger::printErrorDSound(hr);
			return false;
		}
		
		if(directSoundManager->lpDirectSoundCapture!=NULL){
			if(evaluateDSCaptureCAPS(tempDSCaptureCAPS)>evaluateDSCaptureCAPS(directSoundManager->dsCaptureCaps)){
				directSoundManager->lpDirectSoundCapture->Release();
				directSoundManager->lpDirectSoundCapture=NULL;
				directSoundManager->lpDirectSoundCapture=lpTempDirectSoundCapture;
				directSoundManager->dsCaptureCaps=tempDSCaptureCAPS;
			}
		}else{
			evaluateDSCaptureCAPS(tempDSCaptureCAPS);
			directSoundManager->lpDirectSoundCapture=lpTempDirectSoundCapture;
			directSoundManager->dsCaptureCaps=tempDSCaptureCAPS;
		}
	}
	
	/*
	Debugger::println("Driver Name=");
	Debugger::println((char *)lpszDrvName);
	Debugger::println("\n");
	
	Debugger::println("Description=");
	Debugger::println((char *)lpszDesc);
	Debugger::println("\n");
	*/
	return true;
}

DWORD DirectSoundManager::evaluateDSCaptureCAPS(DSCCAPS caps){
	DWORD point=0;
	
	switch(caps.dwFormats){
	case WAVE_FORMAT_1M08:
		point=1;
		break;
	case WAVE_FORMAT_1M16:
		point=2;
		break;
	case WAVE_FORMAT_1S08:
		point=3;
		break;
	case WAVE_FORMAT_1S16:
		point=4;
		break;
	case WAVE_FORMAT_2M08: 
		point=5;
		break;
	case WAVE_FORMAT_2M16: 
		point=6;
		break;
	case WAVE_FORMAT_2S08: 
		point=7;
		break;
	case WAVE_FORMAT_2S16: 
		point=8;
		break;
	case WAVE_FORMAT_4M08: 
		point=9;
		break;
	case WAVE_FORMAT_4M16: 
		point=10;
		break;
	case WAVE_FORMAT_4S08: 
		point=11;
		break;
	case WAVE_FORMAT_4S16: 
		point=12;
		break;
	}

	return point;
}

boolean DirectSoundManager::initDirectSoundCapture(){
	HRESULT hr;
  
	memset(&dsCaptureCaps,0,sizeof(DSCCAPS));
	dsCaptureCaps.dwSize=sizeof(DSCCAPS);
	
	hr=DirectSoundCaptureEnumerate((LPDSENUMCALLBACK)DSCaptureEnumProc,this);
	if(hr!=DS_OK){
		Debugger::printErrorDSound(hr);
		return false;
	}

	return true;
}

boolean DirectSoundManager::initDirectMusic(HWND hWnd){
	CoInitialize(NULL);
	
	if(lpDirectMusic){
		return false;
	}
	
	if(FAILED(CoCreateInstance(
		CLSID_DirectMusic,NULL,
		CLSCTX_INPROC_SERVER,IID_IDirectMusic,
        (VOID**)&lpDirectMusic))){
		Debugger::println("Could not create DirectMusic!");
		return false;
    }
	
    if(FAILED(lpDirectMusic->SetDirectSound(lpDirectSound,hWnd))){
		Debugger::println("Initialization Error");
		return false;
    }
	
	return true;
}

boolean DirectSoundManager::releaseDirectSound(void){
	if(lpDirectSound!=NULL){
		lpDirectSound->Release();
		lpDirectSound=NULL;
		return true;
	}else{
		return false;
	}
}

boolean DirectSoundManager::releaseDirectSoundCapture(void){
	if(lpDirectSoundCapture!=NULL){
		lpDirectSoundCapture->Release();
		lpDirectSoundCapture=NULL;
		return true;
	}else{
		return false;
	}
}

SoundRecorder* DirectSoundManager::getDirectSoundRecorder(
	WaveFormat *waveFormat,
	SoundBuffer *soundBuffer
	){
	
	if(lpDirectSoundCapture==NULL){
		return NULL;
	}

	LPDIRECTSOUNDCAPTUREBUFFER lpDirectSoundCaptureBuffer;
	DSCBUFFERDESC dsCaptureBufferDesc;
	
	memset(&dsCaptureBufferDesc,0,sizeof(DSCBUFFERDESC));
	
	dsCaptureBufferDesc.dwSize=sizeof(DSCBUFFERDESC);
	dsCaptureBufferDesc.dwFlags=0;
	dsCaptureBufferDesc.dwBufferBytes=soundBuffer->getBufferSize();
	dsCaptureBufferDesc.dwReserved=0;
	dsCaptureBufferDesc.lpwfxFormat=&(waveFormat->waveFormatEx);
		
	lpDirectSoundCaptureBuffer=NULL;
		
	HRESULT hr=lpDirectSoundCapture->CreateCaptureBuffer(
		&dsCaptureBufferDesc,
		&lpDirectSoundCaptureBuffer,
		NULL
	);
	
	if(hr!=DS_OK){
		//Debugger::println("init DirectSoundCaptureBuffer:");
		//Debugger::printErrorDSound(hr);
		return NULL;
	}
	

	DirectSoundRecorder *recorder=new DirectSoundRecorder(
		lpDirectSoundCaptureBuffer,
		soundBuffer
	);
	
	return recorder;
}

boolean DirectSoundManager::initWaveformAudioRecorder(WaveFormat *waveFormat){
	MMRESULT mmr=waveInOpen(&hWaveIn,WAVE_MAPPER,&(waveFormat->waveFormatEx),0,0,CALLBACK_NULL);
	if(mmr!=MMSYSERR_NOERROR){
		char buf[128];
		waveInGetErrorText(mmr,buf,128);
		Debugger::println(buf);
		Debugger::println("\n");
		return false;
	}else{
		return true;
	}
}

SoundRecorder* DirectSoundManager::getWaveformAudioRecorder(
	SoundBuffer *soundBuffer,DWORD dwCompressionRate
	){
	return new WaveformAudioRecorder(hWaveIn,dwCompressionRate,soundBuffer);
}

SoundRecorder* DirectSoundManager::getWaveformAudioRecorder(SoundBuffer *soundBuffer){
	return new WaveformAudioRecorder(hWaveIn,soundBuffer);
}
  
StreamSoundPlayer* DirectSoundManager::getStreamSoundPlayer(
	WaveFormat *waveFormat,
	SoundBuffer *soundBuffer
	){
	if(lpDirectSound==NULL){
		return NULL;
	}

	LPDIRECTSOUNDBUFFER lpDirectSoundBuffer;
	DSBUFFERDESC dsBufferDesc;
	
	memset(&dsBufferDesc,0,sizeof(DSBUFFERDESC));
	
	dsBufferDesc.dwSize=sizeof(DSBUFFERDESC);
	dsBufferDesc.dwFlags=DSBCAPS_CTRLPAN | DSBCAPS_CTRLVOLUME | DSBCAPS_CTRLFREQUENCY; 
	dsBufferDesc.dwBufferBytes=soundBuffer->getBufferSize();
	dsBufferDesc.lpwfxFormat=&(waveFormat->waveFormatEx);
	
	lpDirectSoundBuffer=NULL;
	
	HRESULT hr=lpDirectSound->CreateSoundBuffer(
		&dsBufferDesc,
		&lpDirectSoundBuffer,
		NULL
	);
		
	if(hr!=DS_OK){
		//Debugger::println("init DirectSoundBuffer:");
		//Debugger::printErrorDSound(hr);
		return NULL;
	}

	StreamSoundPlayer *player=new StreamSoundPlayer(
		lpDirectSoundBuffer,
		soundBuffer
	);

	return player;
}

StreamSoundPlayer* DirectSoundManager::getStreamSoundPlayer(
	WaveFormat *waveFormat,
	DWORD dwUncompressionRate,
	SoundBuffer *soundBuffer){
	
	if(lpDirectSound==NULL){
		return NULL;
	}

	LPDIRECTSOUNDBUFFER lpDirectSoundBuffer;
	DSBUFFERDESC dsBufferDesc;
	
	memset(&dsBufferDesc,0,sizeof(DSBUFFERDESC));
	
	dsBufferDesc.dwSize=sizeof(DSBUFFERDESC);
	dsBufferDesc.dwFlags=DSBCAPS_CTRLPAN | DSBCAPS_CTRLVOLUME | DSBCAPS_CTRLFREQUENCY; 
	dsBufferDesc.dwBufferBytes=soundBuffer->getBufferSize();
	dsBufferDesc.lpwfxFormat=&(waveFormat->waveFormatEx);
	
	lpDirectSoundBuffer=NULL;
	
	HRESULT hr=lpDirectSound->CreateSoundBuffer(
		&dsBufferDesc,
		&lpDirectSoundBuffer,
		NULL
	);
		
	if(hr!=DS_OK){
		//Debugger::println("init DirectSoundBuffer:");
		//Debugger::printErrorDSound(hr);
		return NULL;
	}

	StreamSoundPlayer *player=new StreamSoundPlayer(
		lpDirectSoundBuffer,
		dwUncompressionRate,
		soundBuffer
	);

	return player;
}

LPDIRECTSOUNDBUFFER DirectSoundManager::getDirectSoundBuffer(DWORD dwFlags){
	LPDIRECTSOUNDBUFFER lpDirectSoundBuffer=NULL;
	DSBUFFERDESC dsBufferDesc;
	
	memset(&dsBufferDesc,0,sizeof(DSBUFFERDESC));
	dsBufferDesc.dwSize  = sizeof(DSBUFFERDESC);
	dsBufferDesc.dwFlags=dwFlags; 
	
	HRESULT hr=lpDirectSound->CreateSoundBuffer(
		&dsBufferDesc,
		&lpDirectSoundBuffer,
		NULL
	);
	
	if(hr!=DS_OK){
		Debugger::println("getDirectSoundBuffer:");
		Debugger::printErrorDSound(hr);
		return NULL;
	}
	
	return lpDirectSoundBuffer;
}

LPDIRECTSOUNDBUFFER DirectSoundManager::getDirectSoundBuffer(SoundData *soundData){
	DSBUFFERDESC dsBufferDesc;
	
	memset(&dsBufferDesc,0,sizeof(DSBUFFERDESC));
	
	dsBufferDesc.dwSize=sizeof(DSBUFFERDESC);
	dsBufferDesc.dwFlags=DSBCAPS_CTRLPAN | DSBCAPS_CTRLVOLUME | DSBCAPS_CTRLFREQUENCY; 
	dsBufferDesc.dwBufferBytes=soundData->soundBuffer->getBufferSize();
	dsBufferDesc.lpwfxFormat=&(soundData->waveFormat->waveFormatEx);
	
	soundData->lpDirectSoundBuffer=NULL;
	
	HRESULT hr=lpDirectSound->CreateSoundBuffer(
		&dsBufferDesc,
		&soundData->lpDirectSoundBuffer,
		NULL
	);
		
	if(hr!=DS_OK){
		//Debugger::println("getDirectSoundBuffer:");
		//Debugger::printErrorDSound(hr);
		return NULL;
	}
	
	return soundData->lpDirectSoundBuffer;
}

LPDIRECTSOUND3DBUFFER DirectSoundManager::getDirectSound3DBuffer(
	SoundData3D *soundData
	){
	return getDirectSound3DBuffer(
		soundData,DSBCAPS_CTRL3D | DSBCAPS_CTRLVOLUME |
		DSBCAPS_CTRLFREQUENCY | DSBCAPS_GETCURRENTPOSITION2 |
		DSBCAPS_STICKYFOCUS 
	);
	//DSBCAPS_STATIC
}

LPDIRECTSOUND3DBUFFER DirectSoundManager::getDirectSound3DBuffer(
	SoundData3D *soundData,
	DWORD dwFlags
	){
	DSBUFFERDESC dsBufferDesc;
	
	memset(&dsBufferDesc,0,sizeof(DSBUFFERDESC));
	
	dsBufferDesc.dwSize=sizeof(DSBUFFERDESC);
	dsBufferDesc.dwFlags=dwFlags | DSBCAPS_STATIC;
	//dsBufferDesc.dwFlags=dwFlags;
	dsBufferDesc.dwBufferBytes=soundData->soundBuffer->getBufferSize();
	dsBufferDesc.lpwfxFormat=&(soundData->waveFormat->waveFormatEx);
	
	int i;
	for(i=0;i<soundData->getDuplicateSoundNum();i++){
		soundData->setDirectSoundBuffer(NULL,i);
	}
	
	HRESULT hr=lpDirectSound->CreateSoundBuffer(
		&dsBufferDesc,
		&soundData->lpDirectSoundBuffer,
		NULL
	);
	
	if(hr!=DS_OK){
		soundData->setStatic(false);
		
		dsBufferDesc.dwFlags=dwFlags;
		
		hr=lpDirectSound->CreateSoundBuffer(
			&dsBufferDesc,
			&soundData->lpDirectSoundBuffer,
			NULL
		);
		
		soundData->lpDirectSoundBuffer->SetVolume(DSBVOLUME_MAX);
		
		if(hr!=DS_OK){
			/*
			Debugger::println("getDirectSound3DBuffer(soundData3D,dwFlags):");
			Debugger::printErrorDSound(hr);
			*/
			return NULL;
		}
	}
	
	for(i=0;i<soundData->getDuplicateSoundNum()-1;i++){
		lpDirectSound->DuplicateSoundBuffer(
			soundData->lpDirectSoundBuffer,
			&(soundData->lpDirectSoundBufferArray[i])
		);
	}
	
	for(i=0;i<soundData->duplicateSoundNum;i++){
		hr=soundData->getDirectSoundBuffer(i)->QueryInterface(
			IID_IDirectSound3DBuffer,
			(void **)&(soundData->lpDirectSound3DBuffer[i])
		);
	}
	
	if(hr!=DS_OK){
		//Debugger::printErrorDSound(hr);
		return NULL;
	}
	
	return soundData->lpDirectSound3DBuffer[0];
}

SoundListener3D *DirectSoundManager::getSoundListener3D(void){
	SoundListener3D *soundListener3D=new SoundListener3D();
	
	LPDIRECTSOUNDBUFFER lpPrimaryDirectSoundBuffer;
	LPDIRECTSOUND3DLISTENER lpDirectSound3DListener;
	
	lpPrimaryDirectSoundBuffer=getDirectSoundBuffer(
		DSBCAPS_CTRL3D | DSBCAPS_CTRLVOLUME |
		DSBCAPS_PRIMARYBUFFER
	);
	
	lpPrimaryDirectSoundBuffer->SetVolume(DSBVOLUME_MAX);
	
	WAVEFORMATEX wfm;

	// Set primary buffer format.
    ZeroMemory( &wfm, sizeof( WAVEFORMATEX)); 
    wfm.wFormatTag      = WAVE_FORMAT_PCM; 
    wfm.nChannels       = 2; 
    wfm.nSamplesPerSec  = 44100; 
    wfm.wBitsPerSample  = 16; 
    wfm.nBlockAlign     = wfm.wBitsPerSample / 8 * wfm.nChannels;
    wfm.nAvgBytesPerSec = wfm.nSamplesPerSec * wfm.nBlockAlign;
	
    lpPrimaryDirectSoundBuffer->SetFormat( &wfm ); 

	HRESULT hr=lpPrimaryDirectSoundBuffer->QueryInterface(
		IID_IDirectSound3DListener, 
		(void **)&lpDirectSound3DListener
	);
	
	if(hr!=DS_OK){
		//Debugger::println("DirectSoundManager::getSoundListener3D()");
		//Debugger::printErrorDSound(hr);
		return NULL;
	}
	
	soundListener3D->lpDirectSoundBuffer=lpPrimaryDirectSoundBuffer;
	//lpPrimaryDirectSoundBuffer->Release();
	//lpPrimaryDirectSoundBuffer=NULL;
	//soundListener3D->lpDirectSoundBuffer=NULL;
	
	soundListener3D->lpDirectSound3DListener=lpDirectSound3DListener;
	
	return soundListener3D;
}

LPDIRECTSOUND DirectSoundManager::getDirectSound(){
  return lpDirectSound;
}

DSCAPS DirectSoundManager::getDirectSoundCaps(){
  return dsCaps;
}

LPDIRECTSOUNDCAPTURE DirectSoundManager::getDirectSoundCapture(){
  return lpDirectSoundCapture;
}

DSCCAPS DirectSoundManager::getDirectSoundCaptureCaps(){
  return dsCaptureCaps;
}

IDirectMusic *DirectSoundManager::getDirectMusic(){
	return this->lpDirectMusic;
}


