/*	 MidiPlayer NX		Ver 0.00 */
/*									*/
/*									*/
/*	 by toyo		1999/7/31		*/

#include "basic.h"

#include "lib/io/InputStream.h"
#include "lib/io/OutputStream.h"
#include "lib/io/ObjectInputStream.h"
#include "lib/io/ObjectOutputStream.h"
#include "lib/io/Serializable.h"

#include "WaveFormat.h"
#include "SoundBuffer.h"
#include "SoundData.h"
#include "SoundData3D.h"
#include "SoundListener3D.h"

#include "SoundRecorder.h"
#include "DirectSoundRecorder.h"
#include "WaveformAudioRecorder.h"
#include "StreamSoundPlayer.h"
#include "DirectSoundManager.h"

#include "lib/math/Vector3.h"

#include "MidiData.h"
#include "MidiPlayer.h"

MidiPlayer::MidiPlayer(DirectSoundManager *directSoundManager){
	this->directSoundManager=directSoundManager;
	this->lpDirectMusic=NULL;
	
	lpDirectMusicLoader=NULL;
	lpDirectMusicPerformance=NULL;
	lpDirectMusicPort=NULL;

	this->playing=false;
	init();
}

MidiPlayer::~MidiPlayer(void){
	if(this->lpDirectMusicLoader){
		this->lpDirectMusicLoader->Release();
		this->lpDirectMusicLoader=NULL;
	}
	
	if(this->lpDirectMusicPort){
		this->lpDirectMusicPort->Release();
		this->lpDirectMusicPort=NULL;
	}

	if(this->lpDirectMusicPerformance){
		lpDirectMusicPerformance->Stop( NULL, NULL, 0, 0 );

        // CloseDown and Release the performance object.
        lpDirectMusicPerformance->CloseDown();
        lpDirectMusicPerformance->Release();
        lpDirectMusicPerformance = NULL;
	}
}

void MidiPlayer::init(void){
	this->lpDirectMusic=directSoundManager->getDirectMusic();
	
	initDirectMusicPerformance();
	initDirectMusicPort();
	initDirectMusicLoader();
	
	if(!lpDirectMusicPerformance || !lpDirectMusicPort){
		Debugger::println("Initializing MidiPlayer failed.");
		return;
	}
	
	if(SUCCEEDED(lpDirectMusicPerformance->AddPort(lpDirectMusicPort))){
        if(SUCCEEDED(lpDirectMusicPerformance->AssignPChannelBlock(0,lpDirectMusicPort,1))){
            
			/*
			defaultVolume=100;
			
			HRESULT returnValue=this->lpDirectMusicPerformance->SetGlobalParam(
				GUID_PerfMasterVolume,
				&defaultVolume,
				sizeof(defaultVolume)
			);	
			
			defaultVolume=0;
			returnValue=this->lpDirectMusicPerformance->GetGlobalParam(
				GUID_PerfMasterVolume,
				&defaultVolume,
				sizeof(defaultVolume)
			);
			
			defaultVolume=50;
			returnValue=this->lpDirectMusicPerformance->SetGlobalParam(
				GUID_PerfMasterVolume,
				&defaultVolume,
				sizeof(defaultVolume)
			);	

			if(returnValue!=S_OK){
				Debugger::printErrorDMusic(returnValue);
			}else{
				Debugger::println("defaultVolume:");
				Debugger::println(defaultVolume);
			}
			*/

			return;
		}else{
			Debugger::println("Initializing MidiPlayer failed.");
		}
    }else{
		Debugger::println("Initializing MidiPlayer failed.");
	}
	
	
}

void MidiPlayer::initDirectMusicLoader(void){
	IDirectMusicLoader* pLoader;
	
	if(FAILED(CoCreateInstance(
		CLSID_DirectMusicLoader,NULL,CLSCTX_INPROC,
		IID_IDirectMusicLoader,(VOID**)&pLoader))){
		pLoader=NULL;
		Debugger::println("Could't create IDirectMusicLoader.");
	}
	
	this->lpDirectMusicLoader=(LPDMUS_LOADER)pLoader;
}

void MidiPlayer::initDirectMusicPort(void){
    if(!lpDirectMusic){
		Debugger::println("DirectMusic is null");
		return;
	}
	
    IDirectMusicPort* pPort = NULL;
    HRESULT           hr;
    GUID              guidPortGUID;
    DMUS_PORTPARAMS   dmos;
    DMUS_PORTCAPS     dmpc;
	
    // See if the default port suits our needs
    if( SUCCEEDED(lpDirectMusic->GetDefaultPort( &guidPortGUID ) ) )
	{
        ZeroMemory( &dmos, sizeof(dmos) );
        dmos.dwSize          = sizeof(DMUS_PORTPARAMS);
        dmos.dwChannelGroups = 1;
        dmos.dwValidParams   = DMUS_PORTPARAMS_CHANNELGROUPS;

        if( SUCCEEDED( lpDirectMusic->CreatePort( guidPortGUID, &dmos, &pPort,
                                              NULL ) ) )
        {
            ZeroMemory( &dmpc, sizeof(dmpc) );
            dmpc.dwSize = sizeof(DMUS_PORTCAPS);

            if( SUCCEEDED( pPort->GetCaps( &dmpc ) ) )
            {
                if( ( dmpc.dwClass != DMUS_PC_OUTPUTCLASS ) ||
                   !( dmpc.dwFlags & DMUS_PC_DLS ) ||
                   !( dmpc.dwFlags & DMUS_PC_DIRECTSOUND ) )
                {
                    pPort->Release();
                    pPort = NULL;
                }
            }
            else
            {
                pPort->Release();
                pPort = NULL;
            }
        }
    }

    if( NULL == pPort )
    {
        // Find a software synth output port.
        for( DWORD index = 0; ; index++ )
        {
            ZeroMemory( &dmpc, sizeof(dmpc) );
            dmpc.dwSize = sizeof(DMUS_PORTCAPS);

            hr = lpDirectMusic->EnumPort( index, &dmpc );

            if( SUCCEEDED(hr) && hr != S_FALSE )
            {
                if( ( dmpc.dwClass == DMUS_PC_OUTPUTCLASS ) &&
                    ( dmpc.dwFlags & DMUS_PC_DLS ) &&
                    ( dmpc.dwFlags & DMUS_PC_DIRECTSOUND ) )
                {
                    CopyMemory( &guidPortGUID, &dmpc.guidPort, sizeof(GUID) );

                    ZeroMemory( &dmos, sizeof(dmos) );
                    dmos.dwSize          = sizeof(DMUS_PORTPARAMS);
                    dmos.dwChannelGroups = 1;
                    dmos.dwValidParams   = DMUS_PORTPARAMS_CHANNELGROUPS;

                    if( FAILED( lpDirectMusic->CreatePort( guidPortGUID, &dmos,
                                                       &pPort, NULL ) ) )
                    {
                        pPort = NULL;
                    }
                    break;
                }
            }
            else
            {
                break;
            }
        }
    }

	this->lpDirectMusicPort=pPort;

	WAVEFORMATEX _WaveFormat;
    ZeroMemory( &_WaveFormat, sizeof(_WaveFormat) );
	
	DSBUFFERDESC dsBD;
	
    ZeroMemory( &dsBD, sizeof(dsBD) );
    dsBD.dwSize      = sizeof(dsBD);
    dsBD.dwFlags     = DSBCAPS_CTRL3D | DSBCAPS_CTRLVOLUME |
                       DSBCAPS_CTRLFREQUENCY | DSBCAPS_GETCURRENTPOSITION2 | 
                       DSBCAPS_STICKYFOCUS;
    dsBD.lpwfxFormat = &_WaveFormat;

    DWORD dwWFXSize = sizeof(_WaveFormat);
    if( FAILED( hr = lpDirectMusicPort->GetFormat( &_WaveFormat, &dwWFXSize,
                                                &dsBD.dwBufferBytes ) ) )
    {
        _WaveFormat.wFormatTag      = WAVE_FORMAT_PCM;
        _WaveFormat.nChannels       = 2;
        _WaveFormat.nSamplesPerSec  = 22050;
        _WaveFormat.wBitsPerSample  = 16;
        _WaveFormat.nBlockAlign     = _WaveFormat.nChannels *  
                                      ( _WaveFormat.wBitsPerSample / 8 );
        _WaveFormat.nAvgBytesPerSec = _WaveFormat.nSamplesPerSec * _WaveFormat.nBlockAlign;
        _WaveFormat.cbSize          = sizeof(_WaveFormat);

        dsBD.dwBufferBytes = _WaveFormat.nAvgBytesPerSec;
    }
	
	IDirectSound *lpDirectSound=directSoundManager->getDirectSound();

    if( SUCCEEDED( hr = IDirectSound_CreateSoundBuffer(lpDirectSound,&dsBD,&lpDirectSoundBuffer,
                                                        NULL ) ) )
    {
        hr = lpDirectMusicPort->SetDirectSound(lpDirectSound,lpDirectSoundBuffer);
    }

    if(lpDirectSoundBuffer && FAILED(hr) )
    {
        lpDirectSoundBuffer->Release();
        lpDirectSoundBuffer = NULL;
    }
	
	lpDirectSoundBuffer->QueryInterface(
		IID_IDirectSound3DBuffer,
		(void **)&lpDirectSound3DBuffer
	);
	
	//lpDirectSound3DBuffer->SetMode(DS3DMODE_HEADRELATIVE,DS3D_DEFERRED);
	lpDirectSound3DBuffer->SetMode(DS3DMODE_NORMAL,DS3D_DEFERRED);
}

void MidiPlayer::setMode(DWORD mode){
	lpDirectSound3DBuffer->SetMode(mode,DS3D_DEFERRED);
}

void MidiPlayer::initDirectMusicPerformance(){
    if(!lpDirectMusic){
		Debugger::println("DirectMusic is null.");
		return;
	}

    IDirectMusicPerformance* pPerf = NULL;

    if( SUCCEEDED(CoCreateInstance( CLSID_DirectMusicPerformance,NULL,
                                    CLSCTX_INPROC, IID_IDirectMusicPerformance,
                                   (VOID**)&pPerf ) ) )
    {
        if( SUCCEEDED(pPerf->Init( &lpDirectMusic, NULL, NULL ) ) )
        {
            // Set autodownloading to be on
            BOOL fAutoDownload = TRUE;
            if( FAILED( pPerf->SetGlobalParam( GUID_PerfAutoDownload,
                                               &fAutoDownload,
                                               sizeof(fAutoDownload) ) ) )
            {
                pPerf->Release();
                pPerf = NULL;
            }
        }
    }
    else
    {
        pPerf = NULL;
    }

	this->lpDirectMusicPerformance=pPerf;
}

void MidiPlayer::play(MidiData *midiData){
	play(midiData,(boolean)true);
}

void MidiPlayer::play(MidiData *midiData,boolean isLoop){
	if(isLoop){
		play(midiData,(DWORD)100000);
	}else{
		play(midiData,(DWORD)1);
	}
}

void MidiPlayer::play(MidiData *midiData,DWORD times){
	midiData->getDirectMusicSegment()->SetRepeats(times);
	
	if(isPlaying()){
		stop();
	}
	
	if(SUCCEEDED(lpDirectMusicPort->Activate(TRUE))){
        if(SUCCEEDED(lpDirectMusicPerformance->PlaySegment(midiData->getDirectMusicSegment(),0,0,NULL))){
			//midiData->lpDirectMusicSegment->Release();
			//midiData->lpDirectMusicSegment = NULL;
			this->playing=true;
		}
    }
}

void MidiPlayer::stop(){
	if(SUCCEEDED(lpDirectMusicPort->Activate(TRUE))){
		lpDirectMusicPerformance->Stop(NULL,NULL,0,NULL);
		this->playing=false;
	}
}

boolean MidiPlayer::isPlaying(){
	return playing;
}

MidiData* MidiPlayer::readMidiData(String *path,String *fileName){
	if(!lpDirectMusicLoader){
		Debugger::println("DirectMusicLoader is null");
		return NULL;
	}
	
    IDirectMusicSegment* pSegment = NULL;
//    char                 szDir[_MAX_PATH];
    WCHAR                wszDir[_MAX_PATH];
    DMUS_OBJECTDESC      ObjDesc;

	mbstowcs(wszDir,path->toCharArray(),_MAX_PATH);
    lpDirectMusicLoader->SetSearchDirectory(GUID_DirectMusicAllTypes,wszDir,FALSE);
	
    ObjDesc.guidClass   = CLSID_DirectMusicSegment;
    ObjDesc.dwSize      = sizeof(DMUS_OBJECTDESC);
    ObjDesc.dwValidData = DMUS_OBJ_CLASS | DMUS_OBJ_FILENAME;
    mbstowcs(ObjDesc.wszFileName,fileName->toCharArray(),DMUS_MAX_FILENAME);
	
	HRESULT hr;
    hr=lpDirectMusicLoader->GetObject(&ObjDesc,IID_IDirectMusicSegment,(VOID**)&pSegment);
	
	if(hr!=S_OK){
		Debugger::printErrorDMusic(hr);
		return NULL;
	}
	
	return new MidiData(pSegment);
}

void MidiPlayer::setPosition(Vector3 *position){
	lpDirectSound3DBuffer->SetPosition(
		position->getX(),position->getY(),position->getZ(),
		DS3D_DEFERRED
	);
}

void MidiPlayer::setVelocity(Vector3 *velocity){
	this->lpDirectSound3DBuffer->SetVelocity(
		velocity->getX(),velocity->getY(),velocity->getZ(),
		DS3D_DEFERRED
	);
}

void MidiPlayer::setOrientation(Vector3 *direction){
	this->lpDirectSound3DBuffer->SetConeOrientation(
		direction->getX(),direction->getY(),direction->getZ(),
		DS3D_DEFERRED
	);
}

void MidiPlayer::setVolume(long volume){
	lpDirectSoundBuffer->SetVolume(volume);
}
