/*	 SoundData3D NX		Ver 0.00 */
/*									*/
/*									*/
/*	 by toyo		1999/5/16		*/

#include "basic.h"

#include "lib/io/InputStream.h"
#include "lib/io/OutputStream.h"
#include "lib/io/ObjectInputStream.h"
#include "lib/io/ObjectOutputStream.h"
#include "lib/io/Serializable.h"

#include "SoundBuffer.h"
#include "WaveFormat.h"
#include "SoundData.h"
#include "lib/math/Vector3.h"
#include "SoundData3D.h"

SoundData3D::SoundData3D(void){
}

SoundData3D::SoundData3D(SoundData *soundData):
	SoundData(soundData->soundBuffer,soundData->waveFormat,soundData->duplicateSoundNum){
	
	this->position=new Vector3*[duplicateSoundNum];
	this->velocity=new Vector3*[duplicateSoundNum];
	this->orientation=new Vector3*[duplicateSoundNum];
	
	int i;
	for(i=0;i<duplicateSoundNum;i++){
		this->position[i]=new Vector3();
		this->velocity[i]=new Vector3();
		this->orientation[i]=new Vector3(0.0f,0.0f,1.0f);
	}
	
	this->isPositionValid=new boolean[duplicateSoundNum];
	this->isVelocityValid=new boolean[duplicateSoundNum];
	this->isOrientationValid=new boolean[duplicateSoundNum];
	
	for(i=0;i<duplicateSoundNum;i++){
		this->isPositionValid[i]=false;
		this->isVelocityValid[i]=false;
		this->isOrientationValid[i]=false;
	}

	this->lpDirectSound3DBuffer=new LPDIRECTSOUND3DBUFFER[duplicateSoundNum];
	for(i=0;i<duplicateSoundNum;i++){
		lpDirectSound3DBuffer[i]=NULL;
	}
}
	
SoundData3D::SoundData3D(
		SoundBuffer *soundBuffer,
		WaveFormat *waveFormat
	):SoundData(soundBuffer,waveFormat){
	
	this->position=new Vector3*[duplicateSoundNum];
	this->velocity=new Vector3*[duplicateSoundNum];
	this->orientation=new Vector3*[duplicateSoundNum];
	
	int i;
	for(i=0;i<duplicateSoundNum;i++){
		this->position[i]=new Vector3();
		this->velocity[i]=new Vector3();
		this->orientation[i]=new Vector3(0.0f,0.0f,1.0f);
	}
	
	this->isPositionValid=new boolean[duplicateSoundNum];
	this->isVelocityValid=new boolean[duplicateSoundNum];
	this->isOrientationValid=new boolean[duplicateSoundNum];
	
	for(i=0;i<duplicateSoundNum;i++){
		this->isPositionValid[i]=false;
		this->isVelocityValid[i]=false;
		this->isOrientationValid[i]=false;
	}

	this->lpDirectSound3DBuffer=new LPDIRECTSOUND3DBUFFER[duplicateSoundNum];
	for(i=0;i<duplicateSoundNum;i++){
		lpDirectSound3DBuffer[i]=NULL;
	}
}

/*
SoundData3D::SoundData3D(
	SoundBuffer *soundBuffer,
	WaveFormat *waveFormat,
	DS3DBUFFER desc3DBuffer
	){
	this->soundBuffer=soundBuffer;
	this->waveFormat=waveFormat;
	this->desc3DBuffer=desc3DBuffer;
}
*/

SoundData3D::~SoundData3D(void){
}

/*
DS3DBUFFER SoundData3D::getDS3DBUFFER(void){
	return this->desc3DBuffer;
}

void SoundData3D::setDS3DBUFFER(DS3DBUFFER desc3DBuffer){
	this->desc3DBuffer=desc3DBuffer;
}
*/

LPDIRECTSOUND3DBUFFER SoundData3D::getDirectSound3DBuffer(){
	return this->getDirectSound3DBuffer(0);
}

void SoundData3D::setDirectSound3DBuffer(LPDIRECTSOUND3DBUFFER lpDirectSound3DBuffer){
	this->setDirectSound3DBuffer(lpDirectSound3DBuffer,0);
}

LPDIRECTSOUND3DBUFFER SoundData3D::getDirectSound3DBuffer(int index){
	return this->lpDirectSound3DBuffer[index];
}

void SoundData3D::setDirectSound3DBuffer(LPDIRECTSOUND3DBUFFER lpDirectSound3DBuffer,int index){
	this->lpDirectSound3DBuffer[index]=lpDirectSound3DBuffer;
}

void SoundData3D::setConeAngles(DWORD dwInsideConeAngle,DWORD dwOutsideConeAngle){
	HRESULT hr;
	
	for(int i=0;i<duplicateSoundNum;i++){
		hr=this->lpDirectSound3DBuffer[i]->SetConeAngles(dwInsideConeAngle,dwOutsideConeAngle,DS3D_DEFERRED);
		if(hr!=DS_OK){
			Debugger::println("SoundData3D::setConeAngles()");
			Debugger::printErrorDSound(hr);
		}
	}
}

void SoundData3D::setConeOrientationSub(Vector3 *orientation,int index){
	HRESULT hr;
	
	hr=this->lpDirectSound3DBuffer[index]->SetConeOrientation(
		orientation->getX(),
		orientation->getY(),
		orientation->getZ(),
		DS3D_DEFERRED
	);
	
	if(hr!=DS_OK){
		Debugger::println("SoundData3D::setConeOrientation()");
		Debugger::printErrorDSound(hr);
	}
}

void SoundData3D::setConeOrientation(Vector3 *orientation){
	setConeOrientation(orientation);
}

void SoundData3D::setConeOrientation(Vector3 *orientation,int index){
	this->orientation[index]->set(orientation);
	this->isOrientationValid[index]=true;
}

void SoundData3D::setConeOutsideVolume(long lConeOutsideVolume){
	HRESULT hr;
	
	for(int i=0;i<duplicateSoundNum;i++){
		hr=this->lpDirectSound3DBuffer[i]->SetConeOutsideVolume(lConeOutsideVolume,DS3D_DEFERRED);
		if(hr!=DS_OK){
			Debugger::println("SoundData3D::setConeOutsideVolume()");
			Debugger::printErrorDSound(hr);
		}
	}
}

void SoundData3D::setMaxDistance(D3DVALUE flMaxDistance){
	HRESULT hr;
	
	for(int i=0;i<duplicateSoundNum;i++){
		hr=this->lpDirectSound3DBuffer[i]->SetMaxDistance(flMaxDistance,DS3D_DEFERRED);
		if(hr!=DS_OK){
			Debugger::println("SoundData3D::setMaxDistance()");
			Debugger::printErrorDSound(hr);
		}
	}
}

void SoundData3D::setMinDistance(D3DVALUE flMinDistance){
	HRESULT hr;
	
	for(int i=0;i<duplicateSoundNum;i++){
		hr=this->lpDirectSound3DBuffer[i]->SetMinDistance(flMinDistance,DS3D_DEFERRED);
		if(hr!=DS_OK){
			Debugger::println("SoundData3D::setMinDistance()");
			Debugger::printErrorDSound(hr);
		}
	}
}

void SoundData3D::setMode(DWORD dwMode){
	HRESULT hr;
	
	for(int i=0;i<duplicateSoundNum;i++){
		hr=this->lpDirectSound3DBuffer[i]->SetMode(dwMode,DS3D_DEFERRED);
		if(hr!=DS_OK){
			Debugger::println("SoundData3D::setMode()");
			Debugger::printErrorDSound(hr);
		}
	}
}

void SoundData3D::validateSettings(){
	this->validateSettings(0);
}

void SoundData3D::validateSettings(int index){
	if(this->isPositionValid[index]){
		setPositionSub(this->position[index],index);	
		this->isPositionValid[index]=false;
	}

	if(this->isVelocityValid[index]){
		setVelocitySub(this->velocity[index],index);
		this->isVelocityValid[index]=false;
	}

	if(this->isOrientationValid[index]){
		setConeOrientationSub(this->orientation[index],index);
		this->isOrientationValid[index]=false;
	}
}

void SoundData3D::setPositionSub(Vector3 *position,int index){
	HRESULT hr;
	//this->desc3DBuffer.vPosition=position->getVector();
	hr=this->lpDirectSound3DBuffer[index]->SetPosition(
		position->getX(),
		position->getY(),
		position->getZ(),
		DS3D_DEFERRED
	);
	if(hr!=DS_OK){
		Debugger::println("SoundData3D::setPosition()");
		Debugger::printErrorDSound(hr);
	}
}

void SoundData3D::setPosition(Vector3 *position,int index){
	this->position[index]->set(position);
	this->isPositionValid[index]=true;
}

void SoundData3D::setPosition(Vector3 *position){
	setPosition(position,0);
}

void SoundData3D::setVelocitySub(Vector3 *velocity,int index){
	HRESULT hr;
	//this->desc3DBuffer.vVelocity=velocity->getVector();
	hr=this->lpDirectSound3DBuffer[index]->SetVelocity(
		velocity->getX(),
		velocity->getY(),
		velocity->getZ(),
		DS3D_DEFERRED
	);

	if(hr!=DS_OK){
		Debugger::println("SoundData3D::setVelocity()");
		Debugger::printErrorDSound(hr);
	}
}

void SoundData3D::setVelocity(Vector3 *velocity,int index){
	this->velocity[index]->set(velocity);
	this->isVelocityValid[index]=true;
}

void SoundData3D::setVelocity(Vector3 *velocity){
	this->setVelocity(velocity,0);
}
