/* SoundPlayer NX	 Ver 0.10   */
/*									*/
/*									*/
/*	 by toyo		1999/5/16	    */

#include "basic.h"

#include "lib/io/InputStream.h"
#include "lib/io/OutputStream.h"
#include "lib/io/ObjectInputStream.h"
#include "lib/io/ObjectOutputStream.h"
#include "lib/io/Serializable.h"

#include "lib/util/PooledObject.h"
#include "lib/util/ObjectPool.h"
#include "lib/math/Vector3.h"

#include "WaveFormat.h"
#include "SoundBuffer.h"
#include "SoundData.h"

#include "SoundRecorder.h"
#include "DirectSoundRecorder.h"
#include "WaveformAudioRecorder.h"
#include "StreamSoundPlayer.h"

#include "DirectSoundManager.h"
#include "SoundData3D.h"
#include "SoundListener3D.h"
#include "SoundPlayer.h"

void SoundPlayer::play(
	LPDIRECTSOUNDBUFFER lpDirectSoundBuffer,
	SoundData *soundData,
	boolean isLoop
	){
	LPVOID lpvPtr1; 
	DWORD dwBytes1; 
	LPVOID lpvPtr2; 
	DWORD dwBytes2; 
	
	HRESULT hr;
	
	lpDirectSoundBuffer->SetVolume(DSBVOLUME_MAX);
	
	//if(soundData->isFirstPlay() || !soundData->isStatic()){
	if(soundData->isFirstPlay()){
		hr=lpDirectSoundBuffer->Lock(
			0,soundData->soundBuffer->getBufferSize(),
			&lpvPtr1,&dwBytes1,
			&lpvPtr2,&dwBytes2,
			0
		);
		
		if(DSERR_BUFFERLOST==hr){ 
			hr=lpDirectSoundBuffer->Lock(
				0,soundData->soundBuffer->getBufferSize(),
				&lpvPtr1,&dwBytes1,
				&lpvPtr2,&dwBytes2,
				0
			);
		} 
	
		if(hr!=DS_OK){
			Debugger::println("SoundPlayer::play()");
			Debugger::printErrorDSound(hr);
			return;
		}
	
		memmove(lpvPtr1,soundData->soundBuffer->getBuffer(),dwBytes1);
		if(lpvPtr2!=NULL){
			memmove(lpvPtr2,soundData->soundBuffer->getBuffer()+dwBytes1,dwBytes2);
		}

		hr=lpDirectSoundBuffer->Unlock(lpvPtr1,dwBytes1,lpvPtr2,dwBytes2);
		if(hr!=DS_OK){
			Debugger::printErrorDSound(hr);
		}
		
		soundData->setFirstPlay(false);

		if(isLoop){
			hr=lpDirectSoundBuffer->Play(0,0,DSBPLAY_LOOPING);
		}else{
			hr=lpDirectSoundBuffer->Play(0,0,0);
		}

		if(hr!=DS_OK){
			Debugger::printErrorDSound(hr);
		}
	}else{
				
		if(isLoop){
			hr=lpDirectSoundBuffer->Play(0,0,DSBPLAY_LOOPING);
		}else{
			hr=lpDirectSoundBuffer->Play(0,0,0);
		}
		
		if(hr!=DS_OK){
			Debugger::printErrorDSound(hr);
		}
	}	
}

boolean SoundPlayer::isPlaying(LPDIRECTSOUNDBUFFER lpDirectSoundBuffer){
	DWORD dwCurrentPlayCursor;
	DWORD dwCurrentWriteCursor;
	
	lpDirectSoundBuffer->GetCurrentPosition(
		&dwCurrentPlayCursor,
		&dwCurrentWriteCursor
	);
	
	if(dwCurrentWriteCursor>0)return true;
	return false;
}

boolean SoundPlayer::isPlaying(SoundData *soundData){
	return isPlaying(soundData,0);
}

boolean SoundPlayer::isPlaying(SoundData *soundData,int index){
	LPDIRECTSOUNDBUFFER lpDirectSoundBuffer=soundData->getDirectSoundBuffer(index);
	if(lpDirectSoundBuffer==NULL){
		return false;
	}else{
		return isPlaying(lpDirectSoundBuffer);
	}
}

void SoundPlayer::play(SoundData *soundData,boolean isLoop){
	play(soundData,isLoop,0);
}

void SoundPlayer::play(SoundData *soundData,boolean isLoop,int index){
	LPDIRECTSOUNDBUFFER lpDirectSoundBuffer=soundData->getDirectSoundBuffer(index);
	if(lpDirectSoundBuffer!=NULL){
		play(lpDirectSoundBuffer,soundData,isLoop);
	}
}

void SoundPlayer::play3D(
	SoundData3D *soundData3D,
	SoundListener3D *soundListener3D,
	boolean isLoop
	){
	play3D(soundData3D,soundListener3D,isLoop,0);
}

void SoundPlayer::play3D(
	SoundData3D *soundData3D,
	SoundListener3D *soundListener3D,
	boolean isLoop,
	int index
	){
	LPDIRECTSOUNDBUFFER lpDirectSoundBuffer=soundData3D->getDirectSoundBuffer(index);
	if(lpDirectSoundBuffer!=NULL){	
		play(lpDirectSoundBuffer,(SoundData*)soundData3D,isLoop);
	}
}

void SoundPlayer::stop(SoundData *soundData,int index){
	if(isPlaying(soundData,index)){
		soundData->getDirectSoundBuffer(index)->Stop();
	}
}
