/* StreamSoundPlayer NX  Ver 0.00  */
/*                                     */
/*                                     */
/*   by toyo    1999/4/10              */

#include "basic.h"

#include "lib/io/InputStream.h"
#include "lib/io/OutputStream.h"
#include "lib/io/ObjectInputStream.h"
#include "lib/io/ObjectOutputStream.h"
#include "lib/io/Serializable.h"

#include "WaveFormat.h"
#include "SoundBuffer.h"
#include "StreamSoundPlayer.h"

StreamSoundPlayer::StreamSoundPlayer(
	LPDIRECTSOUNDBUFFER lpDirectSoundBuffer,
	SoundBuffer *soundBuffer){
	this->lpDirectSoundBuffer=lpDirectSoundBuffer;
	this->soundBuffer=soundBuffer;
	
	blockState=new STATE[soundBuffer->getBlockNum()];
	for(int i=0;i<soundBuffer->getBlockNum();i++){
		blockState[i]=EMPTY;
	}
}

StreamSoundPlayer::StreamSoundPlayer(
	LPDIRECTSOUNDBUFFER lpDirectSoundBuffer,
	DWORD dwUncompressionRate,
	SoundBuffer *soundBuffer){
	this->lpDirectSoundBuffer=lpDirectSoundBuffer;
	this->soundBuffer=soundBuffer;
	
	blockState=new STATE[soundBuffer->getBlockNum()];
	for(int i=0;i<soundBuffer->getBlockNum();i++){
		blockState[i]=EMPTY;
	}
	
	DWORD dwFrequency;
	lpDirectSoundBuffer->GetFrequency(&dwFrequency);
	lpDirectSoundBuffer->SetFrequency(dwFrequency/dwUncompressionRate);
}

StreamSoundPlayer::~StreamSoundPlayer(){
	this->lpDirectSoundBuffer->Release();
}



//
//public method
//

LPDIRECTSOUNDBUFFER StreamSoundPlayer::getDirectSoundBuffer(){
	return this->lpDirectSoundBuffer;
}


boolean StreamSoundPlayer::start(){
	HRESULT hr=lpDirectSoundBuffer->Play(0,0,DSBPLAY_LOOPING);
	if(hr!=DS_OK){
		Debugger::printErrorDSound(hr);
		return false;
	}
	return true;
}

boolean StreamSoundPlayer::appendBlock(DWORD dwBlockNo){
	DWORD dwCurrentWriteCursor;
	
	/*
	Debugger::println("appendBlock");
	Debugger::println((int)dwBlockNo);
	*/
	  
	lpDirectSoundBuffer->GetCurrentPosition(
		NULL,
		&dwCurrentWriteCursor
	);
	
	DWORD dwCurrentWriteBlock=dwCurrentWriteCursor/soundBuffer->getBlockSize();
	DWORD temp;
	
	/*
	Debugger::println("dwCurrentWriteBlock:");
	Debugger::println((int)dwCurrentWriteBlock);
	*/

	for(int i=0;i<soundBuffer->getBlockNum();i++){
		temp=(i+dwCurrentWriteBlock)%soundBuffer->getBlockNum();
		/*
		Debugger::println("temp:");
		Debugger::println((int)temp);
		Debugger::println((int)blockState[temp]);
		*/
		if(blockState[temp]==EMPTY){
			if(writeBlock(dwBlockNo,temp)){
				blockState[temp]=FULL;
				return true;
			}
		}
	}
	
	return false;
}

void StreamSoundPlayer::processPlaying(){
	DWORD dwCurrentPlayCursor;
	DWORD dwCurrentWriteCursor;
	
	lpDirectSoundBuffer->GetCurrentPosition(
		&dwCurrentPlayCursor,
		&dwCurrentWriteCursor
	);
	
	DWORD dwCurrentPlayBlock;
	DWORD dwCurrentWriteBlock;
	DWORD dwBlockSize=soundBuffer->getBlockSize();
	
	dwCurrentPlayBlock=dwCurrentPlayCursor/dwBlockSize;
	dwCurrentWriteBlock=dwCurrentWriteCursor/dwBlockSize;
	
	//step 1
	//
	//BUSY łȂȂubN|ĉ(EMPTY ɂ)B
	//
	if(dwCurrentPlayBlock==dwCurrentWriteBlock){
		for(int i=0;i<soundBuffer->getBlockNum();i++){
			if(i!=dwCurrentPlayBlock && blockState[i]==BUSY){
				/*
				Debugger::println("N[");
				Debugger::println(i);
				*/
				cleanupBlock((DWORD)i);
				blockState[i]=EMPTY;
			}
		}
	}else{
		for(int i=0;i<soundBuffer->getBlockNum();i++){
			if(i<dwCurrentPlayBlock && 
				i>dwCurrentWriteBlock &&
				blockState[i]==BUSY){
				cleanupBlock((DWORD)i);
				blockState[i]=EMPTY;
			}
		}
	}
	
	if(dwLastPlayBlock<dwCurrentPlayBlock){
		for(int i=dwLastPlayBlock;i<dwCurrentPlayBlock;i++){
			cleanupBlock((DWORD)i);
			blockState[i]=EMPTY;
		}
	}else if(dwLastPlayBlock>dwCurrentPlayBlock){
		for(int i=dwLastPlayBlock;i<soundBuffer->getBlockNum();i++){
			cleanupBlock((DWORD)i);
			blockState[i]=EMPTY;
		}

		for(i=0;i<dwCurrentPlayBlock;i++){
			cleanupBlock((DWORD)i);
			blockState[i]=EMPTY;
		}
	}


	//step 2
	//
	//BUSYȃubNBUSYtO ZbgB
	//
	if(dwCurrentPlayBlock==dwCurrentWriteBlock){
		blockState[dwCurrentPlayBlock]=BUSY;
	}else if(dwCurrentPlayBlock<dwCurrentWriteBlock){
		for(int i=dwCurrentPlayBlock;i<dwCurrentWriteBlock-dwCurrentPlayBlock+1;i++){
			blockState[i]=BUSY;
		}
	}else{
		for(int i=0;i<dwCurrentWriteBlock+1;i++){
			blockState[i]=BUSY;
		}

		for(i=dwCurrentPlayBlock;i<soundBuffer->getBlockNum();i++){
			blockState[i]=BUSY;
		}
	}
	
	this->dwLastPlayBlock=dwCurrentPlayBlock;

	char state[32];
	
	for(int i=0;i<32;i++){
		switch(blockState[i]){
		case EMPTY:
			state[i]='E';
			break;
		case FULL:
			state[i]='F';
			break;
		case BUSY:
			state[i]='B';
			break;
		}
	}
	state[32]='\0';

	Debugger::println(state);
	Debugger::println("\n");
}


boolean StreamSoundPlayer::stop(){
	HRESULT hr=lpDirectSoundBuffer->Stop();
	if(hr!=DS_OK){
		Debugger::printErrorDSound(hr);
		return false;
	}else{
		return true;
	}
}



//
//private method
//

boolean StreamSoundPlayer::writeBlock(DWORD dwFrom,DWORD dwTo){
	LPVOID lpvPtr1; 
    DWORD dwBytes1; 
    LPVOID lpvPtr2; 
    DWORD dwBytes2; 
	
	HRESULT hr;
	
	/*
	Debugger::println("dwFrom:");
	Debugger::println((int)dwFrom);
	Debugger::println("dwTo:");
	Debugger::println((int)dwTo);
	*/

	hr=lpDirectSoundBuffer->Lock(
		dwTo*soundBuffer->getBlockSize(),
		soundBuffer->getBlockSize(),
		&lpvPtr1,&dwBytes1,
		&lpvPtr2,&dwBytes2,
		0
	);
	
	if(DSERR_BUFFERLOST==hr){ 
		hr=lpDirectSoundBuffer->Lock(
			dwTo*soundBuffer->getBlockSize(),
			soundBuffer->getBlockSize(),
			&lpvPtr1,&dwBytes1,
			&lpvPtr2,&dwBytes2,
			0
		);
    } 
	
	if(hr!=DS_OK){
		Debugger::printErrorDSound(hr);
		return false;
	}
	
	memmove(lpvPtr1,soundBuffer->getBlock(dwFrom),dwBytes1);
	if(lpvPtr2!=NULL){
		memmove(lpvPtr2,soundBuffer->getBlock(dwFrom)+dwBytes1,dwBytes2);
	}
	
	hr=lpDirectSoundBuffer->Unlock(lpvPtr1,dwBytes1,lpvPtr2,dwBytes2);
	if(hr!=DS_OK){
		Debugger::printErrorDSound(hr);
		return false;
	}
		
	return true;
}

boolean StreamSoundPlayer::cleanupBlock(DWORD dwBlockNo){
	HRESULT hr;
	
	LPVOID lpvPtr1; 
    DWORD dwBytes1; 
    LPVOID lpvPtr2; 
    DWORD dwBytes2; 
	
	hr=lpDirectSoundBuffer->Lock(
		dwBlockNo*soundBuffer->getBlockSize(),
		soundBuffer->getBlockSize(),
		&lpvPtr1,&dwBytes1,
		&lpvPtr2,&dwBytes2,
		0
	);
	
	if(DSERR_BUFFERLOST==hr){ 
		hr=lpDirectSoundBuffer->Lock(
			dwBlockNo*soundBuffer->getBlockSize(),
			soundBuffer->getBlockSize(),
			&lpvPtr1,&dwBytes1,
			&lpvPtr2,&dwBytes2,
			0
		);
    } 
	
	if(hr!=DS_OK){
		Debugger::printErrorDSound(hr);
		return false;
	}
	
	memset(lpvPtr1,0,dwBytes1);
	if(lpvPtr2!=NULL){
		memset(lpvPtr2,0,dwBytes2);
	}
	
	return true;
}




