/*   WaveFormat NX   Ver 0.00 */
/*                                */
/*                                */
/*   by toyo    1999/4/8          */

#include "basic.h"

#include "lib/io/InputStream.h"
#include "lib/io/OutputStream.h"
#include "lib/io/ObjectInputStream.h"
#include "lib/io/ObjectOutputStream.h"
#include "lib/io/Serializable.h"

#include "WaveFormat.h"

WaveFormat::WaveFormat(void){
}

WaveFormat::WaveFormat(WAVEFORMATEX waveFormatEx){
	this->waveFormatEx=waveFormatEx;
}

WaveFormat::WaveFormat(
	WORD wFormatTag,
    WORD nChannels,
    DWORD nSamplesPerSec, 
    DWORD nAvgBytesPerSec, 
    WORD nBlockAlign,
    WORD wBitsPerSample, 
    WORD cbSize
  ){

  initInstance(
    wFormatTag,
    nChannels,
    nSamplesPerSec,
    nAvgBytesPerSec,
    nBlockAlign,
    wBitsPerSample,
    cbSize
  );
}

WaveFormat::WaveFormat(
  WORD nChannels,
  DWORD nSamplesPerSec,
  WORD nBlockAlign,
  WORD wBitsPerSample){
  initInstance(
    WAVE_FORMAT_PCM,
    nChannels,
    nSamplesPerSec,
    nSamplesPerSec*nBlockAlign,
    nBlockAlign,
    wBitsPerSample,
    0
  );
}

void WaveFormat::initInstance(
  WORD wFormatTag,
    WORD nChannels,
    DWORD nSamplesPerSec, 
    DWORD nAvgBytesPerSec, 
    WORD nBlockAlign,
    WORD wBitsPerSample, 
    WORD cbSize
  ){
  memset(&waveFormatEx,0,sizeof(WAVEFORMATEX));
  waveFormatEx.wFormatTag=wFormatTag;
  waveFormatEx.nChannels=nChannels;
  waveFormatEx.nSamplesPerSec=nSamplesPerSec;
  waveFormatEx.nAvgBytesPerSec=nAvgBytesPerSec;
  waveFormatEx.nBlockAlign=nBlockAlign;
  waveFormatEx.wBitsPerSample=wBitsPerSample;
  waveFormatEx.cbSize=cbSize;
}

WaveFormat* WaveFormat::clone(void){
	return new WaveFormat(waveFormatEx);
}

WORD WaveFormat::getFormatTag(void){
  return waveFormatEx.wFormatTag;
}

WORD WaveFormat::getNumOfChannels(void){
  return waveFormatEx.nChannels;
}

DWORD WaveFormat::getNumOfSamplesPerSec(void){
  return waveFormatEx.nSamplesPerSec;
}

DWORD WaveFormat::getNumOfAvgBytesPerSec(void){
  return waveFormatEx.nAvgBytesPerSec;
}

WORD WaveFormat::getNumOfBlockAlign(void){
  return waveFormatEx.nBlockAlign;
}

WORD WaveFormat::getBitsPerSample(void){
  return waveFormatEx.wBitsPerSample;
}

WORD WaveFormat::getCbSize(void){
  return waveFormatEx.cbSize;
}
  
void WaveFormat::setFormatTag(WORD wFormatTag){
  waveFormatEx.wFormatTag=wFormatTag;
}

void WaveFormat::setNumOfChannels(WORD nChannels){
  waveFormatEx.nChannels=nChannels;
}

void WaveFormat::setNumOfSamplesPerSec(DWORD nSamplesPerSec){
  waveFormatEx.nSamplesPerSec=nSamplesPerSec;
}

void WaveFormat::setNumOfAvgBytesPerSec(DWORD nAvgBytesPerSec){
  waveFormatEx.nAvgBytesPerSec=nAvgBytesPerSec;
}

void WaveFormat::setNumOfBlockAlign(WORD nBlockAlign){
  waveFormatEx.nBlockAlign=nBlockAlign;
}

void WaveFormat::setBitsPerSample(WORD wBitsPerSample){
  waveFormatEx.wBitsPerSample=wBitsPerSample;
}

void WaveFormat::setCbSize(WORD cbSize){
  waveFormatEx.cbSize=cbSize;
}

WaveFormat* WaveFormat::newInstance(int id){
	return new WaveFormat();
}

int WaveFormat::getTypeID(void){
	return 0;
}

void WaveFormat::readObject(ObjectInputStream* in){
	byte buffer[sizeof(WAVEFORMATEX)];
	
	for(int i=0;i<sizeof(WAVEFORMATEX);i++){
		buffer[i]=in->readByte();
	}
	
	memmove(&this->waveFormatEx,buffer,sizeof(WAVEFORMATEX));
}

void WaveFormat::writeObject(ObjectOutputStream* out){
	byte buffer[sizeof(WAVEFORMATEX)];
	
	for(int i=0;i<sizeof(WAVEFORMATEX);i++){
		memmove(buffer,&waveFormatEx,sizeof(WAVEFORMATEX));
	}
	
	for(i=0;i<sizeof(WAVEFORMATEX);i++){
		out->writeByte(buffer[i]);
	}
}

