/* WaveformAudioRecorder NX  Ver 0.00  */
/*                                         */
/*                                         */
/*   by toyo    1999/4/8                   */

#include "basic.h"

#include "lib/io/InputStream.h"
#include "lib/io/OutputStream.h"
#include "lib/io/ObjectInputStream.h"
#include "lib/io/ObjectOutputStream.h"
#include "lib/io/Serializable.h"

#include "WaveFormat.h"
#include "SoundBuffer.h"
#include "SoundRecorder.h"
#include "WaveformAudioRecorder.h"

WaveformAudioRecorder::WaveformAudioRecorder(
	HWAVEIN hWaveIn,
	SoundBuffer *soundBuffer
	){
	initInstance(hWaveIn,false,0,soundBuffer);
}

WaveformAudioRecorder::WaveformAudioRecorder(
	HWAVEIN hWaveIn,
	DWORD dwCompressionRate,
	SoundBuffer *soundBuffer
	){
	initInstance(hWaveIn,true,dwCompressionRate,soundBuffer);
}

void WaveformAudioRecorder::initInstance(
	HWAVEIN hWaveIn,
	boolean isCompressionMode,
	DWORD dwCompressionRate,
	SoundBuffer *soundBuffer
	){
	this->hWaveIn=hWaveIn;
	this->isCompressionMode=isCompressionMode;
	this->dwCompressionRate=dwCompressionRate;
	this->soundBuffer=soundBuffer;
	
	this->waveHeader=new WAVEHDR[soundBuffer->getBlockNum()];
		
	if(isCompressionMode){
		subBuffer=new SoundBuffer(
			soundBuffer->getBlockSize()*dwCompressionRate,
			soundBuffer->getBlockNum()
		);
	}
	
	for(int i=0;i<soundBuffer->getBlockNum();i++){
		if(isCompressionMode){
			waveHeader[i].lpData=subBuffer->getBlock(i);
			waveHeader[i].dwBufferLength=subBuffer->getBlockSize();
		}else{
			waveHeader[i].lpData=soundBuffer->getBlock(i);
			waveHeader[i].dwBufferLength=soundBuffer->getBlockSize();
		}
	}
}

WaveformAudioRecorder::~WaveformAudioRecorder(){
	MMRESULT mmr=waveInReset(hWaveIn);
	if(mmr!=MMSYSERR_NOERROR){
		char buf[128];
		waveInGetErrorText(mmr,buf,128);
		Debugger::println(buf);
		Debugger::println("\n");
	}
	
	mmr=waveInStop(hWaveIn);
	if(mmr!=MMSYSERR_NOERROR){
		char buf[128];
		waveInGetErrorText(mmr,buf,128);
		Debugger::println(buf);
		Debugger::println("\n");
	}
	
	mmr=waveInClose(hWaveIn);
	if(mmr!=MMSYSERR_NOERROR){
		char buf[128];
		waveInGetErrorText(mmr,buf,128);
		Debugger::println(buf);
		Debugger::println("\n");
	}
	delete[] waveHeader;
}

boolean WaveformAudioRecorder::start(){
	MMRESULT mmr;
	
	for(int i=0;i<soundBuffer->getBlockNum();i++){
		waveHeader[i].dwBytesRecorded=0;
		waveHeader[i].dwUser=0;
		waveHeader[i].dwFlags=0;
		waveHeader[i].lpNext=NULL;
		waveHeader[i].reserved=0;
		
		mmr=waveInPrepareHeader(hWaveIn,&waveHeader[i],sizeof(WAVEHDR));
		if(mmr!=MMSYSERR_NOERROR){
			char buf[128];
			waveInGetErrorText(mmr,buf,128);
			Debugger::println(buf);
			Debugger::println("\n");
			return false;
		}
	}
	
	for(i=0;i<soundBuffer->getBlockNum();i++){
		mmr=waveInAddBuffer(hWaveIn,&waveHeader[i],sizeof(WAVEHDR));
		if(mmr!=MMSYSERR_NOERROR){
			char buf[128];
			waveInGetErrorText(mmr,buf,128);
			Debugger::println(buf);
			Debugger::println("\n");
			return false;
		}
	}
	this->currentBlockNo=0;
	
	mmr=waveInStart(hWaveIn);
	if(mmr!=MMSYSERR_NOERROR){
		char buf[128];
		waveInGetErrorText(mmr,buf,128);
		Debugger::println(buf);
		Debugger::println("\n");
		return false;
	}

	return true;
}

boolean WaveformAudioRecorder::processRecording(){
	if(waveHeader[currentBlockNo].dwFlags & WHDR_DONE){
		waveHeader[currentBlockNo].dwFlags=WHDR_PREPARED;
		waveHeader[currentBlockNo].dwBytesRecorded=0;
		
		if(isCompressionMode){
			for(int i=0;i<subBuffer->getBlockSize();i++){
				memmove(
					soundBuffer->getBlock(currentBlockNo)+i,
					waveHeader[currentBlockNo].lpData+i*dwCompressionRate,
					1
				);
			}
		}

		MMRESULT mmr=waveInAddBuffer(hWaveIn,&waveHeader[currentBlockNo],sizeof(WAVEHDR));
		if(mmr!=MMSYSERR_NOERROR){
			char buf[128];
			waveInGetErrorText(mmr,buf,128);
			Debugger::println(buf);
			Debugger::println("\n");
		}
		currentBlockNo=(currentBlockNo+1)%soundBuffer->getBlockNum();
		return true;
	}else{
		return false;
	}
}

boolean WaveformAudioRecorder::isProcessReady(){
	if(waveHeader[currentBlockNo].dwFlags & WHDR_DONE){
		return true;
	}else{
		return false;
	}
}

boolean WaveformAudioRecorder::stop(){
	for(int i=0;i<soundBuffer->getBlockNum();i++){
		waveInUnprepareHeader(hWaveIn,&waveHeader[i],sizeof(WAVEHDR));
	}
	
	MMRESULT mmr=waveInReset(hWaveIn);
	if(mmr!=MMSYSERR_NOERROR){
		char buf[128];
		waveInGetErrorText(mmr,buf,128);
		Debugger::println(buf);
		Debugger::println("\n");
	}
	
	mmr=waveInStop(hWaveIn);
	if(mmr!=MMSYSERR_NOERROR){
		char buf[128];
		waveInGetErrorText(mmr,buf,128);
		Debugger::println(buf);
		Debugger::println("\n");
		return false;
	}else{
		return true;
	}
}

int WaveformAudioRecorder::getCurrentBlockNo(){
	return currentBlockNo;
}

int WaveformAudioRecorder::getPastBlockNo(){
	return (soundBuffer->getBlockNum()+currentBlockNo-1)%soundBuffer->getBlockNum();
}

