/*   nbV\          Ver 0.20   */
/*                                  */
/*                                  */
/*   by Atsushi          1999/9/8   */
/*   by Y.Hayashi       1999/6/27   */


/*   ---   RXgN^   ---   */

template <class KeyType,class ValueType> Hashtable<KeyType,ValueType>
    ::Hashtable(int initialCapacity,float loadFactor)
{
	init(initialCapacity,loadFactor);
}

template <class KeyType,class ValueType> Hashtable<KeyType,ValueType>
    ::Hashtable(int initialCapacity)
{
	init(initialCapacity,0.75f);
}

template <class KeyType,class ValueType> Hashtable<KeyType,ValueType>
    ::Hashtable(void)
{
	init(101,0.75f);
}

template <class KeyType,class ValueType> void Hashtable<KeyType,ValueType>
    ::init(int initialCapacity,float loadFactor)
{
	if(initialCapacity<=0||loadFactor<=0.0){
		//error
	}
	this->loadFactor=loadFactor;
	length=initialCapacity;
	table=new Element*[initialCapacity];
	initArray();
	threshhold=(int)(initialCapacity*loadFactor);
	count=0;
}

template <class KeyType,class ValueType> void Hashtable<KeyType,ValueType>
    ::initArray(void)
{
	for(int i=0; i<length; i++)
	{
		table[i]=NULL;
	}
}


/*   ---      ---   */

template <class KeyType,class ValueType> void Hashtable<KeyType,ValueType>::
    clear(void)
{
	for(int index=length; --index>=0;)
	{
		table[index]=NULL;
	}
	count=0;
}

template <class KeyType,class ValueType> void Hashtable<KeyType,ValueType>::
    deleteAll(void)
{
	Element** elements;
	int pointer;
	int i;

	elements=new Element*[count];
	pointer=0;
	
	for (i=length-1;i>=0;i--)
	{
		for (Element* e=table[i];e!=NULL;e=e->next)
		{
			elements[pointer++]=e;
		}
	}
	
	for (i=0;i<=count-1;i++)
	{
		delete elements[i]->key;
		delete elements[i]->value;
	}
	delete[] elements;
	
	clear();
}


/*   ---   ǂݏ   ---   */

template <class KeyType,class ValueType> ValueType* Hashtable<KeyType,ValueType>::
    get(KeyType* key)
{
	int hash=key->getHashCode();
	int index=(hash&0x7FFFFFFF)%length;

	for (Element* e=table[index]; e!=NULL; e=e->next)
	{		
		if ((e->hash==hash)&&e->key->equals(key))
		{
			return e->value;
		}
	}
	
	return NULL;
}

template <class KeyType,class ValueType> ValueType* Hashtable<KeyType,ValueType>::
    put(KeyType* key,ValueType* value)
{
	if(value==NULL){
DEBUG("Error : Hashtable::put(NULL)");
		//error
	}

	int hash=key->getHashCode();
	int index=(hash&0x7FFFFFFF)%length;
	
	for(Element* e=table[index]; e!=NULL; e=e->next)
	{
		if((e->hash==hash)&&e->key->equals(key))
		{
			ValueType* old=e->value;
			e->value=value;
			return old;
		}
	}

	Element* newElement;

	newElement=new Element();
	newElement->hash=hash;
	newElement->key  =key;
	newElement->value=value;
	newElement->next=table[index];
	table[index]=newElement;
	count++;
	return NULL;
}


/*   ---   L   ---   */

template <class KeyType,class ValueType> boolean Hashtable<KeyType,ValueType>::
    containsKey(KeyType* key)
{
	int hash=key->getHashCode();
	int index=(hash&0x7FFFFFFF)%length;

	for (Element* e=table[index]; e!=NULL; e=e->next)
	{
		if((e->hash==hash)&&e->key->equals(key))
		{
			return TRUE;
		}
	}
	
	return FALSE;
}

/*
template <class KeyType,class ValueType> boolean Hashtable<KeyType,ValueType>::
    containsValue(ValueType* value)
{

	for (int i=length; i-->0; )
	{
		for (Element* e=table[i]; e!=NULL; e=e->next)
		{
			if(e->value->equals(value))
			{
				return TRUE;
			}
		}
	}
	
	return FALSE;
}
*/


/*   ---      ---   */

template <class KeyType,class ValueType> HashtableEnumeration<KeyType>*
    Hashtable<KeyType,ValueType>::getKeys(void)
{
	KeyType** keys;
	int pointer;

	keys=new KeyType*[count];
	pointer=0;
	
	for (int i=length-1;i>=0;i--)
	{
		for (Element* e=table[i];e!=NULL;e=e->next)
		{
			keys[pointer++]=e->key;
		}
	}
	
	return new HashtableEnumeration<KeyType>(keys,pointer);
}

template <class KeyType,class ValueType> HashtableEnumeration<ValueType>*
    Hashtable<KeyType,ValueType>::getElements(void)
{
	ValueType** elements;
	int pointer;

	elements=new ValueType*[count];
	pointer=0;
	
	for (int i=length-1;i>=0;i--)
	{
		for (Element* e=table[i];e!=NULL;e=e->next)
		{
			elements[pointer++]=e->value;
		}
	}
	
	return new HashtableEnumeration<ValueType>(elements,pointer);
}


/*   ---      ---   */

template <class KeyType,class ValueType> int Hashtable<KeyType,ValueType>::
    getSize(void)
{
	return count;
}

template <class KeyType,class ValueType> boolean Hashtable<KeyType,ValueType>::
    isEmpty(void)
{
	return count==0;
}


/*   ---   fXgN^   ---   */

template <class KeyType,class ValueType> Hashtable<KeyType,ValueType>::
    ~Hashtable(void)
{
	delete[] table;
}
